/* --------------------------------------------------------------------- */
/* This	 is  SERDEL  --	 A  Supervisor	for Executing Remote Distributed */
/* Experiments on a LAN. This  program	has  been  written  to	organize */
/* SMURPH  experiments,	 but  it  can  be  used in connection with other */
/* programs of a similar nature.					 */
/* --------------------------------------------------------------------- */

/* -------------------------------- */
/* Copyright (C) P. Gburzynski 1991 */
/* -------------------------------- */

#include	"serdel.h"

sigint () {

    closeFiles ();
    logMsg ("Serdel aborted\n");
    exit (1);
}

main (argc, argv) int argc; char *argv[]; {

    if (signal (SIGINT, SIG_IGN) != SIG_IGN) signal (SIGINT, sigint);
    if (argc > 1) {
	Verbose = 1;
	if (*(argv[1]) == 'V') Verbose++;
    } else
	Verbose = 0;
    while (1) {
	if (Verbose > 1) logMsg ("Awakened");
	processThings ();
	removeDefuncts ();
	sleepForAWhile ();
    }
}


sleepForAWhile () {

    if (Verbose > 1) logMsg ("Going to sleep");
    sleep (CHECKINTERVAL);
}

processThings () {

    ExModified = 0;
    if (readHostFile () && readExperimentsFile () && openTempFile ()) {
/*@@@logMsg ("Action sequence entered");*/
	monitorActive ();
	doCheckpointed ();
	tryStartingNew ();
	if (ExModified) writeExperimentsFile ();
    }
    deallocateStorage ();
    closeFiles ();
}

int	openTempFile () {

/* -------------------------------------------------- */
/* Creates a temporary file (for reading and writing) */
/* -------------------------------------------------- */

    mktemp (tmpfn);
    if ((TempFile = fopen (tmpfn, "w+")) == NULL) {
	logMsg ("cannot open tmp file:");
	perror ("->");
	return (0);
    }
    return (1);
}

resetTempFile () {

    rewind (TempFile);
    truncate (tmpfn, 0);
}

rewindTempFile () {

    rewind (TempFile);
}

int	PeekChar, BOLine;

int getC (f) FILE *f; {

/* --------------------------------------------------------------------- */
/* Reads a single character from the file combining multiple spaces into */
/* single spaces and ignoring comment lines				 */
/* --------------------------------------------------------------------- */

    int c, d;

    if (PeekChar != -128) {
	if ((c = PeekChar) != EOF) PeekChar = -128;
	return (c);
    } else {
	while (1) {
	    c = getc (f);
	    if (c == '#' && BOLine) {
		while ((c = getc (f)) != '\n' && c != EOF);
		if (c == EOF) {
		    PeekChar = c;
		    return (c);
		}
	    } else if (isspace (c)) {
		while (1) {	/* Find the last space in sequence */
		    d = getc (f);
		    if (!isspace (d)) break;
		    c = d;
		}
		ungetc (d, f);
		BOLine = (c == '\n');
		return (' ');
	    } else {
		return (c);
	    }
	}
    }
}

int peekC (f) FILE *f; {

/* --------------------------------------------------------------------- */
/* Peeks  at  the next character from the file combining multiple spaces */
/* into single spaces and ignoring comment lines			 */
/* --------------------------------------------------------------------- */

    int c, d;

    if (PeekChar == -128) PeekChar = getC (f);
    return (PeekChar);
}

int getN (f) FILE *f; {

/* --------------------------------------------- */
/* Gets the first nonspace character from file f */
/* --------------------------------------------- */

    int	 c;

    while ((c = getC (f)) == ' ');
    return (c);
}

int peekN (f) FILE *f; {

/* ------------------------------------------------- */
/* Peeks at the first nonspace character from file f */
/* ------------------------------------------------- */

    int	 c;

    while ((c = peekC (f)) == ' ') getC (f);
    return (c);
}

int getK (f, kwd) FILE *f; char *kwd; {

/* --------------------------------------------------------------------- */
/* Gets	 a  "keyword"  from  file  f  (the  size of kwd is assumed to be */
/* MAXPATHLEN)								 */
/* --------------------------------------------------------------------- */

    int c, quoted, nc;

    c = peekN (f);
    if (c == EOF) return (EOF);
    if (c == '"') {
	quoted = 1;
	getC (f); c = peekC (f);
	if (c == EOF) return (EOF);
    } else {
	quoted = 0;
    }

    nc = 0;
    while (1) {
	c = peekC (f);
	if (c == EOF) break;
	if (c == '\\') {
	    getC (f);
	    c = getC (f);
	    if (c == EOF) break;
	} else {
	    if (quoted) {
		if (c == '"') {
		    getC (f);
		    break;
		}
	    } else {
		if (c == ' ' || c == ',' || c == '[' || c == ']') break;
	    }
	    getC (f);
	}
	if (nc < MAXPATHLEN-1) kwd [nc++] = c;
    }
    kwd [nc] = '\0';
/*@@@logMsg (kwd);*/
    return (0);
}

char	*tDate () {

/* ------------------------------------ */
/* Return date and time in textual form */
/* ------------------------------------ */

	long	atime;
	char	*ct, *wt;

	atime = time (0);
	for (wt = ct = ctime (&atime); *ct != '\0'; ct++)
		if (*ct == '\n') {
			*ct = '\0';
			break;
		}
	return (wt);
}

char LastDay [] = "xxx";

logMsg (s) char *s; {

/* --------------------------- */
/* Writes a log line to stderr */
/* --------------------------- */

    char  *d, *pd, car [84];
    int   rem, first;

    d = tDate ();
    pd = d + 11;
    d [19] = '\0';
    d [10] = '\0';
    if (strncmp (d, LastDay, 3) != 0) {
	if (LastDay [0] == 'x') {
	    /* First time */
	    fprintf (stderr, "%s: @@@@@ Started on %s\n", pd, d);
	} else {
	    fprintf (stderr, "00:00:00: @@@@@ BIM! BOM!  --  %s\n", d);
	}
        strncpy (LastDay, d, 3);
    }
    for (first = 0, rem = strlen (s); rem > 0; ) {
	if (rem > 69 - first) {
	    strncpy (car, s, 69-first);
	    rem -= 69-first;
	    s += 69-first;
	    car [69-first] = '\0';
	} else {
	    strcpy (car, s);
	    rem = 0;
	}
	if (first) {
            fprintf (stderr, "%s: +%s\n", pd, car);
        } else {
            fprintf (stderr, "%s: %s\n", pd, car);
	    first = 1;
	}
    }
}

formatError (fname, fptr, lc) char *fname; FILE *fptr; char lc; {

/* ------------------------------------------------- */
/* Writes to the logfile a file format error message */
/* ------------------------------------------------- */

    int	 i;
    char car [MAXPATHLEN];

    sprintf (car, "'%s' format error:", fname);
    logMsg (car);
    if (lc == EOF) {
	logMsg ("  -> unexpected EOF");
    } else {
	strcpy (car, "  -> \"");
	car [6] = lc;
        for (i = 0; i < 9; i++) {
    	    /* Include some context information */
	    if ((lc = getC (fptr)) == EOF) break;
	    car [7+i] = lc;
        }
	car [7+i] = '\0';
	logMsg (car);
    }
}

logFile (fd) FILE *fd; {

/* ----------------------------------------- */
/* Copies the contents of a file to log file */
/* ----------------------------------------- */

    char car [MAXPATHLEN];
    int  nc, i;

    logMsg (" +*****");
    i = 0;
    while ((nc = getc (fd)) != EOF) {
	if (nc == '\n') {
	    car [i] = '\0';
	    logMsg (car);
	    i = 0;
	    continue;
	}
        if (i >= MAXPATHLEN-1) {
	    car [i] = '\0';
	    logMsg (car);
	    i = 0;
	}
	car [i++] = nc;
    }
    if (i > 0) {
	car [i] = '\0';
	logMsg (car);
    }
    logMsg (" +*****");
}
		
int readExperimentsFile () {

/* --------------------------------------------------------------------- */
/* Reads  in  the  contents  of the experiments file. Returns nonzero if */
/* there is any work to be done.					 */
/* --------------------------------------------------------------------- */

    char *s;
    char dir [MAXPATHLEN], tmp [MAXPATHLEN];
    int	 c, and;
    struct ExpClass *ex, *cex;
    struct ProgDesc *pg, *cpg;
    struct RunDesc  *rn, *crn;

    /* The format of the Experiments file is as follows:

	   * dirpath [mtype progname] ... [mtype progname] options
	      @ datafile outfile optional flags added by serdel
	      @ datafile outfile  . . . .
		 . . . . .
	   * dirpath ...
		 . . . . .
    */

    BOLine = PeekChar = -128;
    if ((ExpFile = fopen (EXPFILE, "r+")) == NULL)
	/* The file cannot be opened */
	return (0);

    cex = NULL;
    while (1) {
	if ((c = getN (ExpFile)) == EOF) return (Experiments != NULL);
	if (c != '*') {
FError:
	    formatError (EXPFILE, ExpFile, c);
	    return (0);
	}
	/* Decode directory name */
	if (getK (ExpFile, dir) == EOF) goto FError;
	/* Check if the directory is absolute */
	if (dir [0] != '/') {
	    /* Make it absolute by appending the home directory path */
	    if ((s = getenv ("HOME")) == NULL) {
		logMsg ("Can't get home directory path");
		return (0);
	    }
	    strcpy (tmp, dir);
	    strcpy (dir, s);
	    strcat (dir, "/");
	    strcat (dir, tmp);
	    ExModified++;
	}
	/* Allocate new ExpClass entry */
	ex = (struct ExpClass*) malloc (sizeof (struct ExpClass));
	if (ex == NULL) {
	    logMsg ("No space");
	    exit (1);
	}
	ex->next = NULL;
	ex->Options = NULL;
	ex->ProgList = NULL;
	ex->RunList = NULL;
	ex->DirPath = (char*) malloc (strlen (dir) + 1);
	if (ex->DirPath == NULL) {
	    logMsg ("No space");
	    exit (1);
	}
	strcpy (ex->DirPath, dir);
	if (cex == NULL) Experiments = ex; else cex->next = ex;
	cex = ex;
	/* Now for the machine list */
	if ((c = peekN (ExpFile)) != '[') goto FError;
	cpg = NULL;
	while (c == '[') {
	    getN (ExpFile);
	    /* Machine type */
	    if (getK (ExpFile, tmp) == EOF) goto FError;
	    if ((c = peekN (ExpFile)) == ',') {
		getN (ExpFile);
	    } else {
		if (c == '[' || c == ']') goto FError;
	    }
	    /* Program name */
	    if (getK (ExpFile, dir) == EOF) goto FError;
	    if ((c = getN (ExpFile)) != ']') goto FError;
	    /* Allocate a new program description entry */
	    pg = (struct ProgDesc*) malloc (sizeof (struct ProgDesc));
	    if (pg == NULL) {
		logMsg ("No space");
		exit (1);
	    }
	    pg->next = NULL;
	    if (cpg == NULL) ex->ProgList = pg; else cpg->next = pg;
	    cpg = pg;
	    pg->MType = (char*) malloc (strlen (tmp) + 1);
	    pg->PName = (char*) malloc (strlen (dir) + 1);
	    if (pg->MType == NULL || pg->PName == NULL) {
		logMsg ("No space");
		exit (1);
	    }
	    strcpy (pg->MType, tmp);
	    strcpy (pg->PName, dir);
	    c = peekN (ExpFile);
	}
	/* Check for options */
	if (c == EOF) goto FError;
	if (c != '@') {
	    /* Process options */
	    getK (ExpFile, tmp);
	    ex->Options = (char*) malloc (strlen (tmp) + 1);
	    if (ex->Options == NULL) {
		logMsg ("No space");
		exit (1);
	    }
	    strcpy (ex->Options, tmp);
	    c = peekN (ExpFile);
	    if (c == EOF || c != '@') goto FError;
	}
	/* Now for the run list */
	crn = NULL;
	while (c == '@') {
	    getN (ExpFile);
	    /* Data file */
	    if (getK (ExpFile, tmp) == EOF) goto FError;
	    if ((c = peekN (ExpFile)) == ',') {
		getN (ExpFile);
	    } else {
		if (c == '[' || c == ']') goto FError;
	    }
	    /* Output file */
	    if (getK (ExpFile, dir) == EOF) goto FError;
	    /* Allocate new entry */
	    rn = (struct RunDesc*) malloc (sizeof (struct RunDesc));
	    if (rn == NULL) {
		logMsg ("No space");
		exit (1);
	    }
	    rn->next = NULL;
	    rn->Host = NULL;
	    rn->Status = UNTOUCHED;
	    rn->Pid = -1;
	    rn->NTries = 0;
	    if (crn == NULL) ex->RunList = rn; else crn->next = rn;
	    crn = rn;
	    rn->DataFile = (char*) malloc (strlen (tmp) + 1);
	    rn->OutputFile = (char*) malloc (strlen (dir) + 1);
	    if (rn->DataFile == NULL || rn->OutputFile == NULL) {
		logMsg ("No space");
		exit (1);
	    }
	    strcpy (rn->DataFile, tmp);
	    strcpy (rn->OutputFile, dir);
	    c = peekN (ExpFile);
	    if (c != EOF && c != '@' && c != '*') {
		/* More items follow */
		getK (ExpFile, tmp);	/* Hostname */
		rn->Host = (char*) malloc (strlen (tmp) + 1);
		if (rn->Host == NULL) {
		    logMsg ("No space");
		    exit (1);
		}
		strcpy (rn->Host, tmp);
		c = getN (ExpFile);
		if (c == '-') {
		    rn->Status = STARTED;
		} else if (c == '+') {
		    rn->Status = COMPLETED;
		} else if (c == '!') {
		    rn->Status = CHECKPOINTED;
		} else if (c == '?') {
		    rn->Status = UNCERTAIN;
		} else
		    goto FError;
		/* Now check for the process Id */
		if (rn->Status == STARTED || rn->Status == UNCERTAIN) {
		    /* Now check for the process Id */
		    c = peekN (ExpFile);
		    if (c == EOF || c == '@' || c == '*') goto FError;
		    getK (ExpFile, tmp);
		    rn->Pid = atoi (tmp);
		}
		if (rn->Status == STARTED || rn->Status == CHECKPOINTED ||
	    	  rn->Status == UNCERTAIN) {
		    /* Now expect last checkpoint time */
		    c = peekN (ExpFile);
		    if (c == EOF || c == '@' || c == '*') goto FError;
		    getK (ExpFile, tmp);
		    rn->LastCheckp = atoi (tmp);
		}
		if (rn->Status == UNCERTAIN) {
		    /* Now expect the number of tries */
		    c = peekN (ExpFile);
		    if (c == EOF || c == '@' || c == '*') goto FError;
		    getK (ExpFile, tmp);
		    rn->NTries = atoi (tmp);
		}
		c = peekN (ExpFile);
	    }
	}
    }
}

int readHostFile () {

/* --------------------------------------------------------------------- */
/* Reads  in  the  contents  of	 the  host  file. Returns nonzero if the */
/* operation has been successful.					 */
/* --------------------------------------------------------------------- */

    char h [MAXPATHLEN], m [MAXPATHLEN], *cp;
    int	 c, i, j;
    struct HostDesc *hd, *chd;

    /* The format of the Hosts file is as follows:

	name, type, pscmd, tload, eload, maxexpcnt, int, ckp, atime
	  . . . . . .
	name, type, pscmd, tload, eload, maxexpcnt, int, ckp, atime

    */

    BOLine = PeekChar = -128;
    if ((HostFile = fopen (HOSTFILE, "r")) == NULL)
	/* The file cannot be opened */
	return (0);

    chd = NULL;
    while (1) {
	if ((c = peekN (HostFile)) == EOF) return (Hosts != NULL);
	getK (HostFile, h);
	if ((c = peekN (HostFile)) == ',') getN (HostFile);
	if ((c = peekN (HostFile)) == EOF) {
FError:
	    formatError (HOSTFILE, HostFile, c);
	    return (0);
	}
	/* Machine type */
	getK (HostFile, m);
	hd = (struct HostDesc*) malloc (sizeof (struct HostDesc));
	if (hd == NULL) {
	    logMsg ("No space");
	    exit (1);
	}
	hd->next = NULL;
	if (chd == NULL) Hosts = hd; else chd->next = hd;
	chd = hd;
	hd->PsCommand = NULL;
	hd->Guard = NULL;
	hd->TLoad = 1.0;
	hd->ELoad = 1000.0;
	hd->Pri = 10;
	hd->ECount = 1;
	hd->MIA = 0;
	hd->PERC = 0;	/* Periodic checkpointing */
	/* Zero out availability intervals */
	for (i = 0; i < 7; i++)
	    for (j = 0; j < MXATINTVL; j++)
		hd->STime [i][j] = hd->ETime [i][j] = -1;
	hd->IA = -1;	/* Unknown yet */
	hd->Host = (char*) malloc (strlen (h) + 1);
	hd->MType = (char*) malloc (strlen (m) + 1);
	if (hd->Host == NULL || hd->MType == NULL) {
	    logMsg ("No space");
	    exit (1);
	}
	strcpy (hd->Host, h);
	strcpy (hd->MType, m);
	if ((c = peekN (HostFile)) == ',') getN (HostFile);
	if (getK (HostFile, h) == EOF) goto FError;
	/* ps -x command */
	hd->PsCommand = (char*) malloc (strlen (h) + 1);
	if (hd->PsCommand == NULL) {
	    logMsg ("No space");
	    exit (1);
	}
	strcpy (hd->PsCommand, h);

	if ((c = peekN (HostFile)) == ',') getN (HostFile);
	/* Startup load threshold */
	if (getK (HostFile, h) == EOF) goto FError;
	hd->TLoad = atof (h);

	if ((c = peekN (HostFile)) == ',') getN (HostFile);
	/* Swap-out load threshold */
	if (getK (HostFile, h) == EOF) goto FError;
	hd->ELoad = atof (h);

	if ((c = peekN (HostFile)) == ',') getN (HostFile);
	/* Priority */
	if (getK (HostFile, h) == EOF) goto FError;
	hd->Pri = atoi (h);

	if ((c = peekN (HostFile)) == ',') getN (HostFile);
	/* Maximum experiment count */
	if (getK (HostFile, h) == EOF) goto FError;
	hd->ECount = atoi (h);

	if ((c = peekN (HostFile)) == ',') getN (HostFile);
	/* Determines whether interactive activities are to be detected */
	if (getK (HostFile, h) == EOF) goto FError;
	hd->MIA = (*h == 'y' || *h == 'Y');

	if ((c = peekN (HostFile)) == ',') getN (HostFile);
	if (getK (HostFile, h) == EOF) goto FError;
	/* Guard program name */
	if (*h != '\0' && !isspace (*h)) {
	    hd->Guard = (char*) malloc (strlen (h) + 1);
	    if (hd->Guard == NULL) {
	        logMsg ("No space");
	        exit (1);
	    }
	    strcpy (hd->Guard, h);
        }

	if ((c = peekN (HostFile)) == ',') getN (HostFile);
	/* Determines whether periodic checkpointing is to be performed  */
	if (getK (HostFile, h) == EOF) goto FError;
	hd->PERC = (*h == 'y' || *h == 'Y');

	if ((c = peekN (HostFile)) == ',') getN (HostFile);
	/* Availability intervals */
	if (getK (HostFile, h) == EOF) goto FError;
	/* Example:
	     MWF0800-1200/0200-2400/0200-0400TR0000-0800S0000-0000
	*/
	if (aiParse (hd, h) < 0) goto FError;
    }
}

int	wkDay (c) int c; {

/* ------------------------------------------------- */
/* Converts weekday letter into a number from 0 to 6 */
/* ------------------------------------------------- */

    if (islower (c)) c = toupper (c);
    switch (c) {

	case 'S' : return (0);
	case 'M' : return (1);
	case 'T' : return (2);
	case 'W' : return (3);
	case 'R' : return (4);
	case 'F' : return (5);
	case 'A' : return (6);
        case 'N' : return (0);
	default  : return (-1);
    }
}

int	aiParse (host, buf)  struct HostDesc *host; char *buf; {

/* ------------------------------ */
/* Unpacks availability intervals */
/* ------------------------------ */

    char  *bp, diag [256];
    int	  wk [7], i, intvl, from, to;

    bp = buf;

    while (1) {
	if (*bp == '\0') return (0);
	for (i = 0; i < 7; i++) wk [i] = 0;
	i = 128;
	while (isalpha (*bp)) {
	    if ((i = wkDay (*bp)) < 0) {
		sprintf (diag, "'%c' -- illegal day code", *bp);
PErr:
		logMsg (diag);
		return (-1);
	    }
	    if (wk [i]) {
		sprintf (diag, "'%c' -- day occurs twice", *bp);
		goto PErr;
	    }
	    wk [i] = 1;
	    bp++;
	}
	if (i > 6) {
	    for (i = 0; i < 7; i++) wk [i] = 1;
	}
	/* Check if not already defined */
	for (i = 0; i < 7; i++) {
	    if (wk [i] && host->STime [i][0] != -1) {
		logMsg ("Multiple interval definitions for the same day");
	        return (-1);
	    }
	}
	intvl = 0;
	while (1) {
	    if (!isdigit (*bp)) {
	        if (*bp == '\0') *bp = ' ';
	        sprintf (diag,
		    "'%c' -- day codes or '/' not followed by a digit", *bp);
	        goto PErr;
	    }
	    from = atoi (bp);
	    while (isdigit (*bp)) bp++;
	    if (*bp != '-') {
	        if (*bp == '\0') *bp = ' ';
	        sprintf (diag, "'-' expected instead of '%c'", *bp);
	        goto PErr;
	    }
	    bp++;
	    if (!isdigit (*bp)) {
	        if (*bp == '\0') *bp = ' ';
	        sprintf (diag, "Digit expected after '-' instead of '%c'", *bp);
	        goto PErr;
    	    }
	    to = atoi (bp);
	    while (isdigit (*bp)) bp++;
	    if (intvl >= MXATINTVL) {
		logMsg ("Too many availability intervals");
		return (-1);
	    }
	    for (i = 0; i < 7; i++) {
		if (wk [i]) {
		    host->STime [i][intvl] = from;
		    host->ETime [i][intvl] = to;
		}
	    }
	    intvl++;
	    if (*bp != '/') break;
	    bp++;
	}
    }
}

deallocateStorage () {

/* --------------------------------------- */
/* Deallocate dynamically allocated memory */
/* --------------------------------------- */

    struct ExpClass *ex, *cex;
    struct ProgDesc *pg, *cpg;
    struct RunDesc  *rd, *crd;
    struct HostDesc *ho, *cho;

    for (ex = Experiments; ex != NULL; ) {
	cex = ex; ex = ex->next;
	if (cex->DirPath != NULL) free (cex->DirPath);
	if (cex->Options != NULL) free (cex->Options);

	for (pg = cex->ProgList; pg != NULL; ) {
	    cpg = pg; pg = pg->next;
	    if (cpg->MType != NULL) free (cpg->MType);
	    if (cpg->PName != NULL) free (cpg->PName);
	    free (cpg);
	}
	
	for (rd = cex->RunList; rd != NULL; ) {
	    crd = rd; rd = rd->next;
	    if (crd->DataFile != NULL) free (crd->DataFile);
	    if (crd->OutputFile != NULL) free (crd->OutputFile);
	    if (crd->Host != NULL) free (crd->Host);
	    free (crd);
	}

	free (ex);
    }

    Experiments = NULL;

    for (ho = Hosts; ho != NULL; ) {
	cho = ho; ho = ho->next;
	if (cho->Host != NULL) free (cho->Host);
	if (cho->MType != NULL) free (cho->MType);
	if (cho->PsCommand != NULL) free (cho->PsCommand);
	if (cho->Guard != NULL) free (cho->Guard);
	free (cho);
    }

    Hosts = NULL;
}

closeFiles () {

    fclose (ExpFile);
    fclose (HostFile);
    fclose (TempFile);
    unlink (tmpfn);
}

int	*retjmp;	/* Context to assume after timeout */

timeout () {

/* ---------------------- */
/* Timeout signal service */
/* ---------------------- */

    longjmp (retjmp, 1);
}

setTimer (delay, context) long delay; jmp_buf context; {

/* ------------------ */
/* Set interval timer */
/* ------------------ */

    struct  itimerval	    td;
    td.it_value.tv_sec = delay;
    td.it_value.tv_usec = 0;
    td.it_interval.tv_sec = 0;
    td.it_interval.tv_usec = 0;
    retjmp = (void*)(&(context[0]));
    signal (SIGALRM, timeout);
    signal (SIGPIPE, timeout);
    setitimer (ITIMER_REAL, &td, NULL);
}

clearTimer () {

/* -------------------- */
/* Clear interval timer */
/* -------------------- */

    struct  itimerval	    td;
    td.it_value.tv_sec = 0;
    td.it_value.tv_usec = 0;
    td.it_interval.tv_sec = 0;
    td.it_interval.tv_usec = 0;
    setitimer (ITIMER_REAL, &td, NULL);
    signal (SIGALRM, SIG_DFL);
    signal (SIGPIPE, SIG_DFL);
}

removeDefuncts () {

/* ----------------------------- */
/* Gets rid of defunct processes */
/* ----------------------------- */

#ifdef	SYSV
    int wjunk;
    wait2 (&wjunk, WNOHANG);
#else
    union	wait	wjunk;
    wait3 (&wjunk, WNOHANG, NULL);
#endif
}

int	remoteCommand (host, cmd) char *host, *cmd; {

/* --------------------------------------------------------------------- */
/* Executes a remote command on the specified host. Writes the output of */
/* the command to the temporary file. Times out after one minute.	 */
/* --------------------------------------------------------------------- */

    int	    pip [2], cid;  char cc, cbuf [CBUFSIZE];
    jmp_buf tcntxt;

    if (Verbose > 1) {
	sprintf (cbuf, "Remote command '%s' on host '%s'", cmd, host);
	logMsg (cbuf);
    }
    /* Create a pipe to communicate with the child */
    if (pipe (pip)) {
	/* Failure */
	sprintf (cbuf, "remoteCommand on '%s': cannot open pipe", host);
	logMsg (cbuf);
	perror ("->");
	return (-1);
    }
	
    if ((cid = fork ()) == -1) {
	sprintf (cbuf, "remoteCommand on '%s': cannot fork", host);
	logMsg (cbuf);
	close (pip [0]);
	close (pip [1]);
	return (-1);
    }

    if (cid == 0) {		/* The son */
	setpgrp (0, getpid ());
	close (pip [0]);	/* The input end */
	close (1);		/* The standard output */
	dup (pip [1]);		/* Will write to the pipe */
	/* Create the system command */
	sprintf (cbuf, "rsh -n %s %s", host, cmd);
	system (cbuf);
	close (1);
	exit (0);
    }

    /* In the parent */
    close (pip [1]);		/* Close the output end */
    resetTempFile ();
    if (setjmp (tcntxt)) {
	/* After timeout */
	clearTimer ();
	if (Verbose) {
	    sprintf (cbuf, "remoteCommand on '%s': timed out", host);
	    logMsg (cbuf);
	}
	killpg (cid, SIGKILL);
	removeDefuncts ();
	close (pip [0]);
	return (-1);
    }

    setTimer (RTIMEOUT, tcntxt);

    /* Read the output produced by the child */
    while (read (pip [0], &cc, 1) > 0) putc (cc, TempFile);
    close (pip [0]);

    clearTimer ();

    killpg (cid, SIGKILL);
    removeDefuncts ();
    fflush (TempFile);
    if (Verbose > 1) {
	rewindTempFile ();
        logFile (TempFile);
    }
    rewindTempFile ();
    return (0);
}

int startGuard (hd, pid)

/* ---------------------------------------------- */
/* Starts the guard process on the specified host */
/* ---------------------------------------------- */

  struct HostDesc *hd;
  int		  pid;

{
    int	    cid, pip [2];
    char    cbuf [CBUFSIZE], cc;
    jmp_buf tcntxt;

    if (hd->MIA == 0 || hd->Guard == NULL) return;

    sprintf (cbuf, "rsh -n %s \"%s %1d  >& /dev/null &\"", hd->Host, hd->Guard,
	pid);

    if (Verbose) {
	logMsg ("Trying to start the guard process");
	logMsg (cbuf);
    }

    /* Create a pipe to communicate with the child */
    if (pipe (pip)) {
	/* Failure */
	sprintf (cbuf, "startGuard on '%s': cannot open pipe", hd->Host);
	logMsg (cbuf);
	perror ("->");
	return (-1);
    }
	
    if ((cid = fork ()) == -1) {
	sprintf (cbuf, "startGuard on '%s': cannot fork", hd->Host);
	logMsg (cbuf);
	close (pip [0]);
	close (pip [1]);
	return (-1);
    }

    if (cid == 0) {		/* The son */
	setpgrp (0, getpid ());
	close (pip [0]);	/* The input end */
	close (1);		/* The standard output */
	dup (pip [1]);		/* Will write to the pipe */
	system (cbuf);
	close (1);
	exit (0);
    }

    /* In the parent */
    close (pip [1]);		/* Close the output end */
    if (setjmp (tcntxt)) {
	/* After timeout */
	clearTimer ();
	if (Verbose) {
	    sprintf (cbuf, "startGuard on '%s': timed out", hd->Host);
	    logMsg (cbuf);
	}
	killpg (cid, SIGKILL);
	removeDefuncts ();
	close (pip [0]);
	return (-1);
    }

    setTimer (RTIMEOUT, tcntxt);

    /* Read the output produced by the child */

    resetTempFile ();
    while (read (pip [0], &cc, 1) > 0) {
	putc (cc, TempFile);
	if (cc == '[') break;
    }
	
    /* Find the process number in the response. It should look like: */
    /*		[...] nnnnnnn					     */

    if (cc != '[') {
RexErr:
	/* Something wrong */
	clearTimer ();
	killpg (cid, SIGKILL);
	removeDefuncts ();
	close (pip [0]);
	logMsg ("startGuard: cannot start the process, response:");
	rewindTempFile ();
        logFile (TempFile);
	return (-1);
    }
    while (read (pip [0], &cc, 1) > 0) {
	putc (cc, TempFile);
	if (cc == ']') break;
    }
    if (cc != ']') goto RexErr;
    while (read (pip [0], &cc, 1) > 0) {
	putc (cc, TempFile);
	if (!isspace (cc)) break;
    }
    /* This must be a digit */
    if (!isdigit (cc)) goto RexErr;
    pid = cc - '0';
    while (read (pip [0], &cc, 1) > 0 && isdigit (cc)) {
	putc (cc, TempFile);
	pid = pid * 10 + (cc - '0');
    }
    clearTimer ();
    killpg (cid, SIGKILL);
    removeDefuncts ();
    close (pip [0]);
    if (Verbose) {
	sprintf (cbuf, "Guard process id: %1d", pid);
	logMsg (cbuf);
    }
}

int startExperiment (hd, ed, rd, pd)

/* --------------------------------------------------------------------- */
/* Starts an experiment on the indicated host. Returns the process id or */
/* -1 in case of failure.						 */
/* --------------------------------------------------------------------- */

  struct HostDesc *hd;
  struct ExpClass *ed;
  struct RunDesc  *rd;
  struct ProgDesc *pd;

{
    int	    cid, pid, i, pip [2];
    char    cc, abuf [CBUFSIZE], cbuf [CBUFSIZE], nice [20];

    jmp_buf tcntxt;

    /* Remove any existing output files */

    sprintf (cbuf, "%s/%s", ed->DirPath, rd->OutputFile);
    unlink (cbuf);
    sprintf (cbuf, "%s/errors_%s", ed->DirPath, rd->OutputFile);
    unlink (cbuf);
    sprintf (cbuf, "%s/checkp_%s", ed->DirPath, rd->OutputFile);
    unlink (cbuf);

    /* Prepare the program call line */
    if (hd->Pri == 0) {
	nice [0] = '\0';
    } else {
	if (hd->Pri < 0) hd->Pri == 0; else if (hd->Pri > 20) hd->Pri = 20;
	sprintf (nice, "nice +%1d ", hd->Pri);
    }
#ifdef	NOTIFY
    sprintf (cbuf, "rsh -n %s \"%s%s/%s %s/%s %s/%s -m %s >& %s/errors_%s &\"",
#else
    sprintf (cbuf, "rsh -n %s \"%s%s/%s %s/%s %s/%s %s >& %s/errors_%s &\"",
#endif
	hd->Host, nice, ed->DirPath, pd->PName, ed->DirPath, rd->DataFile,
	    ed->DirPath, rd->OutputFile, ed->Options == NULL ? "" :
		ed->Options, ed->DirPath, rd->OutputFile);

    if (Verbose) {
	logMsg ("Trying to start experiment");
	logMsg (cbuf);
    }

    /* Create a pipe to communicate with the child */
    if (pipe (pip)) {
	/* Failure */
	sprintf (cbuf, "startExperiment on '%s': cannot open pipe", hd->Host);
	logMsg (cbuf);
	perror ("->");
	return (-1);
    }
	
    if ((cid = fork ()) == -1) {
	sprintf (cbuf, "startExperiment on '%s': cannot fork", hd->Host);
	logMsg (cbuf);
	close (pip [0]);
	close (pip [1]);
	return (-1);
    }

    if (cid == 0) {		/* The son */
	setpgrp (0, getpid ());
	close (pip [0]);	/* The input end */
	close (1);		/* The standard output */
	dup (pip [1]);		/* Will write to the pipe */
	system (cbuf);
	close (1);
	exit (0);
    }

    /* In the parent */
    close (pip [1]);		/* Close the output end */
    if (setjmp (tcntxt)) {
	/* After timeout */
	clearTimer ();
	if (Verbose) {
	    sprintf (cbuf, "startExperiment on '%s': timed out", hd->Host);
	    logMsg (cbuf);
	}
	killpg (cid, SIGKILL);
	removeDefuncts ();
	close (pip [0]);
	return (-1);
    }

    setTimer (RTIMEOUT, tcntxt);

    /* Read the output produced by the child */

    resetTempFile ();
    while (read (pip [0], &cc, 1) > 0) {
	putc (cc, TempFile);
	if (cc == '[') break;
    }
	
    /* Find the process number in the response. It should look like: */
    /*		[...] nnnnnnn					     */

    if (cc != '[') {
RexErr:
	/* Something wrong */
	clearTimer ();
	killpg (cid, SIGKILL);
	removeDefuncts ();
	close (pip [0]);
	logMsg ("startExperiment: cannot start remote process, params:");
        sprintf (abuf, "+Host: %s, Cmd: %s, Response:", hd->Host, cbuf);
	logMsg (abuf);
	rewindTempFile ();
        logFile (TempFile);
	return (-1);
    }
    while (read (pip [0], &cc, 1) > 0) {
	putc (cc, TempFile);
	if (cc == ']') break;
    }
    if (cc != ']') goto RexErr;
    while (read (pip [0], &cc, 1) > 0) {
	putc (cc, TempFile);
	if (!isspace (cc)) break;
    }
    /* This must be a digit */
    if (!isdigit (cc)) goto RexErr;
    pid = cc - '0';
    while (read (pip [0], &cc, 1) > 0 && isdigit (cc)) {
	putc (cc, TempFile);
	pid = pid * 10 + (cc - '0');
    }
    clearTimer ();
    killpg (cid, SIGKILL);
    removeDefuncts ();
    close (pip [0]);
    ExModified++;
    rd->Status = STARTED;
    rd->LastCheckp = time (0);
    rd->Pid = pid;
    if (Verbose) {
	sprintf (cbuf, "Remote process id: %1d", pid);
	logMsg (cbuf);
    }
    if (rd->Host != NULL) free (rd->Host);
    rd->Host = (char*) malloc (strlen (hd->Host) + 1);
    strcpy (rd->Host, hd->Host);
    startGuard (hd, pid);
    return (pid);
}

int restartCheckpointed (hd, ed, rd, pd)

/* --------------------------------------------------------------------- */
/* Restarts a checkpointed experiment on the indicated host. Returns the */
/* process id or -1 in case of failure.					 */
/* --------------------------------------------------------------------- */

  struct HostDesc *hd;
  struct ExpClass *ed;
  struct RunDesc  *rd;
  struct ProgDesc *pd;

{
    int	    cid, pid, i; char cc, abuf [CBUFSIZE], cbuf [CBUFSIZE], nice [20];
    int	    pip [2];
    jmp_buf tcntxt;

    /* Save the checkpoint file */
    sprintf (cbuf, "cp %s/%s %s/checkp_%s", ed->DirPath, rd->OutputFile,
	ed->DirPath, rd->OutputFile);
    system (cbuf);

    /* Prepare the program call line */
    if (hd->Pri == 0) {
	nice [0] = '\0';
    } else {
	if (hd->Pri < 0) hd->Pri == 0; else if (hd->Pri > 20) hd->Pri = 20;
	sprintf (nice, "nice +%1d ", hd->Pri);
    }
    sprintf (cbuf, "rsh -n %s \"%s%s/%s -R %s/%s >& %s/errors_%s &\"",
	hd->Host, nice, ed->DirPath, pd->PName, ed->DirPath, rd->OutputFile,
	    ed->DirPath, rd->OutputFile);

    if (Verbose) {
	logMsg ("Trying to restart checkpointed experiment");
	logMsg (cbuf);
    }

    /* Create a pipe to communicate with the child */
    if (pipe (pip)) {
	/* Failure */
	sprintf (cbuf, "restartCheckpointed on '%s': cannot open pipe",
		hd->Host);
	logMsg (cbuf);
	perror ("->");
	return (-1);
    }
	
    if ((cid = fork ()) == -1) {
	sprintf (cbuf, "restartCheckpointed on '%s': cannot fork", hd->Host);
	logMsg (cbuf);
	close (pip [0]);
	close (pip [1]);
	return (-1);
    }

    if (cid == 0) {		/* The son */
	setpgrp (0, getpid ());
	close (pip [0]);	/* The input end */
	close (1);		/* The standard output */
	dup (pip [1]);		/* Will write to the pipe */
	system (cbuf);
	close (1);
	exit (0);
    }

    /* In the parent */
    close (pip [1]);		/* Close the output end */
    if (setjmp (tcntxt)) {
	/* After timeout */
	clearTimer ();
	if (Verbose) {
	   sprintf (cbuf, "restartCheckpointed on '%s': cannot fork", hd->Host);
	   logMsg (cbuf);
	}
	killpg (cid, SIGKILL);
	removeDefuncts ();
	close (pip [0]);
	return (-1);
    }

    setTimer (RTIMEOUT, tcntxt);

    /* Read the output produced by the child */

    resetTempFile ();
    while (read (pip [0], &cc, 1) > 0) {
	putc (cc, TempFile);
	if (cc == '[') break;
    }
	
    /* Find the process number in the response. It should look like: */
    /*		[...] nnnnnnn					     */

    if (cc != '[') {
RexErr:
	/* Something wrong */
	clearTimer ();
	killpg (cid, SIGKILL);
	removeDefuncts ();
	close (pip [0]);
	logMsg ("restartExperiment: cannot start remote process, params:");
        sprintf (abuf, "+Host: %s, Cmd: %s, Response:", hd->Host, cbuf);
	logMsg (abuf);
	rewindTempFile ();
        logFile (TempFile);
	return (-1);
    }
    while (read (pip [0], &cc, 1) > 0) {
	putc (cc, TempFile);
	if (cc == ']') break;
    }
    if (cc != ']') goto RexErr;
    while (read (pip [0], &cc, 1) > 0) {
	putc (cc, TempFile);
	if (!isspace (cc)) break;
    }
    /* This must be a digit */
    if (!isdigit (cc)) goto RexErr;
    pid = cc - '0';
    while (read (pip [0], &cc, 1) > 0 && isdigit (cc)) {
	putc (cc, TempFile);
	pid = pid * 10 + (cc - '0');
    }
    clearTimer ();
    killpg (cid, SIGKILL);
    removeDefuncts ();
    close (pip [0]);
    ExModified++;
    rd->Status = STARTED;
    rd->LastCheckp = time (0);
    rd->Pid = pid;
    if (Verbose) {
	sprintf (cbuf, "Remote process id: %1d", pid);
	logMsg (cbuf);
    }
    if (rd->Host != NULL) free (rd->Host);
    rd->Host = (char*) malloc (strlen (hd->Host) + 1);
    strcpy (rd->Host, hd->Host);
    startGuard (hd, pid);
    return (pid);
}

int	stillAlive (hd, ed, rd)

/* ----------------------------------------------------- */
/* Check if the run is still alive on the specified host */
/* ----------------------------------------------------- */

  struct HostDesc *hd;
  struct ExpClass *ed;
  struct RunDesc  *rd;

{
    FILE  *df; char fnbuf [MAXPATHLEN], *cp;
    int	  nc, pid, sl;

    if (Verbose > 1) {
	sprintf (fnbuf, "Checking alive: %s, %s, %1d (%s, %s)",
	    hd->Host, ed->DirPath, rd->Pid, rd->DataFile, rd->OutputFile);
	logMsg (fnbuf);
    }
    if (remoteCommand (hd->Host, hd->PsCommand)) {
	/* The host appears dead */
AssumeHostDead:
	if (Verbose) {
		sprintf (fnbuf, "Host '%s' appears dead", hd->Host);
		logMsg (fnbuf);
	}
        ExModified++;
	if (rd->Status != UNCERTAIN) {
	    rd->NTries = 0;
	    if (Verbose) {
	        logMsg ("Still trying (0)");
	    }
	    return (rd->Status = UNCERTAIN);
	}
	if (++(rd->NTries) < NTRIES) {
	    /* Give it the benefit of the doubt */
	    if (Verbose) {
	        sprintf (fnbuf, "Still trying (%1d)", rd->NTries);
	        logMsg (fnbuf);
	    }
	    return (UNCERTAIN);
        }
	/* Assume that the experiment is dead */
AssumeDead:
	if (Verbose) {
	    sprintf (fnbuf, "Experiment '%s' on '%s' presumed dead",
		rd->OutputFile, hd->Host);
	    logMsg (fnbuf);
	}
	/* Try to read the output file */
	strcpy (fnbuf, ed->DirPath);
	strcat (fnbuf, "/");
	strcat (fnbuf, rd->OutputFile);
	if ((df = fopen (fnbuf, "r")) == NULL) {
	    if (Verbose) logMsg ("Open for output file fails");
AssumeUnfinished:
	    if (Verbose) logMsg ("Experiment assumed aborted");
	    ExModified++;
	    /* Check if the checkpoint file exists */
            sprintf (fnbuf, "%s/checkp_%s", ed->DirPath, rd->OutputFile);
	    if ((sl = open (fnbuf, O_RDONLY, 0)) >= 0) {
		close (sl);
                sprintf (fnbuf, "cp %s/checkp_%s %s/%s", ed->DirPath,
		    rd->OutputFile, ed->DirPath, rd->OutputFile);
		system (fnbuf);
		sprintf (fnbuf, "%s/checkp_%s", ed->DirPath, rd->OutputFile);
		/* Remove it to avoid looping on a corrupted checkpoint file */
		unlink (fnbuf);
	        if (Verbose) logMsg ("Using existing checkpoint file");
		rd->LastCheckp = time (0);
		return (rd->Status = CHECKPOINTED);
	    }
	    rd->Status = UNTOUCHED;
	    if (rd->Host != NULL) {
		free (rd->Host);
		rd->Host = NULL;
	    }
	    return (UNTOUCHED);
	}
	if (fseek (df, (long)(-(strlen (ENDSTRING) + 4)), 2) < 0) {
	    if (Verbose) logMsg ("fseek on output file fails");
	    fclose (df);
	    goto AssumeUnfinished;
	}
	for (cp = fnbuf; (nc = getc (df)) != EOF; *cp++ = nc);
	*cp = '\0';
	sl = strlen (ENDSTRING);
	for (nc = 0, cp = fnbuf; nc < 8; nc++, cp++) {
	    if (strncmp (cp, ENDSTRING, sl) == 0) {
		/* Finished */
		fclose (df);
		if (Verbose) logMsg ("Experiment completed");
		/* Copy and remove the error file */
		strcpy (fnbuf, ed->DirPath);
		strcat (fnbuf, "/errors_");
		strcat (fnbuf, rd->OutputFile);
		if ((df = fopen (fnbuf, "r")) != NULL) {
		    /* Don't bother if cannot open */
		    if ((nc = getc (df)) != EOF) {
			logMsg ("Completion message:");
			logMsg (fnbuf);
			logFile (df);
		    }
		    fclose (df);
		}
		unlink (fnbuf);
		/* Remove the checkpoint save file */
		strcpy (fnbuf, ed->DirPath);
		strcat (fnbuf, "/checkp_");
		strcat (fnbuf, rd->OutputFile);
		unlink (fnbuf);
		ExModified++;
		return (rd->Status = COMPLETED);
	     }
	}
	/* Checkpointed ? */
	rewind (df);
	while ((nc = getc (df)) != EOF && ((char)nc) != CMAGIC);
	fclose (df);
	if (nc == EOF) {
	    if (Verbose) logMsg ("Output file incomplete");
	    goto AssumeUnfinished;
	}
	/* The experiment has been checkpointed. Mark it as such. It will be */
	/* taken care of later.						     */
	if (Verbose) logMsg ("Checkpointed");
	/* Save the checkpoint time */
	rd->LastCheckp = time (0);
	ExModified++;
	return (rd->Status = CHECKPOINTED);
    }

    /* Read the process list.  We assume that the list consists of a sequence */
    /* of lines, each line whose first nonblank character is a digit contains */
    /* a process id.							      */
	
    sl = 0;
    while (1) {
	/* Find first nonblank character */
	while ((nc = getc (TempFile)) != EOF && isspace (nc));
	if (nc == EOF) {
EndPsResponse:
	    if (sl) {
	        if (rd->Status == UNCERTAIN) {
	            /* Revert to STARTED */
	            ExModified++;
	            rd->Status = STARTED;
		}
	        goto AssumeDead;
            } else
		goto AssumeHostDead;
	}
	if (!isdigit (nc)) {
	    /* Ignore everything until end of line */
SkipToEol:
	    while ((nc = getc (TempFile)) != EOF && nc != '\n');
	    if (nc == EOF) goto EndPsResponse;
	    continue;
	}
	sl = 1;
	for (pid = nc - '0'; (nc = getc (TempFile)) && isdigit (nc);)
	pid = pid * 10 + (nc - '0');
	if (pid == rd->Pid) {
	    if (Verbose > 1) logMsg ("Experiment in progress");
            if (rd->Status == UNCERTAIN) {
                /* Revert to STARTED */
                ExModified++;
                rd->Status = STARTED;
	    }
	    return (STARTED);
	}
	if (nc != '\n') goto SkipToEol;
    }
}

int getTmpLine (buf) char *buf; {

/* -------------------------------- */
/* Reads one line from the tmp file */
/* -------------------------------- */

    int cnt, ch;

    for (cnt = 0; ; cnt++) {
	if ((ch = getc (TempFile)) == EOF) {
	    if (cnt == 0)
		return (EOF);
	    else
		break;
	}
	if (ch == '\n') break;
	if (cnt < MAXPATHLEN-1) *buf++ = ch;
    }
    *buf = '\0';
    return (0);
}

int available (hd) struct HostDesc *hd; {

/* ---------------------------------------- */
/* Determines whether the host is available */
/* ---------------------------------------- */

    struct tm *tim;
    long   clock;
    short  thr;
    int	   wkd, i;

    clock = time (0);
    tim = localtime (&clock);
    thr = (tim->tm_hour) * 100 + tim->tm_min;
    wkd = tim->tm_wday;

    for (i = 0; i < MXATINTVL; i++) {
	if (hd->STime [wkd][i] < 0) return (0);
        if (hd->STime [wkd][i] <= hd->ETime [wkd][i]) {
            if (thr >= hd->STime [wkd][i] && thr <= hd->ETime [wkd][i])
		return (1);
        } else {
            if (thr >= hd->STime [wkd][i] || thr <= hd->ETime [wkd][i])
		return (1);
	}
    }
    return (0);
}
			
int loadCheck (hd) struct HostDesc *hd; {

/* ------------------------------------------- */
/* Determines the load of the specified server */
/* ------------------------------------------- */

    char  tc [MAXPATHLEN], mc [128];
    int	  sc, ec, nc, ac, dg;
    float ld;
    char  cc;

    if (Verbose > 1) {
	sprintf (tc, "Checking load on '%s'", hd->Host);
	logMsg (tc);
    }
    if (hd->IA != -1) {
	/* Already known, so there is no need to go through all this mess */
	return (hd->IA);
    }

    /* Check if the host is available */
    if (!available (hd)) return (hd->IA = UNAVAILABLE);

    if (remoteCommand (hd->Host, "w")) {
	/* No response, assume it is BUSY */
Busy:
	return (hd->IA = BUSY);
    }

    /* Read the first line and locate the position of 'average' */

    while (getTmpLine (tc) != EOF) {
	for (sc = 0, ec = strlen (tc) - 10; sc < ec; sc++)
	    if (strncmp ("average:", &(tc [sc]), 8) == 0) {
		/* Found */
		ld = atof (&(tc [sc + 8]));
/*@@@sprintf (tc, "Load: %f", ld); logMsg (tc);*/
		goto LoadFound;
	}
    }
    /* Couldn't locate this line, assume that the host is busy */
Illegal:
    if (Verbose) {
        sprintf (tc, "Host %s 'w' response screwed up:", hd->Host);
	logMsg (tc);
	rewindTempFile ();
	logFile (TempFile);
    }
    goto Busy;

LoadFound:

    if (ld > hd->ELoad) {
Loaded:
	/* Already know that the host is overloaded, no need to go further */
	return (hd->IA = OVERLOADED);
    } else if (ld >= hd->TLoad) {
	hd->IA = BUSY;
    } else {
	hd->IA = IDLE;
    }
    if (hd->MIA == 0) return (hd->IA);	/* No need to check interactive */

    /* Read the next line and locate the position of 'idle' */

    while (getTmpLine (tc) != EOF) {
	for (sc = 0, ec = strlen (tc) - 4; sc < ec; sc++)
	    if (strncmp ("idle", &(tc [sc]), 4) == 0) {
		/* Found */
		ec = sc + 4;	/* LIMIT */
		/* Backspace with sc through leading blanks */
		if (sc > 0) {
		    for (sc--; sc >= 0 && tc [sc] == ' '; sc--);
		    sc++;
		}
		goto IdleFound;
	}
    }
    goto Illegal;

IdleFound:

    if (ec > sc + 127) ec = sc + 127;
    while (getTmpLine (tc) != EOF) {
	if (strlen (tc) < sc + 4) continue;
	/* Extract the interesting part */
	for (ac = 0, nc = sc; nc < ec; nc++, ac++) {
	    if ((cc = tc [nc]) == '\0') break;
	    mc [ac] = cc;
	}
	mc [ac] = '\0';
/*@@@logMsg ("MC follows:");*/
/*@@@logMsg (mc);*/
	/* Check if there are any characters other than spaces and digits */
	for (nc = 0, ac = 0, dg = 0; ; nc++) {
	    if ((cc = mc [nc]) == '\0') break;
	    if (isdigit (cc))
		dg++;
	    else if (cc != ' ')
		ac++;
	}
	if (ac) continue;	/* Idle for hours or days */
	if (!dg || atoi (mc) < IATHRESHOLD)
	    /* Overloaded */
	    return (hd->IA = OVERLOADED);
    }
    return (hd->IA);
}

checkpointExperiment (hd, ce, cr)

/* ------------------------------------ */
/* Checkpoints the specified experiment */
/* ------------------------------------ */

  struct HostDesc *hd;
  struct ExpClass *ce;
  struct RunDesc  *cr;
{
    char tc [MAXPATHLEN];

    if (Verbose) {
	sprintf (tc, "Checkpointing %1d on '%s'", cr->Pid, hd->Host);
	logMsg (tc);
    }
    sprintf (tc, "kill -USR1 %1d", cr->Pid);
    remoteCommand (hd->Host, tc);
    /* Leave it right here. If the above doesn't work -- tough luck, */
    /* otherwise,  the checkpointed status of the apparently STARTED */
    /* experiment will be detected next time.			     */
}

	
monitorActive () {

/* ---------------------------------------------------- */
/* Go through active experiments and check their status */
/* ---------------------------------------------------- */

    struct ExpClass  *ce;
    struct RunDesc   *cr;
    struct HostDesc  *hd;
    int    li;

    for (ce = Experiments; ce != NULL; ce = ce->next) {
	for (cr = ce->RunList; cr != NULL; cr = cr->next) {
	    if (cr->Status != STARTED && cr->Status != UNCERTAIN) continue;
	    /* Find the host description */
	    for (hd = Hosts; hd != NULL; hd = hd->next)
		if (strcmp (hd->Host, cr->Host) == 0) break;
	    if (hd == NULL) {
		logMsg ("Internal error -- monitorActive can't locate host");
		continue;
	    }
	    if (stillAlive (hd, ce, cr) == STARTED) {
		/* Checkpoint test */
		if ((li = loadCheck (hd)) == OVERLOADED || li == UNAVAILABLE) {
		    if (Verbose) {
		      if (li == OVERLOADED)
			logMsg ("Checkpoint request on high load");
                      else
			logMsg ("Checkpoint request: host becomes unavailable");
		    }
		    checkpointExperiment (hd, ce, cr);
		} else if (hd->PERC && (time (0) - cr->LastCheckp >
		  CHKPINTERVAL)) {
		    if (Verbose)
			logMsg ("Periodic checkpoint request");
		    checkpointExperiment (hd, ce, cr);
		}
	    }
	}
    }
}

int usedHost (hd) struct HostDesc *hd; {

/* --------------------------------------------------- */
/* Checks whether the specified host is already "used" */
/* --------------------------------------------------- */

    struct ExpClass  *ce;
    struct RunDesc   *cr;
    int		     nu;

    for (nu = 0, ce = Experiments; ce != NULL; ce = ce->next)
	for (cr = ce->RunList; cr != NULL; cr = cr->next)
	    if (cr->Status == STARTED && strcmp (cr->Host, hd->Host) == 0) {
		if (++nu >= hd->ECount) return (1);
	    }
    return (0);
}

struct RunDesc *findUntouched (hd, ce, pd)

/* ------------------------------------ */
/* Finds a run to be started on host hd */
/* ------------------------------------ */

  struct HostDesc  *hd;
  struct ExpClass  **ce;
  struct ProgDesc  **pd;
{
    struct RunDesc   *cr;

    for ((*ce) = Experiments; (*ce) != NULL; (*ce) = (*ce)->next) {
	for (cr = (*ce)->RunList; cr != NULL; cr = cr->next) {
	    if (cr->Status != UNTOUCHED) continue;
	    for ((*pd) = (*ce)->ProgList; (*pd) != NULL; (*pd) = (*pd)->next)
		if (strcmp ((*pd)->MType, hd->MType) == 0) return (cr);
	}
    }
    return (NULL);
}

struct RunDesc *findCheckpointed (hd, ce, pd)

/* --------------------------------------------------- */
/* Finds a checkpointed run to be restarted on host hd */
/* --------------------------------------------------- */

  struct HostDesc  *hd;
  struct ExpClass  **ce;
  struct ProgDesc  **pd;
{
    struct RunDesc   *cr;      struct HostDesc   *hh;

    for ((*ce) = Experiments; (*ce) != NULL; (*ce) = (*ce)->next) {
	for (cr = (*ce)->RunList; cr != NULL; cr = cr->next) {
	    if (cr->Status != CHECKPOINTED) continue;
	    for (hh = Hosts; hh != NULL; hh = hh->next)
		if (strcmp (hh->Host, cr->Host) == 0) break;
	    if (hh == NULL) {
		logMsg ("Internal error -- findCheckpointed can't locate host");
		continue;
	    }
	    if (strcmp (hh->MType, hd->MType) != 0)  continue;
	    for ((*pd) = (*ce)->ProgList; (*pd) != NULL; (*pd) = (*pd)->next)
		if (strcmp ((*pd)->MType, hd->MType) == 0) return (cr);
	}
    }
    return (NULL);
}

doCheckpointed () {

/* ------------------------------------ */
/* Try to resume checkpointed processes */
/* ------------------------------------ */

    struct ExpClass  *ce;
    struct RunDesc   *cr;
    struct HostDesc  *hd;
    struct ProgDesc  *pd;

    for (hd = Hosts; hd != NULL; hd = hd->next) {
	if (usedHost (hd)) continue;
	if ((cr = findCheckpointed (hd, &ce, &pd)) == NULL) continue;
	if (loadCheck (hd) != IDLE) continue;
	restartCheckpointed (hd, ce, cr, pd);
    }
}

tryStartingNew () {

/* ---------------------------- */
/* Try starting new experiments */
/* ---------------------------- */

    struct ExpClass  *ce;
    struct RunDesc   *cr;
    struct HostDesc  *hd;
    struct ProgDesc  *pd;

    for (hd = Hosts; hd != NULL; hd = hd->next) {
	if (usedHost (hd)) continue;
	if ((cr = findUntouched (hd, &ce, &pd)) == NULL) continue;
	if (loadCheck (hd) != IDLE) continue;
	startExperiment (hd, ce, cr, pd);
    }
}

writeExperimentsFile () {

/* -------------------------------------------------------------- */
/* Writes back a possibly updated version of the experiments file */
/* -------------------------------------------------------------- */

    struct ExpClass *ex;
    struct ProgDesc *pg;
    struct RunDesc  *rn;

    rewind (ExpFile);
    for (ex = Experiments; ex != NULL; ex = ex->next) {

	fprintf (ExpFile, "* %s", ex->DirPath);
	for (pg = ex->ProgList; pg != NULL; pg = pg->next)
	    fprintf (ExpFile, " [%s %s]", pg->MType, pg->PName);
	if (ex->Options != NULL)
	    fprintf (ExpFile, " %s", ex->Options);
	fprintf (ExpFile, "\n\n");

	for (rn = ex->RunList; rn != NULL; rn = rn->next) {
	    fprintf (ExpFile, "	 @ %s %s", rn->DataFile, rn->OutputFile);
	    if (rn->Status != UNTOUCHED) {
		fprintf (ExpFile, " %s", rn->Host);
		if (rn->Status == STARTED)
		    fprintf (ExpFile, " -");
		else if (rn->Status == COMPLETED)
		    fprintf (ExpFile, " +");
		else if (rn->Status == CHECKPOINTED)
		    fprintf (ExpFile, " !");
		else if (rn->Status == UNCERTAIN)
		    fprintf (ExpFile, " ?");
	    }

	    /* Process id */
	    if (rn->Status == STARTED || rn->Status == UNCERTAIN)
		fprintf (ExpFile, " %1d", rn->Pid);
            /* Last checkpoint time */
	    if (rn->Status == STARTED || rn->Status == CHECKPOINTED ||
	      rn->Status == UNCERTAIN)
		fprintf (ExpFile, " %1d", rn->LastCheckp);
	    /* Number of tries on uncertain status */
	    if (rn->Status == UNCERTAIN)
		fprintf (ExpFile, " %1d", rn->NTries);
	    fprintf (ExpFile, "\n");
	}
	fprintf (ExpFile, "\n");
    }
    truncate (EXPFILE, ftell (ExpFile));
}
