#include	<sys/param.h>
#include	<sys/time.h>
#include	<sys/types.h>
#include	<sys/file.h>
#include	<sys/stat.h>
#include	<sys/wait.h>
#include	<stdio.h>
#include	<errno.h>
#include	<signal.h>
#include	<ctype.h>
#include	<setjmp.h>
#include	<math.h>

#define	NOTIFY		1	/* Call smurph/lansf with -m */

#define CHECKINTERVAL	(150)	/* Seconds */
#define	CHKPINTERVAL	(10800) /* Three hours */
#define IATHRESHOLD	20	/* Twenty minutes */
#define	NTRIES		5	/* Number of tries on UNCERTAIN status */
#define RTIMEOUT	120L	/* Remote command timeout in seconds */
#define CBUFSIZE	(4*MAXPATHLEN+256)
#define	MXATINTVL	4	/* Maximum number of availability intervals */

#define UNTOUCHED	0	/* Run status codes */
#define STARTED		1
#define COMPLETED	2
#define CHECKPOINTED	3
#define	UNCERTAIN	4

/* --------------- */
/* Load indication */
/* --------------- */
#define IDLE		0	/* Can try starting new experiments */
#define BUSY		1	/* Do not start new experiments */
#define OVERLOADED	2	/* Swap out existing experiments */
#define UNAVAILABLE     3       /* Outside availability hours */

/* --------------------------------------------------------------------- */
/* All the following file names are relative to the directory serdel has */
/* been called from							 */
/* --------------------------------------------------------------------- */
#define EXPFILE		"Experiments"
#define HOSTFILE	"Hosts"

/* --------------------------------------------- */
/* The string terminating a complete output file */
/* --------------------------------------------- */
#define ENDSTRING	"@@@ End of output"

/* --------------------------------------------------------------------- */
/* The magic byte to tell the beginning of checkpoint information in the */
/* output file								 */
/* --------------------------------------------------------------------- */
#define CMAGIC		((char)(-2))

struct	ProgDesc {

/* --------------------------------------------------------------------- */
/* A  program description. Consists of a program name and the associated */
/* machine type.							 */
/* --------------------------------------------------------------------- */

    struct  ProgDesc *next;	/* Pointer to the next description (or NULL) */
    char    *MType, *PName;
};

struct	RunDesc {

/* --------------------------------------------------------------------- */
/* Description	of  a single run. Consists of the data file name and the */
/* output file name.							 */
/* --------------------------------------------------------------------- */

    struct  RunDesc *next;	/* These things are linked into lists */
    char    *DataFile, *OutputFile;
    int	    Status;		/* Started, Completed, Untouched */
    int	    Pid,		/* Relevant if Status == STARTED */
	    NTries;		/* Relevant if Status == UNCERTAIN */
    char    *Host;		/* Relevant if Status == STARTED or COMPLETED */
    unsigned long LastCheckp;	/* Last checkpoint time */
};

struct	HostDesc {

/* ---------------- */
/* Host description */
/* ---------------- */

    struct  HostDesc *next;
    char    *Host, *MType,
	    *PsCommand,		/* How to get the ps -x info */
	    *Guard;		/* Guard process name */
    float   TLoad,		/* Threshold load to start */
	    ELoad;		/* Threshold load to swap out */
    int	    ECount,		/* Maximum number of active experiments */
	    Pri;		/* Priority 0-20 */
    int	    MIA,		/* Monitor interactive activities */
	    PERC;		/* Periodic checkpointing */
    short   STime [7][MXATINTVL],   /* Availability time */
	    ETime [7][MXATINTVL];   /* from STime to ETime */
    int	    IA;			/* Used locally -- load flag */
};

struct	ExpClass {

/* --------------------------------------------------------------------- */
/* This	 structure describes a single experiment class. This description */
/* consists of a directory path, program names associated  with	 machine */
/* types,  and the list of options to be passed to the programs. This is */
/* followed by a list of data files together  with  their  corresponding */
/* output files.							 */
/* --------------------------------------------------------------------- */

    struct  ExpClass  *next;	/* Linked */
    char    *DirPath, *Options;
    struct  ProgDesc  *ProgList;
    struct  RunDesc   *RunList;
};


struct	ExpClass  *Experiments = NULL;
struct	HostDesc  *Hosts = NULL;

FILE	*ExpFile = NULL,
	*HostFile = NULL,
	*TempFile = NULL;

char	tmpfn [] = "/tmp/srdlXXXXXX";	/* Tmp file name */

int	ExModified,			/* Experiments file change flag */
	Verbose;

char	*malloc (), *getenv ();
