/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* --------------------------- */
/* Smurph communication module */
/* --------------------------- */

#include        "types.h"
#include        "var.h"
#include        "../SIMULATOR/display.h"

static  int     peekoctet;              // Current byte read from smurph

int     getOctet () {

/* -------------------------- */
/* Reads one byte from smurph */
/* -------------------------- */

	int     r;
	char    c;

	if (peekoctet != NONE) {
		// Never get past ESP
		if ((r = peekoctet) != PH_ESP) peekoctet = NONE;
		return (r);
	}

	if (read (SmrSocket, &c, 1) < 1)
		abtDisplay ("Read failed: connection broken by smurph");

	if ((r = ((int)c) & 0377) == PH_ESP) peekoctet = PH_ESP;
	return (r);
}

int     peekOctet () {

/* ---------------------------------------------------------- */
/* Look  at  the  next  octet  arriving  from  smurph without */
/* reading it                                                 */
/* ---------------------------------------------------------- */

	char    c;

	if (peekoctet == NONE) {
		if (read (SmrSocket, &c, 1) < 1)
			abtDisplay ("Read failed: connection broken by smurph");
		peekoctet = ((int)c) & 0377;
	}

	return (peekoctet);
}

void    sendOctet (int val) {

/* ----------------------- */
/* Send one byte to smurph */
/* ----------------------- */

	char    c;

	c = val;
	if (write (SmrSocket, &c, 1) < 0)
		abtDisplay ("Write failed: connection broken by smurph");
}

static  int     dscrhalf, dscrlast;

#ifndef ZZ_ATT
    inline
#endif
	static  int     getDescriptor () {

/* --------------------------------------------------- */
/* Gets one numeric descriptor from the input sequence */
/* --------------------------------------------------- */

	if (dscrhalf) {
		// Take the second half
		dscrhalf = NO;
		return (dscrlast & 017);
	}

	// Get a new byte

	dscrlast = getOctet ();
	dscrhalf = YES;
	return ((dscrlast >> 4) & 017);
}

void    sendText (char *t) {

/* -------------------------- */
/* Send text string to smurph */
/* -------------------------- */

	int     l;

	sendOctet (PH_TXT);
	while (*t != '\0') {
		sendOctet ((l = *t) & 0377);
		t++;
	}
	sendOctet (0);
}

long    getInt () {

/* --------------------------------------------- */
/* Reads an integer number from the input stream */
/* --------------------------------------------- */

	long    res;
	int     des;

	// This shouldn't really happen
	if (peekOctet () != PH_INM) return (0);
	getOctet ();

	res = 0; dscrhalf = NO;
	while (1) {
		if ((des = getDescriptor ()) == DSC_ENP)
			// Positive number
			return (-res);
		else if (des == DSC_ENN)
			// Negative number
			return (res);
		else if (des > 9)
	abtDisplay ("Illegal integer number sent by smurph (internal error)");

		res = res * 10 - des;
	}
}

BIG     getBIG () {

/* ---------------------------------------- */
/* Reads a BIG number from the input stream */
/* ---------------------------------------- */

	BIG     res;
	int     des;

	// This shouldn't really happen
	if (peekOctet () != PH_BNM) return (BIG_0);
	getOctet ();

	if (BigPrecision < 0)
		abtDisplay ("Big precision range not declared by smurph");

	res = BIG_0; dscrhalf = NO;
	while (1) {
		if ((des = getDescriptor ()) == DSC_ENP) {
			return (res);
			// Note: no negative big numbers allowed
		} else if (des > 9)
	abtDisplay ("illegal BIG number sent by smurph (internal error)");
		res = (res * 10) + des;
	}
}

double  getFloat () {

/* --------------------------------------------------- */
/* Reads a floating point number from the input stream */
/* --------------------------------------------------- */

	double  res, fac, frac;
	int     des, exp;

	if (peekOctet () != PH_FNM) return (0.0);
	getOctet ();

	res = 0.0; dscrhalf = NO;

	// Before dot

	while (1) {
		if ((des = getDescriptor ()) == DSC_ENP)
			break;
		else if (des == DSC_DOT)
			goto GF_DOT;
		else if (des == DSC_ENN) {
			res = -res;
			break;
		} else if (des > 9) {
GF_SPERR:
	abtDisplay ("Illegal fp number sent by smurph (internal error)");
		}
		res = res * 10.0 + (double) des;
	}

	// Now for the fractional part

	if (getDescriptor () != DSC_DOT) goto GF_SPERR;
GF_DOT:
	fac = 1.0;
	frac = 0.0;

	while (1) {
		if ((des = getDescriptor ()) == DSC_ENP)
			break;
		else if (des == DSC_EXP)
			goto GF_EXP;
		else if (des == DSC_ENN) {
			frac = -frac;
			break;
		} else if (des > 9) goto GF_SPERR;

		frac = frac + (((double) des) * (fac *= 0.1));
	}

	// Now for the exponent

	if (getDescriptor () != DSC_EXP) goto GF_SPERR;
GF_EXP:
	res += frac;
	exp = 0;

	while (1) {
		if ((des = getDescriptor ()) == DSC_ENP)
			break;
		else if (des == DSC_ENN) {
			exp = -exp;
			break;
		} else if (des > 9) goto GF_SPERR;

		exp = exp * 10 + des;
	}

	fac = pow (10.0, exp);
	res *= fac;
	return (res);
}

static  int     lastsdscr, lastshalf;

inline  static  void    sendDescriptor (int d) {

/* ----------------------------- */
/* Send one numerical descriptor */
/* ----------------------------- */

	if (lastshalf) {
		sendOctet (lastsdscr | (d & 017));
		lastshalf = NO;
	} else {
		lastshalf = YES;
		lastsdscr = (d & 017) << 4;
	}
}

inline  static  void    flushDescriptor () {

/* -------------------------------------------------- */
/* Flush a partially filled descriptor byte to smurph */
/* -------------------------------------------------- */

	if (lastshalf) {
		sendOctet (lastsdscr);
		lastshalf = NO;
	}
}

void    sendInt (long val) {

/* -------------------------------- */
/* Send an integer number to smurph */
/* -------------------------------- */

	int     digits [16], ndigits, sig;

	sendOctet (PH_INM);

	lastshalf = NO;

	if (val >= 0) {
		sig = NO;
		val = -val;
	} else
		sig = YES;

	for (ndigits = 0; val; ) {
		digits [ndigits++] = - (int)(val % 10);
		val /= 10;
	}

	while (ndigits--) sendDescriptor (digits [ndigits]);
	sendDescriptor (sig ? DSC_ENN : DSC_ENP);
	flushDescriptor ();
}

void    sendBIG (BIG val) {

/* --------------------------- */
/* Send a BIG number to smurph */
/* --------------------------- */

	int     digits [256], ndigits;

	sendOctet (PH_BNM);

	lastshalf = NO;

	for (ndigits = 0; val != BIG_0; ) {
		digits [ndigits++] = val % 10;
		val /= 10;
	}

	while (ndigits--) sendDescriptor (digits [ndigits]);
	sendDescriptor (DSC_ENP);
	flushDescriptor ();
}

#if     ZZ_ATT && ZZ_OWNLIB
extern "C" {
#if	ZZ_ATT_SUN
char    *sprintf (char*, const char*, ...);
#else
int     sprintf (char*, const char*, ...);
#endif
}
#endif

void    sendFloat (double val) {

/* -------------------------------------- */
/* Send a floating point number to smurph */
/* -------------------------------------- */

	char    *digits;
	int     sign;
#if ZZ_OWNLIB == 1
	char    tmp [32];               // Do not use 'dtoa'
#endif

	sendOctet (PH_FNM);
	lastshalf = NO;
#if ZZ_OWNLIB == 1
	sprintf (tmp, "%g", val);       // Does the same as 'dtoa'
	digits = tmp;
#else
	digits = dtoa (val);
#endif
	sign = NO;

	while (*digits != '\0') {
		if (*digits == ' ' || *digits == '+') {
			digits++;
			continue;
		}

		if (*digits == '-') {
			sign = YES;
			digits++;
			continue;
		}

		if (*digits == '.' || *digits == 'e' || *digits == 'E') {
			// No leading part
			sendDescriptor (DSC_ENP);
			sendDescriptor (DSC_DOT);
			if (*digits != '.') {
				sendDescriptor (DSC_ENP);
				sendDescriptor (DSC_EXP);
				sendDescriptor (DSC_ENP);
				flushDescriptor ();
				return;
			}
			break;          // Go process fractional part
		}

		// Process leading part (we assume that the dtoa output
		// is sane

		while (*digits <= '9' && *digits >= '0')
			sendDescriptor ((*digits++) - '0');
		sendDescriptor (sign ? DSC_ENN : DSC_ENP);
		sendDescriptor (DSC_DOT);
		break;
	}

	// The integer part is done

	if (*digits == '\0') {
		sendDescriptor (DSC_ENP);
		sendDescriptor (DSC_EXP);
		sendDescriptor (DSC_ENP);
		flushDescriptor ();
		return;
	}

	if (*digits == '.') {
		// Process fractional part
		digits++;
		while (*digits <= '9' && *digits >= '0')
			sendDescriptor ((*digits++) - '0');
		sendDescriptor (sign ? DSC_ENN : DSC_ENP);
	} else {
		// Null fractional part
		sendDescriptor (DSC_ENP);
	}

	sendDescriptor (DSC_EXP);

	// Now the exponent

	if (*digits != 'e' && *digits != 'E') {
		sendDescriptor (DSC_ENP);
		flushDescriptor ();
		return;
	}

	digits++;

	sign = NO;

	while (*digits != '\0') {
		if (*digits == '+') {
			digits++;
			continue;
		}
		if (*digits == '-') {
			sign = YES;
			digits++;
			continue;
		}
		while (*digits <= '9' && *digits >= '0')
			sendDescriptor ((*digits++) - '0');
		break;
	}
	sendDescriptor (sign ? DSC_ENN : DSC_ENP);
	flushDescriptor ();
}

void    Window::requestDisplay () {

/* ------------------------------------- */
/* Sends a request to display the window */
/* ------------------------------------- */

	int       fc;

	sendOctet (PH_WND);
	sendText (objct->SName);
	sendInt (templ->mode);
	// Should be station relative ?
	if (sttn >= 0) {
		sendInt (sttn);
	}
	sendOctet (PH_TRN);
	sendInt (fc = activeFields ());
	sendOctet (PH_EPH);
}

static  void    addObject () {

/* ---------------------- */
/* Process the ADD phrase */
/* ---------------------- */

	char    on [SANE], sn [SANE], bn [SANE], nn [SANE];
	int     cnt;

	if (getOctet () != PH_TXT)
		abtDisplay ("No owner's name in ADD phrase (internal error)");

	for (cnt = 0; cnt < SANE; cnt++)
		if ((on [cnt] = getOctet ()) == '\0') break;

	if (cnt >= SANE)
		abtDisplay ("Owner's name in ADD phrase is too long");

	if (getOctet () != PH_TXT)
		abtDisplay ("No object name in ADD phrase (internal error)");

	for (cnt = 0; cnt < SANE; cnt++)
		if ((sn [cnt] = getOctet ()) == '\0') break;

	if (cnt >= SANE)
		abtDisplay ("Object name in ADD phrase is too long");

	// Get the base name
	if (getOctet () != PH_TXT)
		abtDisplay ("No base name in ADD phrase (internal error)");

	for (cnt = 0; cnt < SANE; cnt++)
		if ((bn [cnt] = getOctet ()) == '\0') break;

	if (cnt >= SANE)
		abtDisplay ("Object base name in ADD phrase is too long");

	// Get the nickname
	if (getOctet () != PH_TXT)
		abtDisplay ("No nickname in ADD phrase (internal error)");

	for (cnt = 0; cnt < SANE; cnt++)
		if ((nn [cnt] = getOctet ()) == '\0') break;

	if (cnt >= SANE)
		abtDisplay ("Object nickname in ADD phrase is too long");

	new TItem (on, sn, bn, nn);
	if (strncmp (bn, "Station ", 8) == 0) NStations++;
}

void    removeObject () {

/* ---------------------- */
/* Process the REM phrase */
/* ---------------------- */

	char    on [SANE];
	int     cnt;
	TItem   *obj;
	Window  *w, *u;

	if (getOctet () != PH_TXT)
		abtDisplay ("No object name in REM phrase (internal error)");

	for (cnt = 0; cnt < SANE; cnt++)
		if ((on [cnt] = getOctet ()) == '\0') break;

	if (cnt >= SANE)
		abtDisplay ("Object name in REM phrase is too long");

	if ((obj = findObj (on)) == NULL) {
		// This may be an EObject (or something even stranger) that
		// is sought by a BName; give it a second chance
		if ((obj = findByBName (on)) == NULL)
	      abtDisplay ("REM phrase for nonexistent object (internal error)");
	}

	// Remove windows associated with the object

	for (w = Active; w != NULL; w = u) {
		u = w->next;
		if (w->objct == obj) removeWindow (w);
	}
	// This won't happen: stations are never removed
	if (strncmp (obj->BName, "Station ", 8) == 0) NStations--;
	delete (obj);
}

void    updateWindow () {

/* -------------------------------------------------- */
/* Read the WCN phrase and update the window contents */
/* -------------------------------------------------- */

	char    sn [SANE], em [SANE];
	int     cnt, mode, sttn, code;
	Window  *w;

	if (getOctet () != PH_TXT)
		abtDisplay ("No object name in WCN phrase (internal error)");

	for (cnt = 0; cnt < SANE; cnt++)
		if ((sn [cnt] = getOctet ()) == '\0') break;

	if (cnt >= SANE)
		abtDisplay ("Object name in WCN phrase is too long");

	while (peekOctet () == PH_EPH) getOctet ();

	if (peekOctet () != PH_INM)
		abtDisplay ("Illegal format of the WCN phrase");

	mode = (int) getInt ();

	while (peekOctet () == PH_EPH) getOctet ();

	if (peekOctet () != PH_INM)
		abtDisplay ("Illegal format of the WCN phrase");

	sttn = (int) getInt ();

	// Check if it is an error message

	while (peekOctet () == PH_EPH) getOctet ();

	if (peekOctet () == PH_ERR) {
		// Yes, it is
		getOctet ();
		code = getOctet ();     // The code
		getOctet ();            // This was PH_TXT
		for (cnt = 0; cnt < SANE; cnt++)
			if ((em [cnt] = getOctet ()) == '\0') break;
		outRLC (form ("%s: (%1d) %s", sn, code, em));
		return;
	}

	// Find the window
	if (!WCNAdmitted) return;

	if (TermFlag) {
		skipWindow ();
		return;
	}

	for (w = Active; w != NULL; w = w->next) {

		if (w->sttn != sttn) continue;
		if (w->templ->mode != mode) continue;
		if (strcmp (w->objct->SName, sn) == 0) break;
	}

	if (w == NULL)
		skipWindow ();   // No such window
	else
		w->displayNew ();
}

static  void    errorPhrase (int type) {

/* ---------------------------------------- */
/* Displays the error message on the screen */
/* ---------------------------------------- */

	char    em [SANE];
	int     cnt, code;

	if (type == PH_ERR) code = getOctet ();
	getOctet ();            // This was PH_TXT
	for (cnt = 0; cnt < SANE; cnt++)
		if ((em [cnt] = getOctet ()) == '\0') break;
	if (type == PH_ERR)
		outRLC (form ("(%1d) %s", code, em));
	else {
		if (em [0] == '*') {
			TermFlag = YES;
			outRLC (&(em[1]));
		} else
			outRLC (em);
	}
}

void    readSmurphSpurt () {

/* ------------------------------------------------------- */
/* This function reads and processes one spurt from smurph */
/* ------------------------------------------------------- */

	int     pb, more;

	peekoctet = NONE;
	Stepped = NO;
	more = YES;

	while (more) {          // More stuff coming in

	 switch (pb = getOctet ()) {

	  int   d;

	  case  PH_ESP:         // The end of spurt

		more = NO;

	  case  PH_ATT:         // Attention -- ignore
	  case  PH_EPH:         // Spurious endphrase -- as above

		break;

	  case  PH_STP:         // Stepped windows included

		if (Stepped == NO) Stepped = YES;
		break;

	  case  PH_DIN:         // New display interval

		if ((d = peekOctet ()) != PH_INM)
		 abtDisplay ("Illegal DIN phrase from smurph (internal error)");

		DisplayInterval = getInt ();
		break;

	  case  PH_BPR:         // Big precision announcement

		if ((d = peekOctet ()) != PH_INM)
		 abtDisplay ("Illegal BPR phrase from smurph (internal error)");
		d = (int) getInt ();
		if (BigPrecision == -1) {       // Undefined yet
			if (d == 0)
				BigPrecision = 3;       // Double
			else
				BigPrecision = (d + 30) / 31;
			initBig ();
		}
		break;

	  case  PH_NST:         // Number of stations (ignored)

		if ((d = peekOctet ()) != PH_INM)
		 abtDisplay ("Illegal NST phrase from smurph (internal error)");
		d = (int) getInt ();
		break;

	  case  PH_EXI:         // Smurph terminates

		Stepped = NONE;
		break;

	  case  PH_ADD:         // Add new object

		addObject ();
		break;

	  case  PH_REM:         // Remove an object

		removeObject ();
		break;

	  case  PH_WCN:         // Window contents phrase

		updateWindow ();
		break;

	  case  PH_LCK:         // As error, but lock the simulator

		Stepped = 2;

	  case  PH_ERR:         // Error message

		// Process error phrase
		errorPhrase (pb);
		break;

	  default:

	   abtDisplay (form ("Illegal phrase (%d) from smurph (internal error)",
		pb));
	 }
	 while (peekOctet () == PH_EPH) getOctet ();
	}
}
