/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ---------------------------------------------------------- */
/* This  program monitors the execution of smurphs on a local */
/* network of computers  running  BSD  UNIX.  Each  simulator */
/* instance  is  supposed  to  report  to  the monitor, which */
/* operation  establishes  a  stream  socket  connecting  the */
/* simulator  to  the  monitor.  This socket is alive for the */
/* entire lifespan of the simulator instance and it  is  used */
/* by  the  monitor to identify the simulator instance within */
/* the domain of the local network.                           */
/* ---------------------------------------------------------- */

#include        <sys/types.h>
#include        <sys/socket.h>
#include        <sys/time.h>
#include        <netinet/in.h>
#include        <setjmp.h>
#include        <fcntl.h>
#include        <signal.h>

#include        "../version.h"

#ifdef	sgi
typedef	void (*SIGARG) (int, ...);
#else
#ifdef	ZZ_ATT_SUN
#define	SIGARG	SIG_PF
#else
typedef	void  (*SIGARG) ();
#endif
#endif

#ifdef	ZZ_ATT_SUN
#define	RWCAST	(char*)
#define	SCAST	(struct sockaddr*)
#else
#define	RWCAST
#define	SCAST
#endif

#include        "../SIMULATOR/display.h"
#include        "monitor.h"


int                     Master,         // The master socket
			LSk,            // Last "local" socket
			NDesc;          // Maximum number of descriptors
fd_set                  Alive;          // The set of alive sockets
struct  sockaddr_in     master_id;
jmp_buf                 restart;        // For timeouts

long                    NextHandle;     // Next handle value

#include	"smdesc.c"

void    setTimer () {

/* ------------------- */
/* Set the alarm clock */
/* ------------------- */

	struct  itimerval       td;

	td.it_value.tv_sec = GTIMEOUT;
	td.it_value.tv_usec = 0;
	td.it_interval.tv_sec = 0;
	td.it_interval.tv_usec = 0;
	setitimer (ITIMER_REAL, &td, NULL);
}

void    clearTimer () {

/* ---------------------- */
/* Clears the alarm clock */
/* ---------------------- */

	struct  itimerval       td;

	td.it_value.tv_sec = 0;
	td.it_value.tv_usec = 0;
	td.it_interval.tv_sec = 0;
	td.it_interval.tv_usec = 0;
	setitimer (ITIMER_REAL, &td, NULL);
}

void    TimeOut () {

/* --------------- */
/* Timeout service */
/* --------------- */

	longjmp (restart, 0);
}

void    processSmurph () {

/* --------------------------------------- */
/* Process a new smurph connection request */
/* --------------------------------------- */

	char    HName [128],    // Hostname
		PArgs [128],    // Program call arguments
		DTime [32] ;    // Startup date/time

	long    PId,            // Process ID
		dummy;

	int     i;
	char    c;

	SmDesc  *sm;

	// Read the host name

	for (i = 0; ;) {
		if (read (LSk, &c, 1) <= 0) longjmp (restart, 0);
		if (c == '\0') break;
		if (i < 127) HName [i++] = c;
	}

	HName [i] = '\0';

	// Read the program call line

	for (i = 0; ;) {
		if (read (LSk, &c, 1) <= 0) longjmp (restart, 0);
		if (c == '\0') break;
		if (i < 127) PArgs [i++] = c;
	}

	PArgs [i] = '\0';
		
	// Read the startup date/time

	for (i = 0; ;) {
		if (read (LSk, &c, 1) <= 0) longjmp (restart, 0);
		if (c == '\0') break;
		if (i < 31) DTime [i++] = c;
	}

	DTime [i] = '\0';

	// Read the process ID

	if (read (LSk, RWCAST &dummy, sizeof (long)) < sizeof (long))
		longjmp (restart, 0);

	PId = ntohl (dummy);

	clearTimer ();          // Critical section

	sm = new SmDesc;

	strcpy (sm->HName, HName);
	strcpy (sm->PArgs, PArgs);
	strcpy (sm->DTime, DTime);
	sm->PId = PId;
	sm->Socket = LSk;
	sm->Handle = NextHandle++;
	// Make sure handles are always nonnegative
	if (NextHandle < 0) NextHandle = 0;
	FD_SET (LSk, &Alive);
	LSk = -1;
}

void    processInfAll () {

/* ---------------------------------------------------- */
/* Inform the server about all smurphs currently active */
/* ---------------------------------------------------- */

	SmDesc  *sm;
	long    dummy;
	char    c;

	setTimer ();    // Give it more time

	for (sm = Smurphs; sm != NULL; sm = sm->next) {

		if (write (LSk, sm->HName, strlen (sm->HName)+1) < 0)
			longjmp (restart, 0);

		if (write (LSk, sm->PArgs, strlen (sm->PArgs)+1) < 0)
			longjmp (restart, 0);

		if (write (LSk, sm->DTime, strlen (sm->DTime)+1) < 0)
			longjmp (restart, 0);

		dummy = htonl (sm->PId);
		if (write (LSk, RWCAST &dummy, sizeof (dummy)) < 0)
			longjmp (restart, 0);

		dummy = htonl (sm->Handle);
		if (write (LSk, RWCAST &dummy, sizeof (dummy)) < 0)
			longjmp (restart, 0);

	}

	// Follow it by the ESP byte

	c = PH_ESP;

	if (write (LSk, &c, 1) < 0) longjmp (restart, 0);
}

void    processDisply () {

/* ----------------------- */
/* Process display request */
/* ----------------------- */

	SmDesc  *sm;
	long    sh, dummy;
	char    c;

	// Read the handle
	if (read (LSk, RWCAST &dummy, sizeof (long)) < sizeof (long))
		longjmp (restart, 0);

	setTimer ();            // This is getting tough, give them more time

	sh = ntohl (dummy);     // This is the handle as it should look

	for (sm = Smurphs; sm != NULL; sm = sm->next) {
		// Find it
		if (sm->Handle == sh) break;
	}

	if (sm == NULL) {
		// No such smurph: respond with -1
		dummy = htonl (-1);
		write (LSk, RWCAST &dummy, sizeof (long));
		return;
	}

	// Issue a write request to interrupt the smurph

	c = PH_ATT;             // Tells connection request from status request
	if (write (sm->Socket, &c, 1) < 1) {
		// Something wrong, assume the smurph is dead
		clearTimer ();
		delete (sm);
		setTimer ();
		dummy = htonl (-1);
		write (LSk, RWCAST &dummy, sizeof (long));
		return;
	};

	// The smurph is supposed to respond with socket port number

	if (read (sm->Socket, RWCAST &dummy, sizeof (long)) < sizeof (long)) {
		// Something wrong again, same thing
		clearTimer ();
		delete (sm);
		setTimer ();
		dummy = htonl (-1);
	};

	// The socket number is in the network format, send it as is

	if (write (LSk, RWCAST &dummy, sizeof (long)) < 0) longjmp (restart, 0);

	// Done
}

void    forwardText (SmDesc *sm) {

/* ------------------------------------------- */
/* Forwards a string from smurph to the server */
/* ------------------------------------------- */

	char    txt [128], c;
	int     n;

	for (n = 0; ;) {
		if (read (sm->Socket, &c, 1) < 1) {
			// Assume the smurph is dead
			clearTimer ();
			delete (sm);
			longjmp (restart, 0);
		}
		if (c == '\0') break;
		if (n < 127) txt [n++] = c;
	}
	txt [n++] = '\0';
	if (write (LSk, txt, n) < 0) longjmp (restart, 0);
}

void    processStatus () {

/* ---------------------- */
/* Process status request */
/* ---------------------- */

	// The status information contains:
	// 
	//      The number of received messages
	//      The simulated time
	//      The CPU execution time
	//
	// All items are textual (i.e. strings)

	SmDesc  *sm;
	long    sh, dummy;
	char    c;

	// Read the handle
	if (read (LSk, RWCAST &dummy, sizeof (long)) < sizeof (long))
		longjmp (restart, 0);

	setTimer ();

	sh = ntohl (dummy);     // This is the handle as it should look

	for (sm = Smurphs; sm != NULL; sm = sm->next) {
		// Find it
		if (sm->Handle == sh) break;
	}

	if (sm == NULL) {
		// No such smurph
		c = PH_ESP;
		// Respond with ESP byte
		write (LSk, &c, 1);
		return;
	}

	// Issue a write request to interrupt the smurph

	c = PH_RPL;             // Respond with status
	if (write (sm->Socket, &c, 1) < 1) {
		// Something wrong, assume the smurph is dead
		delete (sm);
		c = PH_ESP;
		write (LSk, &c, 1);
		return;
	};

	// The smurph is supposed to respond with three pieces of text
	// that are immediately forwarded to the caller

	forwardText (sm);
	forwardText (sm);
	forwardText (sm);
}

void    processSttall () {

/* -------------------------------------- */
/* Process status request for all smurphs */
/* -------------------------------------- */

	// The status information for each smurrph contains:
	// 
	//      The host name
	//      The call line
	//      The call date/time
	//      The process id
	//      The handle      (this and the preceding items are numerical)
	//      The number of received messages
	//      The simulated time
	//      The CPU execution time
	//
	// All items are textual (i.e. strings)

	SmDesc  *sm;
	long    dummy;
	char    c;

	for (sm = Smurphs; sm != NULL; sm = sm->next) {

		if (write (LSk, sm->HName, strlen (sm->HName)+1) < 0)
			longjmp (restart, 0);

		if (write (LSk, sm->PArgs, strlen (sm->PArgs)+1) < 0)
			longjmp (restart, 0);

		if (write (LSk, sm->DTime, strlen (sm->DTime)+1) < 0)
			longjmp (restart, 0);

		dummy = htonl (sm->PId);
		if (write (LSk, RWCAST &dummy, sizeof (dummy)) < 0)
			longjmp (restart, 0);

		dummy = htonl (sm->Handle);
		if (write (LSk, RWCAST &dummy, sizeof (dummy)) < 0)
			longjmp (restart, 0);

		setTimer ();

		// Issue a write request to interrupt the smurph

		c = PH_RPL;             // Respond with status
		if (write (sm->Socket, &c, 1) < 1) {
			// Something wrong, assume the smurph is dead
			delete (sm);
			c = PH_ESP;
			write (LSk, &c, 1);
			return;
		};

		// The smurph is supposed to respond with three pieces of text
		// that are immediately forwarded to the caller

		forwardText (sm);
		forwardText (sm);
		forwardText (sm);
	}

	// Follow it by the ESP byte

	c = PH_ESP;

	if (write (LSk, &c, 1) < 0) longjmp (restart, 0);
}
	
void    processServer () {

/* ------------------------------------- */
/* Process a request from display server */
/* ------------------------------------- */

	char    rtype;          // Request type

	if (read (LSk, &rtype, 1) < 1) longjmp (restart, 0);

	switch (rtype) {

	  case DR_INFALL:

	    // Information about all smurphs

	    processInfAll ();
	    break;

	  case DR_DISPLY:

	    // Connect request

	    processDisply ();
	    break;

	  case DR_STATUS:

	    // Status information of one smurph

	    processStatus ();
	    break;

	  case DR_STTALL:

	    // Status information of all smurphs

	    processSttall ();
	    break;

	  default:

	    // Some kind of junk -- ignore it
	    longjmp (restart, 0);
	}

	close (LSk);
}

main    () {

	// Initialization

	char	*st;
	char	pname [SIGNLENGTH+1];

	close (0);
	close (1);

	if ((Master = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
		// This shouldn't really happen
		st = "Monitor startup error: cannot open master socket\n";
		write (2, st, strlen (st));
		exit (-1);
	}

	master_id.sin_family = AF_INET;
	master_id.sin_addr.s_addr = htonl (INADDR_ANY);
	master_id.sin_port = htons (ZZ_MONSOCK);

	if (bind (Master, SCAST &master_id, sizeof (master_id))) {
		st =
"Monitor startup error: port probably taken, rebuild with another socket port\n"
	;
		write (2, st, strlen (st));
		exit (-2);
	}

	LSk = 2;                // Standard error not needed any longer

	NextHandle = 0;         // Start numbering smurphs

	NDesc = getdtablesize ();

	listen (Master, 5);

	FD_ZERO (&Alive);       // Initialize the socket set
	FD_SET  (Master, &Alive);

	setjmp (restart);       // For timeouts

	clearTimer ();

	signal (SIGALRM, (SIGARG) TimeOut);      // Timer service

	close (LSk);            // Close last "local" socket

	while (Master >= 0) {

		fd_set  rr;     // Ready for reading
		SmDesc  *sp, *spp;
		int     dummy;

		// Wait for a new connection or for a disappearance of a smurph
		bcopy (&Alive, &rr, sizeof (Alive));
		if (select (NDesc, &rr, NULL, NULL, NULL) <= 0) {
			// Something wrong (shouldn't really happen)
			sleep (2);
			continue;
		}

		setTimer ();    // Start counting

		// Check if a closing request

		for (sp = Smurphs; sp != NULL; sp = spp) {
			spp = sp->next;
			if (FD_ISSET (sp->Socket, &rr)) {

				// Issue a dummy read request
				if (read (sp->Socket, RWCAST &dummy, 1) <= 0) {
					close (sp->Socket);
					FD_CLR (sp->Socket, &Alive);
					delete (sp);
				}
			}
		}

		if (FD_ISSET (Master, &rr)) {
			// Incoming connection
			if ((LSk = accept (Master, NULL, &dummy)) < 0)
				// Ignore if fails
				longjmp (restart, 0);

			// Read the signature
			if (read (LSk, pname, SIGNLENGTH) < SIGNLENGTH)
				longjmp (restart, 0);

			pname [ SIGNLENGTH ] = '\0';
			if (strcmp (pname, SMRPHSIGN) == 0) {
				// A new smurph
				processSmurph ();
			} else if (strcmp (pname, SERVRSIGN) == 0) {
				// A server request
				processServer ();
			} else
				longjmp (restart, 0);

		}

	}
}
