/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ---------------------- */
/* The standard Client AI */
/* ---------------------- */

#include        <stdarg.h>
#include        "system.h"

/* --------------------------------------------- */
/* Create with nickname for Traffic's RVariables */
/* --------------------------------------------- */

#define rvcreate(vn,nn,txp)\
  { char *zz_cc; zz_cc = (nn); tp->vn = new RVariable; \
    tp->vn->zz_start (); tp->vn->zz_nickname = \
    new char [strlen(zz_cc)+1]; strcpy (tp->vn->zz_nickname, zz_cc);\
    tp->vn->setup (txp, 2); }

/* ---------------------------------------------------------- */
/* Create  for temporary RVariables (bypasses the nonstandard */
/* constructor                                                */
/* ---------------------------------------------------------- */

#define tcreate(t) { trv = new ZZ_TRVariable; trv->setup (t, 2); }

zz_client::zz_client () {

/* --------------- */
/* The constructor */
/* --------------- */

	Class = AIC_client;
	Id = NONE;
}

static  long cp2m1 (long n) {

/* ---------------------------------------------------------- */
/* Calculates  the  smallest  power  of two minus one greater */
/* than or equal to the nonnegative argument                  */
/* ---------------------------------------------------------- */

	long    res;

	for (res = 16; res <= n; res += res);

	return (res - 1);
}

/* --------------------------------------- */
/* Client service process type declaration */
/* --------------------------------------- */

class   ClientService : public ZZ_SProcess {

	public:

	Station *S;

	static  int     zz_ClientService_prcs;

	ClientService () {
		S = TheStation;
		zz_typeid = (void*) (&zz_ClientService_prcs);
	};

	virtual char *getTName () { return ("ClientService"); };

	enum {Start, GenMessage, GenBurst};

	private:

		Traffic         *mst;
		int             st, sr, i, tp;
		long            ml;
		TIME            t;

	public:

	void setup () { 

		// Initialize state name list

		zz_ns = 3;
		zz_sl = new char* [GenBurst + 1];
		zz_sl [0] = "Start";
		zz_sl [1] = "GenMessage";
		zz_sl [2] = "GenBurst";
	};

	void zz_code ();
};

void    ClientService::zz_code () {

     switch (TheState) {

	  case Start:

		// Initialize the process execution
		for (i = 0; i < NTraffics; i++) {
			// Schedule initial events for all traffic patterns
			mst = idToTraffic (i);
			if (mst->FlgSCL == OFF) continue;
			if (mst->DstBIT != UNDEFINED) {
				// Schedule the burst generation event
				new ZZ_EVENT (zz_sentinel_event,
					Time + mst->genBIT (), System, NULL,
						(void*) i, TheProcess, Client,
							ARR_BST, GenBurst,
								NULL);
			} else if (mst->DstMIT != UNDEFINED) {
				new ZZ_EVENT (Time + mst->genMIT (), System,
					NULL, (void*) i, TheProcess, Client,
						ARR_MSG, GenMessage, NULL);
			}
		}
			
	  break; case GenMessage:

		// Generate new message
		mst = idToTraffic (tp = TheTraffic);
		assert (mst->DstMIT != UNDEFINED,
		form ("Client: MIT distribution for Traffic %d undefined",
			tp));
		st = mst->genSND ();            // Generate sender
		assert (isStationId (st),
		form ("Client: cannot generate sender for Traffic %s", 
			mst->getSName ()));
		ml = mst->genMLE ();            // Generate message length

		mst->genMSG (st, mst->genRCV (), ml);

		// Schedule next message generation event
		if (mst->DstBIT == UNDEFINED || --(mst->RemBSI) > 0) {
			new ZZ_EVENT (Time + mst->genMIT (), System, NULL,
				(void*) tp, TheProcess, Client,
					ARR_MSG, GenMessage, NULL);
		}

	  break; case GenBurst:

		// Generate new burst
		mst = idToTraffic (tp = TheTraffic);
		assert (mst->DstBIT != UNDEFINED,
		form ("Client: Traffic %d is not bursty", tp));
		ml = mst->genBSI ();            // Generate burst size
		if (! mst->RemBSI && ml)
			// Schedule message generation event
			new ZZ_EVENT (Time + mst->genMIT (), System, NULL,
				(void*) tp, TheProcess, Client,
					ARR_MSG, GenMessage, NULL);
		mst -> RemBSI += ml;
		// Schedule next burst generation event
		new ZZ_EVENT (zz_sentinel_event, Time + mst->genBIT (), System,
			NULL, (void*) tp, TheProcess, Client, ARR_BST,
				GenBurst, NULL);
     }
};

void    zz_start_client () {            // Starts the standard Client

	Process         *phandle;
	Traffic         *tp;
	int             i, j;
	CGroup          *cg;
	SGroup          *sg;

	// Postprocess traffic patterns

	for (i = 0; i < NTraffics; i++) {
		tp = idToTraffic (i);
		for (j = 0; j < tp->NCGroups; j++) {
			cg = tp->CGCluster [j];
			Assert (cg != NULL, form (
"Client: Traffic %1d -- CGroup %1d undefined, but a higher group defined", i,
j));
			sg = cg->Senders;
			Assert (sg != NULL, form (
"Client: Traffic %1d, CGroup %1d -- Senders not defined", i, j));
			zz_sort_group (sg, cg->SWeights);
			Assert (sg->NStations, form (
"Client: Traffic %1d, CGroup %1d -- The sender group has zero weight", i, j));

			if ((sg = cg->Receivers) != NULL) {
			    if (cg->RWeights != (float*) (-1)) {
				zz_sort_group (sg, cg->RWeights);
				Assert (sg->NStations, form (
"Client: Traffic %1d, CGroup %1d -- The receiver group has zero weight", i,j));
			    } else 
				zz_sort_group (sg);
			}
		}
		tp->preprocess_weights ();
	}

	// Initialize counters (global)

	zz_NGMessages = zz_NQMessages = zz_NTMessages = zz_NRMessages =
	zz_NTPackets = zz_NRPackets = 0;
	zz_NQBits = zz_NTBits = zz_NRBits = BITCOUNT_0;

	for (i = 0; i < NTraffics; i++) {

		tp = idToTraffic (i);

		// Initialize counters (local)

		tp->NQMessages = tp->NTMessages = tp->NRMessages =
		tp->NTPackets = tp->NRPackets = 0;
		tp->NQBits = tp->NTBits = tp->NRBits = BITCOUNT_0;
		tp->RemBSI = 0;        // Remaining burst size

#if     ZZ_CLI
	    if (zz_flg_stdClient) {

		// Determine if standard performance measures are to be
		// calculated

		if (tp->FlgSPF == ON) {

			// Generate standard RVariables and make them belong
			// to the Traffic

			rvcreate (RVAMD, "AMDelay", TYPE_long);
			// Move it to Traffic
			queue_out ((ZZ_Object*)(tp->RVAMD));
			queue_head ((ZZ_Object*)(tp->RVAMD), tp->ChList,
				ZZ_Object);
	
			rvcreate (RVAPD, "APDelay", TYPE_long);
			queue_out ((ZZ_Object*)(tp->RVAPD));
			queue_head ((ZZ_Object*)(tp->RVAPD), tp->ChList,
				ZZ_Object);
	
			rvcreate (RVWMD, "WMDelay", TYPE_BITCOUNT);
			queue_out ((ZZ_Object*)(tp->RVWMD));
			queue_head ((ZZ_Object*)(tp->RVWMD), tp->ChList,
				ZZ_Object);
	
			rvcreate (RVMAT, "MAcTime", TYPE_long);
			queue_out ((ZZ_Object*)(tp->RVMAT));
			queue_head ((ZZ_Object*)(tp->RVMAT), tp->ChList,
				ZZ_Object);
	
			rvcreate (RVPAT, "PAcTime", TYPE_long);
			queue_out ((ZZ_Object*)(tp->RVPAT));
			queue_head ((ZZ_Object*)(tp->RVPAT), tp->ChList,
				ZZ_Object);
	
			rvcreate (RVMLS, "MsgStat", TYPE_long);
			queue_out ((ZZ_Object*)(tp->RVMLS));
			queue_head ((ZZ_Object*)(tp->RVMLS), tp->ChList,
				ZZ_Object);
	
		}
	    }
#endif

	}

#if     ZZ_CLI
	if (zz_flg_stdClient) {
		// Create the Client service process
		// phandle = create ClientService;
		phandle = new ClientService;
		((ClientService*)phandle)->zz_start ();
		((ClientService*)phandle)->setup ();
	}
#endif

}

void    zz_client::zz_wait (int ev, int pstate) {

/* ----------------------- */
/* The Client wait request */
/* ----------------------- */

	if_from_observer ("Client->wait: called from an observer");
	assert (ev >= -2 && ev < NTraffics, "Client->wait: illegal event");

	if (zz_c_first_wait) {

		zz_c_other = NULL;    // No requests so far

		// Create new simulation event
		zz_c_wait_event = new ZZ_EVENT;
		zz_c_wait_event -> station = TheStation;
		zz_c_wait_event -> process = TheProcess;

		if (ev < 0)     // Unspecified Traffic
			new ZZ_REQUEST (&(TheStation->CWList), this, ev,pstate);
		else            // Force a Traffic wait request
			new ZZ_REQUEST (&(TheStation->CWList), idToTraffic (ev),
				ARRIVAL, pstate);

		assert (zz_c_other != NULL,
			"Client->wait: internal error -- null request");

		zz_c_wait_event -> pstate = pstate;
		zz_c_wait_event -> ai = zz_c_other->ai;
		zz_c_wait_event -> event_id = zz_c_other->event_id;
		zz_c_wait_event -> chain = zz_c_other;
		zz_c_wait_event -> Info01 = NULL;
		zz_c_wait_event -> Info02 = NULL;

		zz_c_whead = zz_c_other;
		zz_c_wait_event -> waketime =
			zz_c_wait_tmin = zz_c_other -> when = TIME_inf;
		zz_c_other -> event = zz_c_wait_event;
		zz_c_other -> other = zz_c_whead;

		zz_c_first_wait = NO;
		zz_c_wait_event->store ();

	} else {

		if (ev < 0)     // Unspecified Traffic
			new ZZ_REQUEST (&(TheStation->CWList), this, ev,
				pstate);
		else            // Force a Traffic wait request
			new ZZ_REQUEST (&(TheStation->CWList), idToTraffic (ev),
				ARRIVAL, pstate);

		assert (zz_c_other != NULL,
			"Client->wait: internal error -- null request");

		zz_c_other -> when = TIME_inf;
		zz_c_other -> event = zz_c_wait_event;
		zz_c_other -> other = zz_c_whead;
	}
}

void    Traffic::setQSLimit (long lim) {

/* ---------------------------------------------------------- */
/* Sets the limit on the number of messages belonging to this */
/* traffic  pattern  that  can  be  queued  at (all) stations */
/* awaiting transmission. When this  limit  is  reached,  new */
/* generated messages will be ignored.                        */
/* ---------------------------------------------------------- */

#if     ZZ_QSL
	Assert (FlgSPF == ON,
			"Traffic->setQSLimit illegal: SPF_off was selected");
	QSLimit = lim;
#else
	lim++;
	excptn ("Traffic->setQSLimit illegal: smurph not created with '-q'");
#endif
}

void    setQSLimit (long lim) {

/* ---------------------------------------------------------- */
/* Sets  the  global limit on the number of messages that can */
/* be queued at (all) stations  awaiting  transmission.  When */
/* this  limit  is  reached,  new  generated messages will be */
/* ignored.                                                   */
/* ---------------------------------------------------------- */

#if     ZZ_QSL
	zz_QSLimit = lim;
#else
	lim++;
	excptn ("setQSLimit illegal: smurph not created with '-q'");
#endif
}

void    Traffic::zz_wait (int ev, int pstate) {

/* ------------------------------------------------ */
/* The Client wait request (addressed to a Traffic) */
/* ------------------------------------------------ */

	if_from_observer ("Traffic->wait: called from an observer");
	assert (ev == INTERCEPT || ev == ARRIVAL,
		"Traffic->wait: illegal event");

	if (zz_c_first_wait) {

		zz_c_other = NULL;    // No requests so far

		// Create new simulation event
		zz_c_wait_event = new ZZ_EVENT;
		zz_c_wait_event -> station = TheStation;
		zz_c_wait_event -> process = TheProcess;

		new ZZ_REQUEST (&(TheStation->CWList), this, ev, pstate);

		assert (zz_c_other != NULL,
			"Traffic->wait: internal error -- null request");

		zz_c_wait_event -> pstate = pstate;
		zz_c_wait_event -> ai = this;
		zz_c_wait_event -> event_id = ev;
		zz_c_wait_event -> waketime = TIME_inf;
		zz_c_wait_event -> chain = zz_c_other;
		zz_c_wait_event -> Info01 = NULL;
		zz_c_wait_event -> Info02 = NULL;

		zz_c_whead = zz_c_other;
		zz_c_wait_event -> waketime =
			zz_c_wait_tmin = zz_c_other -> when = TIME_inf;
		zz_c_other -> event = zz_c_wait_event;
		zz_c_other -> other = zz_c_whead;

		zz_c_first_wait = NO;
		zz_c_wait_event->store ();

	} else {

		new ZZ_REQUEST (&(TheStation->CWList), this, ev, pstate);

		assert (zz_c_other != NULL,
			"Traffic->wait: internal error -- null request");

		zz_c_other -> when = TIME_inf;
		zz_c_other -> event = zz_c_wait_event;
		zz_c_other -> other = zz_c_whead;
	}
}

void    SGroup::setup (int n, int *sl) {

/* --------------------------- */
/* Initializes a station group */
/* --------------------------- */

	int     i, j, k;

	Assert ( n > -::NStations && ::NStations >= n,
		"SGroup: illegal number of stations");

	if (n == 0) {                   // Null group
		NStations = (int) ::NStations;
		SIdents = NULL;
		sorted = YES;
		return;
	}

	if (n < 0) {

		// Check if exceptions are legal station ids
		for (i = 0; i < -n; i++)
			if (!isStationId (sl [i]))
				excptn ("SGroup: illegal exception");

		// Check exception list against duplicates
		for (k = n, i = 0; i < -n; i++) {
			if (sl [i] == NONE) continue;
			for (j = i+1; j < -n; j++)
				if (sl [i] == sl [j]) k++;
		}

		NStations = (int) (::NStations + k);
	} else
		NStations = n;

	SIdents = new short [cp2m1 (NStations)];

	if (n < 0) {
		// The array contains list of exceptions
		n = -n;
		for (i = 0, k = 0; i < ::NStations; i++) {

			for (j = 0; j < n; j++)
				 if (sl [j] == i) break;

			if (j >= n) SIdents [k++] = i;
		}
	} else {
		// The array contains list of stations
		for (i = 0; i < NStations; i++) {
			Assert (isStationId (sl [i]),
				"SGroup: illegal station id");
			SIdents [i] = sl [i];
		}
	}
}

void    SGroup::setup (int n, int, ...) {

/* ------------------------------------------- */
/* As above, but with an explicit station list */
/* ------------------------------------------- */

	int     i;
	char    *ap;    // For processing variable-length argument list
	narray  (slist, int, n);

	Assert (n > 0, "SGroup: illegal station count");

	va_start (ap, n);   // Start variable-length argument processing

	for (i = 0; i < n; i++) slist [i] = va_arg (ap, int);
	va_end (ap);
	setup (n, slist);
	darray (slist);
}

void    SGroup::setup (int n, Station*, ...) {

/* -------------------------------------------------- */
/* As above, but with station pointers instead of ids */
/* -------------------------------------------------- */

	int     i;
	char    *ap;    // For processing variable-length argument list
	narray  (slist, int, n);

	Assert (n > 0, "SGroup: illegal station count");

	va_start (ap, n);   // Start variable-length argument processing

	for (i = 0; i < n; i++) {
		slist [i] = (int)((va_arg (ap, Station*))->Id);
	}
	va_end (ap);
	setup (n, slist);
	darray (slist);
}

static  short   *s_idents;      // Global variables to reduce the number of
static  float   *s_weights;     // parameters.

static void group_sorter (int lo, int up) {

/* -----------------------------------*/
/* The proper sorter for 'sort_group' */
/* -----------------------------------*/

int     i, j;
short   t;
float   s;

	// This is a simple variation of quicksort

	while (up > lo) {
		t = s_idents [i = (lo + up) / 2];
		s_idents [i] = s_idents [lo];
		s_idents [lo] = t;
		if (s_weights != NULL) {
			s = s_weights [i];
			s_weights [i] = s_weights [lo];
			s_weights [lo] = s;
		}
		i = lo;
		j = up;

		while (i < j) {
			while (s_idents [j] > t) j--;
			s_idents [i] = s_idents [j];
			if (s_weights != NULL)
				s_weights [i] = s_weights [j];
			while ((i < j) && (s_idents [i] <= t)) i ++;
			s_idents [j] = s_idents [i];
			if (s_weights != NULL)
				s_weights [j] = s_weights [i];
		}
		s_idents [i] = t;
		if (s_weights != NULL) s_weights [i] = s;
		group_sorter (lo, i-1);
		lo = i + 1;
	}
}

void zz_sort_group (SGroup *sg, float *w) {

/* ---------------------------------------------------------- */
/* Sort  station  id's  (and  weights)  in the non-decreasing */
/* order                                                      */
/* ---------------------------------------------------------- */

	int     n, i, j;
	short   l;

	if (sg->sorted) return;
	sg->sorted = YES;
	if ((s_idents = sg->SIdents) == NULL) return;   // Nothing to sort
	s_weights = w;
	n = sg->NStations;
	group_sorter (0, n-1);

	// Remove duplicates

	for (i = 1, j = 0, l = s_idents [0]; i < n; i++) {
		if (s_idents [i] != l) {
			s_idents [++j] = l = s_idents [i];
			if (s_weights != NULL) s_weights [j] = s_weights [i];
		} else {
			if (s_weights != NULL) s_weights [j] += s_weights [i];
		}
	}

	n = j + 1;

	if (s_weights != NULL) {
		for (i = 0, j = 0; i < n; i++) {
			// Remove stations with zero weight
			if (s_weights [i] > 0.0) {
				s_idents [j] = s_idents [i];
				s_weights [j++] = s_weights [i];
			}
		}
		n = j;
	}
	sg->NStations = n;
}

void    CGroup::setup (SGroup *snd, float *snw, SGroup *rcv, float *rcw) {

/* --------------------------------- */
/* Initializes a communication group */
/* --------------------------------- */

	Senders = snd;
	Receivers = rcv;
	CGWeight = 0.0;
	
	if (snw != NULL && Senders->NStations) {
		SWeights = new float [cp2m1 (Senders->NStations)];
		for (int i = 0; i < Senders->NStations; i++) {
			Assert (snw [i] >= 0.0,
				"CGroup: negative sender weight");
			CGWeight += (SWeights [i] = snw [i]);
		}
	} else {
			SWeights = NULL;
			CGWeight = 1.0;
	}

	// Assume all stations
	if (Senders->NStations == 0) Senders->NStations = (int) NStations;

	if (rcw != NULL && Receivers->NStations) {
		RWeights = new float [cp2m1 (Receivers->NStations)];
		for (int i = 0; i < Receivers->NStations; i++) {
			Assert (snw [i] >= 0.0,
				"CGroup: negative receiver weight");
			RWeights [i] = rcw [i];
		}
	} else
			RWeights = NULL;
};

void    CGroup::setup (SGroup *snd, float *snw, SGroup *rcv, int bdcst) {

/* ------------------------------------------------ */
/* Initializes a broadcast-type communication group */
/* ------------------------------------------------ */

	Senders = snd;
	Receivers = rcv;
	CGWeight = 0.0;
	
	if (snw != NULL && Senders->NStations) {
		SWeights = new float [cp2m1 (Senders->NStations)];
		for (int i = 0; i < Senders->NStations; i++) {
			Assert (snw [i] >= 0.0,
				"CGroup: negative sender weight");
			CGWeight += (SWeights [i] = snw [i]);
		}
	} else {
			SWeights = NULL;
			CGWeight = 1.0;
	}
	// Assume all stations
	if (Senders->NStations == 0) Senders->NStations = (int) NStations;

	Assert (bdcst == GT_broadcast, "CGroup: illegal argument");

	RWeights = (float*) (-1);       // Flag as broadcast type
};

void    Traffic::zz_addinit (int gr) {

/* -------------------------------------------- */
/* A preprocessor for addSender and addReceiver */
/* -------------------------------------------- */

	int     mgr, i;
	CGroup  **tcg, *cg;

	if ((mgr = gr+1) > NCGroups) {
		// A brand new group to be created
		if (CGCluster == NULL) {
			// Creating from scratch
			Assert (NCGroups == 0,
	  "Traffic->add...: internal error -- inconsistent initialization");
			CGCluster = new CGroup* [cp2m1 (mgr)];
		} else if (mgr > cp2m1 (NCGroups)) {
			// The group array must be extended
			tcg = new CGroup* [NCGroups];
			for (i = 0; i < NCGroups; i++) tcg [i] = CGCluster [i];
			delete (CGCluster);
			CGCluster = new CGroup* [cp2m1 (mgr)];
			for (i = 0; i < NCGroups; i++) CGCluster [i] = tcg [i];
			delete (tcg);
		}
		for (i = NCGroups; i < mgr; i++) CGCluster [i] = NULL;
		NCGroups = mgr;
	}

	if ((cg = CGCluster [gr]) == NULL) {
		// Must create the communication group
		cg = new CGroup;
		cg->Senders = cg->Receivers = NULL;
		cg->SWeights = cg->RWeights = NULL;
		cg->CGWeight = 0.0;
		CGCluster [gr] = cg;
	}
}

void    Traffic::addSender (Station *s, double w, int gr) {

/* ---------------------------------------------------------- */
/* Adds a sender to dynamically created/updated communication */
/* group                                                      */
/* ---------------------------------------------------------- */

	CGroup  *cg;
	SGroup  *sg;
	short   *sl;
	float   *fl;
	int     i;

	Assert (w != BROADCAST,
			  "Traffic->addSender: BROADCAST illegal for senders");
	Assert (w >= 0.0, "Traffic->addSender: negative weight");
	Assert (gr >=0, "Traffic->addSender: negative group");

	zz_addinit (gr);
	cg = CGCluster [gr];
	if ((sg = cg->Senders) == NULL) {
		sg = new SGroup;
		sg->NStations = 0;
		sg->SIdents = NULL;
		sg->sorted = NO;
		cg->Senders = sg;
	}

	if (s == ALL) {

		Assert (w == 1.0,
	"Traffic->addSender: weight must be 1.0 if ALL stations are included");
		// All stations
		Assert (sg->NStations == 0, form (
      "Traffic->addSender: senders of group %1d -- duplicate definition", gr));
		sg->NStations = (int) ::NStations;
		cg->CGWeight = 1.0;
		sg->sorted = YES;
		cg->SWeights = NULL;
		return;
	}

	Assert (sg->NStations != ::NStations || sg->SIdents != NULL, form (
      "Traffic->addSender: senders of group %1d -- cannot extend ALL", gr));

	// Add a new entry

	if (sg->SIdents == NULL) {
		// Creating from scratch
		Assert (sg->NStations == 0 && cg->CGWeight == 0.0 &&
			cg->SWeights == NULL,
	  "Traffic->addSender: internal error -- inconsistent initialization");
		sg->SIdents = new short [cp2m1 (1)];
		cg->SWeights = new float [cp2m1 (1)];
	} else if (sg->NStations + 1 > cp2m1 (sg->NStations)) {
		// The group array must be extended
		sl = new short [sg->NStations];
		for (i = 0; i < sg->NStations; i++) sl [i] = sg->SIdents [i];
		delete (sg->SIdents);
		sg->SIdents = new short [cp2m1 (sg->NStations + 1)];
		for (i = 0; i < sg->NStations; i++) sg->SIdents [i] = sl [i];
		delete (sl);

		fl = new float [sg->NStations];
		for (i = 0; i < sg->NStations; i++) fl [i] = cg->SWeights [i];
		delete (cg->SWeights);
		cg->SWeights = new float [cp2m1 (sg->NStations + 1)];
		for (i = 0; i < sg->NStations; i++) cg->SWeights [i] = fl [i];
		delete (fl);
	}

	sg->SIdents [sg->NStations] = (short) ident (s);
	cg->SWeights [sg->NStations] = w;
	sg->NStations ++;
	cg->CGWeight += w;
	sg->sorted = NO;
}

void    Traffic::addSender (long s, double w, int gr) {

/* ---------------------------------------------------------- */
/* Another version of the above -- with station id instead of */
/* pointer                                                    */
/* ---------------------------------------------------------- */

	Assert (isStationId (s), form (
     "Traffic->addSender: group %1d -- %1d is not a legal station id", gr, s));

	addSender (idToStation (s), w, gr);
}

void    Traffic::addReceiver (Station *s, double w, int gr) {

/* ---------------------------------------------------------- */
/* Adds    a    receiver   to   dynamically   created/updated */
/* communication group                                        */
/* ---------------------------------------------------------- */

	CGroup  *cg;
	SGroup  *sg;
	short   *sl;
	float   *fl;
	int     i;

	Assert (w >= 0.0 || w == BROADCAST,
				      "Traffic->addReceiver: negative weight");
	Assert (gr >=0, "Traffic->addReceiver: negative group");

	zz_addinit (gr);
	cg = CGCluster [gr];
	if ((sg = cg->Receivers) == NULL) {
		sg = new SGroup;
		sg->NStations = 0;
		sg->SIdents = NULL;
		sg->sorted = NO;
		cg->Receivers = sg;
	}

	if (s == ALL) {

		Assert (w == 1.0 || w == BROADCAST,
	      "Traffic->addSender: with ALL, weight must be 1.0 or BROADCAST");
		// All stations
		Assert (sg->NStations == 0, form (
    "Traffic->addSender: receivers of group %1d -- duplicate definition", gr));
		sg->NStations = (int) ::NStations;
		sg->sorted = YES;
		cg->RWeights = (w == BROADCAST) ? (float*) (-1) : NULL;
		return;
	}

	Assert (sg->NStations != ::NStations || sg->SIdents != NULL, form (
     "Traffic->addReceiver: receivers of group %1d -- cannot extend ALL", gr));

	// Add a new entry

	if (sg->SIdents == NULL) {
		// Creating from scratch
		Assert (sg->NStations == 0 && cg->RWeights == NULL,
	"Traffic->addReceiver: internal error -- inconsistent initialization");
		sg->SIdents = new short [cp2m1 (1)];
		cg->RWeights = (w == BROADCAST) ? (float*) (-1) :
			new float [cp2m1 (1)];
	} else if (sg->NStations + 1 > cp2m1 (sg->NStations)) {
		// The group array must be extended
		sl = new short [sg->NStations];
		for (i = 0; i < sg->NStations; i++) sl [i] = sg->SIdents [i];
		delete (sg->SIdents);
		sg->SIdents = new short [cp2m1 (sg->NStations + 1)];
		for (i = 0; i < sg->NStations; i++) sg->SIdents [i] = sl [i];
		delete (sl);

		if (cg->RWeights != (float*) (-1)) {
			fl = new float [sg->NStations];
			for (i = 0; i < sg->NStations; i++)
				fl [i] = cg->RWeights [i];
			delete (cg->RWeights);
			cg->RWeights = new float [cp2m1 (sg->NStations + 1)];
			for (i = 0; i < sg->NStations; i++)
				cg->RWeights [i] = fl [i];
			delete (fl);
		}
	}

	sg->SIdents [sg->NStations] = (short) ident (s);
	if (w == BROADCAST) {
		Assert (cg->RWeights == (float*) (-1), form (
"Traffic->addReceiver: group %1d nonempty -- cannot change to BROADCAST", gr));
	} else {
		Assert (cg->RWeights != (float*) (-1), form (
"Traffic->addReceiver: group %1d nonempty -- cannot change to non-BROADCAST",
			gr));
		cg->RWeights [sg->NStations] = w;
	}
	sg->NStations ++;
	sg->sorted = NO;
}

void    Traffic::addReceiver (long s, double w, int gr) {

/* ---------------------------------------------------------- */
/* Another version of the above -- with station id instead of */
/* pointer                                                    */
/* ---------------------------------------------------------- */

	Assert (isStationId (s), form (
   "Traffic->addReceiver: group %1d -- %1d is not a legal station id", gr, s));

	addReceiver (idToStation (s), w, gr);
}

void    Traffic::zz_start () {

/* --------------------------- */
/* The nonstandard constructor */
/* --------------------------- */

	static  int     asize = 15;     // A small power of two - 1 (initial
					// size of zz_tp)
	Traffic         **scratch;

	Assert (!zz_flg_started,
	    "Traffic: cannot create Traffics after the protocol has started");

	ChList = NULL;
	last_sender = NONE;
	last_CGroup = NULL;
#if     ZZ_QSL
	QSLimit = MAX_long;
	QSize = 0;
#endif

	// Obtain the pointer to the virtual function table of the packet
	zz_mkp ();
	ptviptr = zz_setptr (ThePacket);
	delete (ThePacket);

	if ((Id = sernum++) == 0) {
		// The first time around -- create the array of Traffics
		zz_tp = new Traffic* [asize];
	} else if (Id >= asize) {
		// The array must be enlarged
		scratch = new Traffic* [asize];
		for (int i = 0; i < asize; i++)
			// Backup copy
			scratch [i] = zz_tp [i];

		delete (zz_tp);         // Deallocate previous array
		zz_tp = new Traffic* [asize = (asize+1) * 2 - 1];
		while (i--) zz_tp [i] = scratch [i];
		delete (scratch);
	}

	zz_tp [Id] = this;
	NTraffics = sernum;

	// Setup for class
	Class = AIC_traffic;

	// Add the object to the owner (at this moment it is Kernel)
	zz_queue_head (this, TheProcess->ChList, ZZ_Object);
	// Notify the server about a new displayable object.
	zz_DADD (TheProcess, this);
}

void    Traffic::setup (CGroup **cgl, int ncgl, int flags, ...) {

/* ---------------------------- */
/* Initialize a traffic pattern */
/* ---------------------------- */

	char    *ap;    // For processing variable-length argument list

	Assert (ncgl >= 0, "Traffic: illegal group count");

	if (ncgl) {
		// The cluster is nonempty
		CGCluster = new CGroup* [ncgl];
		for (int i = 0; i < ncgl; i++) CGCluster [i] = cgl [i];
	} else
		CGCluster = NULL;

	NCGroups = ncgl;

	// Setup for flags

	DstMIT = DstMLE = DstBIT = DstBSI = UNDEFINED;
	FlgSCL = FlgSPF = ON;

	if (flags & MIT_exp) {
		Assert (! (flags & MIT_unf), "Traffic: illegal options");
		DstMIT = EXPONENTIAL;
	}
	
	if (flags & MIT_unf) {
		Assert (! (flags & MIT_exp), "Traffic: illegal options");
		DstMIT = UNIFORM;
	}
	
	if (flags & MLE_exp) {
		Assert (! (flags & MLE_unf), "Traffic: illegal options");
		DstMLE = EXPONENTIAL;
	}
	
	if (flags & MLE_unf) {
		Assert (! (flags & MLE_exp), "Traffic: illegal options");
		DstMLE = UNIFORM;
	}
	
	if (flags & BIT_exp) {
		Assert (! (flags & BIT_unf), "Traffic: illegal options");
		DstBIT = EXPONENTIAL;
	}
	
	if (flags & BIT_unf) {
		Assert (! (flags & BIT_exp), "Traffic: illegal options");
		DstBIT = UNIFORM;
	}
	
	if (flags & BSI_exp) {
		Assert (! (flags & BSI_unf), "Traffic: illegal options");
		DstBSI = EXPONENTIAL;
	}
	
	if (flags & BSI_unf) {
		Assert (! (flags & BSI_exp), "Traffic: illegal options");
		DstBSI = UNIFORM;
	}

	if (flags & SCL_on) {
		Assert (! (flags & SCL_off), "Traffic: illegal options");
		FlgSCL = ON;
	}

	if (flags & SCL_off) {
		Assert (! (flags & SCL_on), "Traffic: illegal options");
		FlgSCL = OFF;
	}

	if (flags & SPF_on) {
		Assert (! (flags & SPF_off), "Traffic: illegal options");
		FlgSPF = ON;
	}

	if (flags & SPF_off) {
		Assert (! (flags & SPF_on), "Traffic: illegal options");
		FlgSPF = OFF;
	}

	va_start (ap, flags);   // Start variable-length argument processing

	if (DstMIT != UNDEFINED) {
		ParMnMIT = va_arg (ap, double) * Etu;
		if (DstMIT == UNIFORM) {
			ParMxMIT = va_arg (ap, double) * Etu;
		} else {
			ParMxMIT = ParMnMIT;
		}
	}

	if (DstMLE != UNDEFINED) {
		ParMnMLE = va_arg (ap, double);
		if (DstMLE == UNIFORM) {
			ParMxMLE = va_arg (ap, double);
		} else {
			ParMxMLE = ParMnMLE;
		}
	}

	if (DstBIT != UNDEFINED) {
		ParMnBIT = va_arg (ap, double) * Etu;
		if (DstBIT == UNIFORM) {
			ParMxBIT = va_arg (ap, double) * Etu;
		} else {
			ParMxBIT = ParMnBIT;
		}
	}

	if (DstBSI != UNDEFINED) {
		ParMnBSI = va_arg (ap, double);
		if (DstBSI == UNIFORM) {
			ParMxBSI = va_arg (ap, double);
		} else {
			ParMxBSI = ParMnBSI;
		}
	}

	va_end (ap);
}

void    Traffic::setup (int flags, ...) {

/* ---------------------------------------------------------- */
/* Initialize  a  traffic pattern without groups. They can be */
/* later defined dynamically.                                 */
/* ---------------------------------------------------------- */

	char    *ap;    // For processing variable-length argument list

	// Initialize group list
	CGCluster = NULL;
	NCGroups = 0;

	// Setup for flags
	DstMIT = DstMLE = DstBIT = DstBSI = UNDEFINED;
	FlgSCL = FlgSPF = ON;

	if (flags & MIT_exp) {
		Assert (! (flags & MIT_unf), "Traffic: illegal options");
		DstMIT = EXPONENTIAL;
	}
	
	if (flags & MIT_unf) {
		Assert (! (flags & MIT_exp), "Traffic: illegal options");
		DstMIT = UNIFORM;
	}
	
	if (flags & MLE_exp) {
		Assert (! (flags & MLE_unf), "Traffic: illegal options");
		DstMLE = EXPONENTIAL;
	}
	
	if (flags & MLE_unf) {
		Assert (! (flags & MLE_exp), "Traffic: illegal options");
		DstMLE = UNIFORM;
	}
	
	if (flags & BIT_exp) {
		Assert (! (flags & BIT_unf), "Traffic: illegal options");
		DstBIT = EXPONENTIAL;
	}
	
	if (flags & BIT_unf) {
		Assert (! (flags & BIT_exp), "Traffic: illegal options");
		DstBIT = UNIFORM;
	}
	
	if (flags & BSI_exp) {
		Assert (! (flags & BSI_unf), "Traffic: illegal options");
		DstBSI = EXPONENTIAL;
	}
	
	if (flags & BSI_unf) {
		Assert (! (flags & BSI_exp), "Traffic: illegal options");
		DstBSI = UNIFORM;
	}
	if (flags & SCL_on) {
		Assert (! (flags & SCL_off), "Traffic: illegal options");
		FlgSCL = ON;
	}

	if (flags & SCL_off) {
		Assert (! (flags & SCL_on), "Traffic: illegal options");
		FlgSCL = OFF;
	}

	if (flags & SPF_on) {
		Assert (! (flags & SPF_off), "Traffic: illegal options");
		FlgSPF = ON;
	}

	if (flags & SPF_off) {
		Assert (! (flags & SPF_on), "Traffic: illegal options");
		FlgSPF = OFF;
	}

	va_start (ap, flags);   // Start variable-length argument processing

	if (DstMIT != UNDEFINED) {
		ParMnMIT = va_arg (ap, double) * Etu;
		if (DstMIT == UNIFORM) {
			ParMxMIT = va_arg (ap, double) * Etu;
		} else {
			ParMxMIT = ParMnMIT;
		}
	}

	if (DstMLE != UNDEFINED) {
		ParMnMLE = va_arg (ap, double);
		if (DstMLE == UNIFORM) {
			ParMxMLE = va_arg (ap, double);
		} else {
			ParMxMLE = ParMnMLE;
		}
	}

	if (DstBIT != UNDEFINED) {
		ParMnBIT = va_arg (ap, double) * Etu;
		if (DstBIT == UNIFORM) {
			ParMxBIT = va_arg (ap, double) * Etu;
		} else {
			ParMxBIT = ParMnBIT;
		}
	}

	if (DstBSI != UNDEFINED) {
		ParMnBSI = va_arg (ap, double);
		if (DstBSI == UNIFORM) {
			ParMxBSI = va_arg (ap, double);
		} else {
			ParMxBSI = ParMnBSI;
		}
	}

	va_end (ap);
}

void    Traffic::setup (CGroup *cg, int flags, ...) {

/* ---------------------------------------------------- */
/* Initialize a traffic pattern (single CGroup version) */
/* ---------------------------------------------------- */

	char    *ap;    // For processing variable-length argument list

	CGCluster = new CGroup* [1];
	CGCluster [0] = cg;
	NCGroups = 1;

	// Setup for flags

	DstMIT = DstMLE = DstBIT = DstBSI = UNDEFINED;
	FlgSCL = FlgSPF = ON;

	if (flags & MIT_exp) {
		Assert (! (flags & MIT_unf), "Traffic: illegal options");
		DstMIT = EXPONENTIAL;
	}
	
	if (flags & MIT_unf) {
		Assert (! (flags & MIT_exp), "Traffic: illegal options");
		DstMIT = UNIFORM;
	}
	
	if (flags & MLE_exp) {
		Assert (! (flags & MLE_unf), "Traffic: illegal options");
		DstMLE = EXPONENTIAL;
	}
	
	if (flags & MLE_unf) {
		Assert (! (flags & MLE_exp), "Traffic: illegal options");
		DstMLE = UNIFORM;
	}
	
	if (flags & BIT_exp) {
		Assert (! (flags & BIT_unf), "Traffic: illegal options");
		DstBIT = EXPONENTIAL;
	}
	
	if (flags & BIT_unf) {
		Assert (! (flags & BIT_exp), "Traffic: illegal options");
		DstBIT = UNIFORM;
	}
	
	if (flags & BSI_exp) {
		Assert (! (flags & BSI_unf), "Traffic: illegal options");
		DstBSI = EXPONENTIAL;
	}
	
	if (flags & BSI_unf) {
		Assert (! (flags & BSI_exp), "Traffic: illegal options");
		DstBSI = UNIFORM;
	}

	if (flags & SCL_on) {
		Assert (! (flags & SCL_off), "Traffic: illegal options");
		FlgSCL = ON;
	}

	if (flags & SCL_off) {
		Assert (! (flags & SCL_on), "Traffic: illegal options");
		FlgSCL = OFF;
	}

	if (flags & SPF_on) {
		Assert (! (flags & SPF_off), "Traffic: illegal options");
		FlgSPF = ON;
	}

	if (flags & SPF_off) {
		Assert (! (flags & SPF_on), "Traffic: illegal options");
		FlgSPF = OFF;
	}

	va_start (ap, flags);   // Start variable-length argument processing
	if (DstMIT != UNDEFINED) {
		ParMnMIT = va_arg (ap, double) * Etu;
		if (DstMIT == UNIFORM) {
			ParMxMIT = va_arg (ap, double) * Etu;
		} else {
			ParMxMIT = ParMnMIT;
		}
	}

	if (DstMLE != UNDEFINED) {
		ParMnMLE = va_arg (ap, double);
		if (DstMLE == UNIFORM) {
			ParMxMLE = va_arg (ap, double);
		} else {
			ParMxMLE = ParMnMLE;
		}
	}

	if (DstBIT != UNDEFINED) {
		ParMnBIT = va_arg (ap, double) * Etu;
		if (DstBIT == UNIFORM) {
			ParMxBIT = va_arg (ap, double) * Etu;
		} else {
			ParMxBIT = ParMnBIT;
		}
	}

	if (DstBSI != UNDEFINED) {
		ParMnBSI = va_arg (ap, double);
		if (DstBSI == UNIFORM) {
			ParMxBSI = va_arg (ap, double);
		} else {
			ParMxBSI = ParMnBSI;
		}
	}

	va_end (ap);
}

void    Traffic::preprocess_weights () {

/* ---------------------------------------------------------- */
/* Goes  through  the communication groups and calculates new */
/* weights, so that interpolation can  be  used  to  generate */
/* senders and receivers                                      */
/* ---------------------------------------------------------- */

	CGroup  *cg;
	SGroup  *sg;
	int     k, j;
	double  w;

	TCGWeight = 0.0;
	for (j = 0; j < NCGroups; j++) {
		cg = CGCluster [j];
		if (cg->SWeights == NULL) {
			// Equal weights of 1/n
			TCGWeight += 1.0;
			cg->CGWeight = TCGWeight;
		} else {
			sg = cg->Senders;
			for (cg->CGWeight = 0.0, k = 0; k < sg->NStations; k++)
				// Convert weights to density
				cg->SWeights [k] = (cg->CGWeight +=
					cg->SWeights [k]);
			cg->CGWeight = (TCGWeight += cg->CGWeight);
		}

		// Take care of receivers
		if ((int) (cg->RWeights) == -1 || cg->RWeights == NULL)
			continue;
		sg = cg->Receivers;
		for (w = 0.0, k = 0; k < sg->NStations; k++)
			// Convert weights to density
			cg->RWeights [k] = (w += cg->RWeights [k]);
	}
}

TIME    Traffic::genMIT () {

/* ----------------------------------- */
/* Generates message interarrival time */
/* ----------------------------------- */

	assert (DstMIT != UNDEFINED, "genMIT: MIT distribution is undefined");
		if (DstMIT == EXPONENTIAL)
			return (tRndPoisson (ParMnMIT));
		else
			return (tRndUniform (ParMnMIT, ParMxMIT));
}

TIME    Traffic::genBIT () {

/* --------------------------------- */
/* Generates burst interarrival time */
/* --------------------------------- */

	assert (DstBIT != UNDEFINED, "genBIT: BIT distribution is undefined");
		if (DstBIT == EXPONENTIAL)
			return (tRndPoisson (ParMnBIT));
		else
			return (tRndUniform (ParMnBIT, ParMxBIT));
}

long    Traffic::genMLE () {

/* ------------------------ */
/* Generates message length */
/* ------------------------ */

	long    ml;

	assert (DstMLE != UNDEFINED, "genMLE: MLE distribution is undefined");
	if (DstMLE == EXPONENTIAL)
		ml = lRndPoisson (ParMnMLE);
	else
		ml = lRndUniform (ParMnMLE, ParMxMLE);
	if (ml < 8)
		ml = 8;                 // At least one byte
	else
		ml = (ml + 7) & ~07;    // Round up to bytes
	return (ml);
}

long    Traffic::genBSI () {

/* -------------------- */
/* Generates burst size */
/* -------------------- */

	assert (DstBSI != UNDEFINED, "genBSI: BSI distribution is undefined");
		if (DstBSI == EXPONENTIAL)
			return (lRndPoisson (ParMnBSI));
		else
			return (lRndUniform (ParMnBSI, ParMxBSI));
}

static  select_station (int ns, short *idents, float *weights, double w) {

/* ---------------------------------------------------------- */
/* Selects  one  station  from  the  group  according  to the */
/* weights                                                    */
/* ---------------------------------------------------------- */

	int             j, es, ef;
	double          l, u;

	if (weights == NULL) {
		// All stations have equal weights of 1/n
		if (ns == 0) ns = (int) NStations;
		assert ((w < 1.01) && (w > -0.01), "genXXX: bad weight");
		if ((j = (int) (ns * w)) >= ns) j = ns - 1; else
			if (j < 0) j = 0;
		return ((idents == NULL) ? j : (int)(idents [j]));
	}

	if (ns < 5) {
		// Do it "manually", if the number of stations is not too big
		for (j = 0; j < ns; j++) {
			if (weights [j] >= w)
				return ((int) (idents [j]));
		}
		return ((int) (idents [j-1]));
	} else {
		// Interpolate
		l = weights [es = 0];
		u = weights [ef = ns - 1];

		while (YES) {

			// Interpolation loop
			if ((j = (int)(((w - l) / (u - l)) * (ef - es) +
				es + 1)) < es)
					j = es;
			else if (j > ef) j = ef;

			if (weights [j] < w) {
				// Search on the right
				if ((es = j + 1) >= ef) {
					j = ef;
					break;
				}
				l = weights [es];
			} else if ((j <= es) || (weights [j - 1] < w)) {
				break;
			} else {
				// Search on the left
				if ((ef = j - 1) <= es) {
					j = es;
					break;
				}
				u = weights [ef];
			}
		}
	}

	return ((int) (idents [j]));
}

int     Traffic::genSND () {

/* --------------- */
/* Generate sender */
/* --------------- */

	CGroup  *cg;
	int     j, es, ef;
	double  w, l, u;

	if (NCGroups == 0) return (last_sender = NONE); // Nothing to do

	w = rnd (SEED_traffic) * TCGWeight;

	// Determine the group; do it "manually", if the number of groups
	// is not too big

	if (NCGroups < 5) {
		for (j = 0, cg = CGCluster [0]; j < NCGroups; j++)
			if ((cg = CGCluster [j]) -> CGWeight >= w) goto FI;
			// Please, don't tell me that goto's are not nice.
			// I've been told this nonsense long enough.
		cg = CGCluster [--j];
FI: ;
	} else {
		// Too many groups -- interpolate
		l = CGCluster[0]->CGWeight;
		u = TCGWeight;                  // Bounds
		es = 0;
		ef = NCGroups - 1;
		while (YES) {
			// Interpolation loop
			j = (int)(((w - l) / (u - l)) * (ef - es) + es + 1);
			if (j > ef) j = ef; else if (j < es) j = es;
			if ((cg = CGCluster [j]) -> CGWeight < w) {
				// Search on the right
				if ((es = j+1) >= ef) {
					cg = CGCluster [j = ef];
					break;
				}
				l = CGCluster [es] -> CGWeight;
			} else if ((j <= es) || (CGCluster [j-1] -> CGWeight <
				w)) {
					break;
			} else {
				// Search on the left
				if ((ef = j-1) <= es) {
					cg = CGCluster [j = es];
					break;
				}
				u = CGCluster [ef] -> CGWeight;
			}
		}
	}

	if (j) w -= CGCluster [j-1] -> CGWeight;

	// Set last_sender and last_CGroup to be used by genRCV. genRCV
	// must select the receiver from the same group and the receiver must
	// be different from the sender

	last_sender = select_station ((cg->Senders)->NStations,
		(cg->Senders)->SIdents, cg->SWeights, w);
	last_CGroup = cg;
	return (last_sender);
}

static  int     sender_index;       // Set by 'within_group'

CGroup *Traffic::genCGR (long sender) {

/* ---------------------------------------------------------- */
/* Generate  a  selection  group  containing the given sender */
/* (according  to  the  probability  of  the  sender's  being */
/* selected from that group)                                  */
/* ---------------------------------------------------------- */

	double  sp [EETSIZE];
	CGroup  *cg;
	int     j, ng;
	double  tsw, w, l;

	if (NCGroups == 0) {
		last_sender = NONE;
		last_CGroup = NULL;
		return (NULL);
	}

	// Determine the probability of the sender in each selection group
	for (j = 0, tsw = 0.0, ng = NCGroups; j < ng; j++) {
		cg = CGCluster [j];

		if ((cg->Senders)->occurs (sender)) {
			if (sender_index == NONE)       // All stations
				l = 1.0 / (double) NStations;
			else if (cg->SWeights == NULL)
				l = ((cg->Senders)->NStations) ?
					1.0/((cg->Senders)->NStations) :
						0;
			
			else if (sender_index)
				l = cg->SWeights [sender_index] -
					cg->SWeights [sender_index-1];
			else
				l = cg->SWeights [0];
		} else
			l = 0.0;

		sp [j] = (tsw += l);
	}

	w = rnd (SEED_traffic) * tsw;

	// Determine the group

	for (j = 0; j < ng; j++)
		if (sp [j] >= w) {
			last_sender = sender;
			last_CGroup = CGCluster [j];
			return (last_CGroup);
		}
}

long    Traffic::genRCV () {

/* ------------------- */
/* Generate a receiver */
/* ------------------- */

	SGroup  *receivers;
	int     ns, i, nr, f, l;
	double  w, wbs, was, wg;

	if (last_sender == NONE) {
		// The sender was not generated, so generate it first
		genSND ();
		if (last_sender == NONE) return (NONE);
	}

	if (((receivers = last_CGroup -> Receivers) == NULL) ||
		((nr = receivers -> NStations) == 0)) {
		// No receiver
		return (NONE);

	}

	if ((long) (last_CGroup->RWeights) == -1) {
		// A broadcast group -- return the whole group
		return ((long) receivers);
	}

	// Select the receiver at random

	wg = (last_CGroup->RWeights == NULL) ? 1.0 :
		last_CGroup->RWeights [ nr - 1 ];
RCV_RETRY:
	for (i = 0; i < 10; i++) {
		// Give it 10 tries
		w = rnd (SEED_traffic) * wg;
		if ((ns = select_station (receivers->NStations,
			receivers->SIdents, last_CGroup->RWeights, w)) !=
				last_sender)
					return (ns);
	}

	// We have tried it 10 times and the receiver was always
	// the same as the sender. Now let us take a closer look
	// at what is going on.

	if (nr <= 1)
	    excptn (form ("genRCV: cannot generate receiver for Traffic %s",
		getSName ()));

	// All stations - keep trying
	if (last_CGroup->RWeights == NULL) goto RCV_RETRY;

	// Eliminate the sender from the receiver list

	f = 0; l = nr-1;

	while (1) {

		// Find the sender. Note that it occurs exactly once
		// as all duplicates were removed when the group was
		// read from the input.

		assert (f <= l, "genRCV: sender not found");

		i = (f + l) / 2;

		// We use bisection here. Note that the receivers are
		// sorted according to their id's.

		if (receivers -> SIdents [i] < last_sender) {
			f = i + 1;
		} else if (receivers -> SIdents [i] > last_sender) {
			l = i - 1;
		} else
			break;
	}

	// Total weight of receivers preceding the sender
	wbs = i ? last_CGroup -> RWeights [i - 1] : 0.0;

	// Total weight of receivers following the sender
	was = wg - last_CGroup -> RWeights [i];

	w = rnd (SEED_traffic) * (wbs + was);

	if ((w <= wbs) && i) {
		// Search on the left
		ns = select_station (i, receivers->SIdents,
			last_CGroup->RWeights, w);
	} else {
		// Search on the right
		ns = select_station (nr - i - 1,
			&(receivers->SIdents [i+1]),
				&(last_CGroup->RWeights [i+1]),
					w + last_CGroup->RWeights [i]);
	}
	return (ns);
}

Message  *Traffic::genMSG (long sender, long receiver, long length) {

/* ---------------------------------------------------------- */
/* Generates  a  message  with  the  specified parameters and */
/* queues it at the end of the corresponding sender's  queue. */
/* A pointer to the message is returned as the function value */
/* and via 'TheMessage' (Info01).                             */
/* ---------------------------------------------------------- */

	Message            *m;
	Station            *s;
	ZZ_REQUEST         *cw;
	ZZ_EVENT           *e;
	int                i;

	assert (length >= 0, "genMSG: negative length");

	s = idToStation (sender);
#if     ZZ_QSL
	// Check if any of the limits has been reached
	if (FlgSPF == ON) {
	    if (zz_NQMessages >= zz_QSLimit || QSize >= QSLimit || s->QSize >=
		s->QSLimit) {
		// Ignore this generation
		return ((Message*)(Info01 = NULL));
	    }
	    // These rightfully belong to stdpfmMQU (see stdpfmMTR), but
	    // stdpfmMQU knows nothing about the sender 's'
	    QSize++;
	    s->QSize++;
	}
#endif
	m = zz_mkm ();                          // Make the message

	m -> Receiver = receiver;
	m -> QTime = Time;
	m -> TP = (int) Id;
	m -> Length = length;

	// Put the message into queue

	if (s -> MQHead == NULL) {
		// The message queue does not exist yet -- create it
		s->MQHead = new Message* [NTraffics];
		s->MQTail = new Message* [NTraffics];
		for (i = 0; i < NTraffics; i++)
			s->MQHead [i] = s->MQTail [i] = NULL;
	}

	if (s->MQHead [Id] == NULL) {
		assert ((s->MQTail [Id] == NULL),
				"genMSG: message queue screwup");
		s->MQHead [Id] = s->MQTail [Id] = m;
		m->prev = (Message*) (&(s->MQHead [Id]));
	} else {
		assert (s->MQTail [Id] != NULL,
			"genMSG: message queue screwup");
		(s->MQTail [Id])->next = m;
		m->prev = s->MQTail [Id];
		s->MQTail [Id] = m;
	}

	m -> next = NULL;       // Make sure the queue is terminated properly

	// Restart processes waiting for message arrival

	for (cw = s -> CWList; cw != NULL; cw = cw -> next) {
		if (cw -> ai == Client) {
			if (cw->event_id == ARRIVAL ||
				cw -> event_id == Id) {

				cw -> when = Time;
				if ((e = cw-> event) -> waketime > Time ||
					flip ()) {
						cw -> Info01 = (void*) m;
						cw -> Info02 = (void*) Id;
						e->new_top_request (cw);
				}

			} else if (cw -> event_id == INTERCEPT) {

				cw -> when = Time;
				cw -> Info01 = (void*) m;
				cw -> Info02 = (void*) Id;

				assert (zz_pe == NULL,
			"genMSG: more than one wait for INTERCEPT");

				zz_pe = (zz_pr = cw) -> event;
			}
		} else {
			if (((Traffic*)(cw->ai))->Id != Id) continue;
			if (cw -> event_id == INTERCEPT) {

				cw -> when = Time;
				cw -> Info01 = (void*) m;
				cw -> Info02 = (void*) Id;

				assert (zz_pe == NULL,
			"genMSG: more than one wait for INTERCEPT");

				zz_pe = (zz_pr = cw) -> event;
			} else {
				cw -> when = Time;
				if ((e = cw-> event) -> waketime > Time ||
					flip ()) {
						cw -> Info01 = (void*) m;
						cw -> Info02 = (void*) Id;
						e->new_top_request (cw);
				}
			}
		}
	}

	m->setup ();    // User-defined part
	spfmMQU (m);    // Performance
	return (m);
}

int     SGroup::occurs (long sid) {

/* ----------------------------------------------- */
/* Determine if station id sid occurs in the group */
/* ----------------------------------------------- */

	int     nf, ns;
	short   *idents;

	if (! sorted) zz_sort_group (this);


	if (SIdents == NULL) {
		// All stations belong to this group
		assert (NStations == ::NStations,
			"SGroup: 'occurs (...)' inconsistent station group");
		sender_index = NONE;
		return (YES);
	}

	if ((ns = NStations) < 5) {
		// Search sequentially
		for (idents = SIdents; ns--; )
			if (*idents++ == sid) return (YES);
		return (NO);
	}

	// Use bisection
	nf = 0; ns--;

	while (1) {

		// Find the current station id. Note that it occurs
		// at most once in the idents table.

		if (nf > ns) return (NO);

		sender_index = (nf + ns) / 2;

		// sender_index is defined as static above generate_group

		if (SIdents [sender_index] < sid) {
			nf = sender_index + 1;
		} else if (SIdents [sender_index] > sid) {
			ns = sender_index - 1;
		} else
			return (YES);
	}
}

void    Traffic::stdpfmMQU (Message *m) {

/* ---------------------------------------------------------- */
/* Updates  standard  performance measures after a message is */
/* queued at the sender                                       */
/* ---------------------------------------------------------- */

	assert (RVMLS != NULL,
		"Traffic->stdpfmMQU: standard Client not started");
	RVMLS->update (m->Length, 1);
	// Update counters
	NQMessages++;
	zz_NQMessages++;
	zz_NGMessages++;
	NQBits += m->Length;
	zz_NQBits += m->Length;
}

void    Traffic::stdpfmMTR (Packet *p) {

/* ---------------------------------------------------------- */
/* Updates  standard  performance  measures  after  the  last */
/* packet of a message is transmitted                         */
/* ---------------------------------------------------------- */

	assert (RVMAT != NULL,
		"Traffic->stdpfmMTR: standard Client not started");
	RVMAT->update (Itu * ((double) (Time - p->QTime)), 1);
	NQMessages--;
	zz_NQMessages--;
	NTMessages++;
	zz_NTMessages++;
#if     ZZ_QSL
	QSize--;
	(zz_st [p->Sender]->QSize)--;
#endif
}

void    Traffic::stdpfmPTR (Packet *p) {

/* ---------------------------------------------------------- */
/* Updates  standard  performance  measures after a packet is */
/* transmitted                                                */
/* ---------------------------------------------------------- */

	assert (RVPAT != NULL,
		"Traffic->stdpfmPTR: standard Client not started");
	RVPAT->update (Itu * ((double) (Time - p->TTime)), 1);
	NQBits -= p->ILength;
	zz_NQBits -= p->ILength;
	NTBits += p->ILength;
	zz_NTBits += p->ILength;
	NTPackets++;
	zz_NTPackets++;
}

void    Traffic::stdpfmPRC (Packet *p) {

/* ---------------------------------------------------------- */
/* Updates  standard  performance  measures after a packet is */
/* received                                                   */
/* ---------------------------------------------------------- */

	assert (RVWMD != NULL,
		"Traffic->stdpfmPRC: standard Client not started");
	if (p->ILength > 0)
		// Weighted message delay
		RVWMD->update (Itu * ((double) (Time - p->QTime)), p->ILength);

	// Absolute packet delay
	assert (RVAPD != NULL,
		"Traffic->stdpfmPRC: standard Client not started");
	RVAPD->update (Itu * ((double) (Time - p->TTime)), 1);
	NRBits += p->ILength;
	zz_NRBits += p->ILength;
	NRPackets++;
	zz_NRPackets++;
}

void    Traffic::stdpfmMRC (Packet *p) {

/* ---------------------------------------------------------- */
/* Updates  standard  performance measures after a message is */
/* received                                                   */
/* ---------------------------------------------------------- */

	assert (RVAMD != NULL,
		"Traffic->stdpfmMRC: standard Client not started");
	RVAMD->update (Itu * ((double) (Time - p->QTime)), 1);
	NRMessages++;
	zz_NRMessages++;
}

Packet  *Traffic::zz_mkp () {

/* --------------------------------------- */
/* Creates a packet (used for tricks only) */
/* --------------------------------------- */

	Info01 = (void*) new Packet;    // ThePacket
	zz_mkpclean ();
	return (ThePacket);
}

void    zz_mkpclean () {

/* ---------------------------------------------------------- */
/* Cleans  up  after  zz_mkp.  The Packet constructor assumes */
/* that the packet is  a  statically  declared  buffer.  This */
/* function removes the buffer description from the temporary */
/* list.                                                      */
/* ---------------------------------------------------------- */

	assert (zz_ncpframe != NULL && zz_ncpframe->next == NULL,
		"Traffic::mkpclean: packet buffer declared outside station");

	delete (zz_ncpframe);
	zz_ncpframe = NULL;
}

Packet::Packet () {

/* --------------------------------------------------- */
/* Constructor for creating packet buffers at stations */
/* --------------------------------------------------- */

	new ZZ_PFItem (this);
}

ZZ_PFItem::ZZ_PFItem (Packet *p) {

/* ---------------------------------------------------------- */
/* Constructor  for  a  packet  buffer  description: puts the */
/* description into a temporary list                          */
/* ---------------------------------------------------------- */

	(buf = p)->Flags = 0L;
	p->QTime = p->TTime = TIME_0;
	p->Sender = p->Receiver = NONE;
	p->ILength = p->TLength = 0;
	p->TP = NONE;
	next = NULL;
	if (zz_ncpframe == NULL)
		zz_ncpframe = this;
	else {

		for (ZZ_PFItem *it = zz_ncpframe; it->next != NULL;
			it = it->next);
		it->next = this;
	}
}

#if     ZZ_ASR
void    Packet::setup (Message *m) {
#else
void    Packet::setup (Message*) {
#endif

/* ---------------------------------------------------------- */
/* A  virtual  function  to  be  redefined  in a user-defined */
/* Packet subtype. Called when a packet  is  created  from  a */
/* message   and  allows  the  user  to  set  up  nonstandard */
/* attributes of the Packet subtype.                          */
/* ---------------------------------------------------------- */

	// Make sure that the user defines 'setup' for all extended
	// subtypes
	assert (sizeof (Message) == m->frameSize (),
	"Packet: 'setup (Message*)' undefined for an extended Packet subtype");
}

void    Packet::fill (Station *s, Station *r, long tl, long il, long tp) {

/* ---------------------------------------------- */
/* Fills a row packet with the specified contents */
/* ---------------------------------------------- */

	Sender = (s != NULL) ? s->Id : NONE;
	Receiver = (r != NULL) ? r->Id : NONE;
	TLength = tl;
	assert (TLength >= 0, "Packet: 'fill' negative TLength");
	if (il != NONE) {
		assert (il <= tl, "Packet: 'fill' ILength > TLength");
		ILength = il;
	} else
		ILength = TLength;

	assert (!isTrafficId (tp),
	form ("Packet: 'fill' illegal TP (%1d) for a non-standard packet", tp));

	TP = (int) tp;
};

void    Packet::fill (long s, long r, long tl, long il, long tp) {

/* ---------------------------------------------------------- */
/* Fills  a  row  packet with the specified contents (another */
/* version)                                                   */
/* ---------------------------------------------------------- */

	assert (isStationId (s) || s == NONE,
		form ("Packet: 'fill' illegal sender id (%1d)", s));
	assert (isStationId (r) || r == NONE,
		form ("Packet: 'fill' illegal receiver id (%1d)", r));
	Sender = s;
	Receiver = r;
	TLength = tl;
	assert (TLength >= 0, "Packet: 'fill' negative TLength");
	if (il != NONE) {
		assert (il <= tl, "Packet: 'fill' ILength > TLength");
		ILength = il;
	} else
		ILength = TLength;

	assert (!isTrafficId (tp),
	form ("Packet: 'fill' illegal TP (%1d) for a non-standard packet", tp));

	TP = (int) tp;
};

int Traffic::getPacket (Packet *p, long min, long max, long frm) {

/* ---------------------------------------------------------- */
/* Acquires  a  packet (of "this" type) and puts it into `p'. */
/* Returns NO, if there is no packet of "this" type           */
/* ---------------------------------------------------------- */

	Message *m;
	long    l;

	if_from_observer ("Traffic->getPacket: called from an observer");

	assert (p->isEmpty (), "Traffic->getPacket: the buffer is full");

	if (TheStation->MQHead == NULL || (m = TheStation->MQHead [Id]) ==
		NULL) {
		// No message queued
		return (NO);
	}

	assert (m->Length >= 0,
		"Traffic->getPacket: negative message length");

	p->QTime = m->QTime;
	if (p->QTime > p->TTime) p->TTime = p->QTime;
	p->Sender      = TheStation->Id;
	p->Receiver    = m->Receiver;
	p->TP          = (int)Id;

	if (max && (m -> Length > max))
		l = p->ILength = max;
	else
		l = p->ILength = m -> Length;

	p->Flags = 1 << PF_full;        // The only flag
	if (!isStationId (p->Receiver)) {
		// NOTE !!!!  This trick only works under the assumption that
		// an address cannot be confused with a (not-too-big) integer
		// number representing a station Id.
		if (p->Receiver != NONE) setFlag (p->Flags, PF_broadcast);
	}

	p->TLength = frm + ((l < min) ? min : l);
	Info02 = (void*) Id;
	zz_setptr (p) = ptviptr;        // The virtual functions
	p->setup (m);                   // Private attributes

	if ((m->Length -= l) == 0) {
		// The last packet of the message
		setFlag (p->Flags, PF_last);
		Info01 = NULL;

		// Deallocate the message
		queue_out (m);
		if (TheStation->MQTail [Id] == m)
			// The last message being removed - adjust the tail
			TheStation->MQTail [Id] = NULL;
	
		delete (m);
	} else
		Info01 = (void*)m;
	return (YES);
}

int Traffic::getPacket (Packet *p, ZZ_MTTYPE tf, long min, long max, long frm) {

/* ---------------------------------------------------------- */
/* Acquires a packet (of "this" type) whose message satisfies */
/* condition  'tf' and puts it into `p'. Returns NO, if there */
/* is no such packet                                          */
/* ---------------------------------------------------------- */

	Message *m;
	long    l;

	if_from_observer ("Traffic->getPacket: called from an observer");

	assert (p->isEmpty (), "Traffic->getPacket: the buffer is full");

	if (TheStation->MQHead == NULL) {
		// No message queued
		return (NO);
	}

	for (m = TheStation->MQHead [Id]; m != NULL && !tf (m); m = m->next);

	if (m == NULL) {
		// No message satisfying 'tf'
		return (NO);
	}

	assert (m->Length >= 0,
		"Traffic->getPacket: negative message length");

	p->QTime = m->QTime;
	if (p->QTime > p->TTime) p->TTime = p->QTime;
	p->Sender      = TheStation->Id;
	p->Receiver    = m->Receiver;
	p->TP          = (int) Id;

	if (max && (m -> Length > max))
		l = p->ILength = max;
	else
		l = p->ILength = m -> Length;

	p->Flags = 1 << PF_full;        // The only flag
	if (!isStationId (p->Receiver)) {
		// NOTE !!!!  This trick only works under the assumption that
		// an address cannot be confused with a (not-too-big) integer
		// number representing a station Id.
		if (p->Receiver != NONE) setFlag (p->Flags, PF_broadcast);
	}

	p->TLength = frm + ((l < min) ? min : l);
	Info02 = (void*) Id;
	zz_setptr (p) = ptviptr;        // The virtual functions
	p->setup (m);                   // Private attributes

	if ((m->Length -= l) == 0) {
		// The last packet of the message
		setFlag (p->Flags, PF_last);
		Info01 = NULL;

		queue_out (m);
		// Deallocate the message
		if (TheStation->MQTail [Id] == m)
			// The last message being removed - adjust the tail
			TheStation->MQTail [Id] =
				(TheStation->MQHead [Id] == NULL) ? NULL :
					m->prev;
	
		delete (m);
	} else
		Info01 = (void*)m;
	return (YES);
}

int     Traffic::isPacket () {

/* ---------------------------------------------------------- */
/* Returns  YES,  if  there  is a packet of "this" type to be */
/* acquired from the Client                                   */
/* ---------------------------------------------------------- */

	if (TheStation->MQHead == NULL || TheStation->MQHead [Id] == NULL) {
		Info01 = NULL;
		Info02 = (void*) NONE;
		return (NO);
	}

	Info01 = (void*) (TheStation->MQHead [Id]);
	Info02 = (void*) Id;
	return (YES);
};

int     Traffic::isPacket (ZZ_MTTYPE tf) {

/* ---------------------------------------------------------- */
/* Returns   YES,  if  there  is  a  packet  of  "this"  type */
/* satisfying 'tf' to be acquired from the Client             */
/* ---------------------------------------------------------- */

	Message *m;

	if (TheStation->MQHead == NULL) {
		Info01 = NULL;
		Info02 = (void*) NONE;
		return (NO);
	}

	for (m = TheStation->MQHead [Id]; m != NULL && !tf (m); m = m->next);

	if (m == NULL) {
		Info01 = NULL;
		Info02 = (void*) NONE;
		return (NO);
	}

	Info01 = (void*) m;
	Info02 = (void*) Id;
	return (YES);
};

int     zz_client::isPacket () {

/* ---------------------------------------------------------- */
/* Returns  YES,  if  there  is  a  packet  of any type to be */
/* acquired from the Client                                   */
/* ---------------------------------------------------------- */

	int     i;

	if (TheStation->MQHead == NULL) {
		Info01 = NULL;
		Info02 = (void*) NONE;
		return (NO);
	}

	for (i = 0; i < NTraffics; i++) {
		if (TheStation->MQHead [i] != NULL) {
			Info01 = (void*) (TheStation->MQHead [i]);
			Info02 = (void*) i;
			return (YES);
		}
	}

	Info01 = NULL;
	Info02 = (void*) NONE;
	return (NO);
};

int     zz_client::isPacket (ZZ_MTTYPE tf) {

/* ---------------------------------------------------------- */
/* Returns  YES,  if there is a packet of any type satisfying */
/* 'tf' to be acquired from the Client                        */
/* ---------------------------------------------------------- */

	int     i;
	Message *m;

	if (TheStation->MQHead == NULL) {
		Info01 = NULL;
		Info02 = (void*) NONE;
		return (NO);
	}

	for (i = 0; i < NTraffics; i++) {
		for (m = TheStation->MQHead [i]; m != NULL && !tf (m);
			m = m->next);
		if (m != NULL) {
			Info01 = (void*) (TheStation->MQHead [i]);
			Info02 = (void*) i;
			return (YES);
		}
	}

	Info01 = NULL;
	Info02 = (void*) NONE;
	return (NO);
};

int zz_client::getPacket (Packet *p, long min, long max, long frm) {

/* ---------------------------------------------------------- */
/* Acquires  a  packet  (of  any  type) and puts it into `p'. */
/* Returns NO, if no message is queued at the station         */
/* ---------------------------------------------------------- */


	Message *m, *mm;
	long    l;
	int     i, mtp, mtype;
	TIME    qt;

	if_from_observer ("getPacket: called from an observer");
	assert (p->isEmpty (), "getPacket: the buffer is full");

	if (TheStation->MQHead == NULL) {
		return (NO);
	}

	if (NTraffics == 1) {
		// A special (and frequent) case worth speeding up
		if ((m = *(TheStation->MQHead)) == NULL) {
			return (NO);
		}
		mtype = 0;
	} else {
		// More message types - determine the earliest one
		for (mtp = i = (int) toss (NTraffics), qt = TIME_inf, m = NULL;
			i < NTraffics; i++) {
			// Circular search started at random
			if ((mm = TheStation->MQHead [i]) == NULL) continue;
			if (qt <= mm->QTime) continue;
			qt = (m = mm) -> QTime;
			mtype = i;
		}

		for (i = 0; i < mtp; i++) {
			// The second part of the search
			if ((mm = TheStation->MQHead [i]) == NULL) continue;
			if (qt <= mm->QTime) continue;
			qt = (m = mm) -> QTime;
			mtype = i;
		}

		if (m == NULL) {
			return (NO);
		}
	}

	assert (m->Length >= 0, "getPacket: negative message length");

	p->QTime = m->QTime;
	if (p->QTime > p->TTime) p->TTime = p->QTime;
	p->Sender      = TheStation->Id;
	p->Receiver    = m->Receiver;
	p->TP          = mtype;

	if (max && (m -> Length > max))
		l = p->ILength = max;
	else
		l = p->ILength = m -> Length;

	p->Flags = 1 << PF_full;        // The only flag
	if (!isStationId (p->Receiver)) {
		// NOTE !!!!  This trick only works under the assumption that
		// an address cannot be confused with a (not-too-big) integer
		// number representing a station Id.
		if (p->Receiver != NONE) setFlag (p->Flags, PF_broadcast);
	}

	p->TLength = frm + ((l < min) ? min : l);
	Info02 = (void*) mtype;
	zz_setptr (p) = idToTraffic (mtype)->ptviptr;
	p->setup (m);                   // Private attributes

	if ((m->Length -= l) == 0) {
		// The last packet of the message
		setFlag (p->Flags, PF_last);
		Info01 = NULL;

		// Deallocate the message
		queue_out (m);
		if (TheStation->MQTail [mtype] == m)
			// The last message being removed - adjust the tail
			TheStation->MQTail [mtype] = NULL;
	
		delete (m);
	} else
		Info01 = (void*)m;
	return (YES);
}

int zz_client::getPacket (Packet *p, ZZ_MTTYPE tf, long min, long max,
	long frm) {

/* ---------------------------------------------------------- */
/* Acquires  a  packet  (of any type) whose message satisfies */
/* condition 'tf' and puts it into `p'.  Returns  NO,  if  no */
/* message satisfying 'tf' is queued at the station           */
/* ---------------------------------------------------------- */

	Message *m, *mm;
	long    l;
	int     i, mtp, mtype;
	TIME    qt;

	if_from_observer ("getPacket: called from an observer");
	assert (p->isEmpty (), "getPacket: the buffer is full");

	if (TheStation->MQHead == NULL) {
		return (NO);
	}

	if (NTraffics == 1) {
		// A special (and frequent) case worth speeding up
		for (m = *(TheStation->MQHead); m != NULL && !tf (m);
			m = m->next);
		if (m == NULL) {
			return (NO);
		}
		mtype = 0;
	} else {
		// More message types - determine the earliest one
		for (mtp = i = (int) toss (NTraffics), qt = TIME_inf, m = NULL;
			i < NTraffics; i++) {
			// Circular search started at random
			for (mm = TheStation->MQHead [i]; ; mm = mm->next) {
				if (mm == NULL || qt < mm->QTime) goto ES;
				if (tf (mm)) break;
			}
			qt = (m = mm) -> QTime;
			mtype = i;
		  ES:   continue;
		}

		for (i = 0; i < mtp; i++) {
			for (mm = TheStation->MQHead [i]; ; mm = mm->next) {
				if (mm == NULL || qt < mm->QTime) goto ET;
				if (tf (mm)) break;
			}
			qt = (m = mm) -> QTime;
			mtype = i;
		  ET:   continue;
		}

		if (m == NULL) {
			return (NO);
		}
	}

	assert (m->Length >= 0, "getPacket: negative message length");

	p->QTime = m->QTime;
	if (p->QTime > p->TTime) p->TTime = p->QTime;
	p->Sender      = TheStation->Id;
	p->Receiver    = m->Receiver;
	p->TP          = mtype;

	if (max && (m -> Length > max))
		l = p->ILength = max;
	else
		l = p->ILength = m -> Length;

	p->Flags = 1 << PF_full;        // The only flag
	if (!isStationId (p->Receiver)) {
		// NOTE !!!!  This trick only works under the assumption that
		// an address cannot be confused with a (not-too-big) integer
		// number representing a station Id.
		if (p->Receiver != NONE) setFlag (p->Flags, PF_broadcast);
	}

	p->TLength = frm + ((l < min) ? min : l);
	Info02 = (void*) mtype;
	zz_setptr (p) = idToTraffic (mtype)->ptviptr;
	p->setup (m);                   // Private attributes

	if ((m->Length -= l) == 0) {
		// The last packet of the message
		setFlag (p->Flags, PF_last);
		Info01 = NULL;

		// Deallocate the message
		queue_out (m);
		if (TheStation->MQTail [mtype] == m)
			// The last message being removed - adjust the tail
			TheStation->MQTail [mtype] =
				(TheStation->MQHead [mtype] == NULL) ? NULL :
					m->prev;
		delete (m);
	} else
		Info01 = (void*)m;
	return (YES);
}

int     Message::getPacket (Packet *p, long min, long max, long frm) {

/* ---------------------------------------------------------- */
/* Acquires  a  packet from the specific  message and puts it */
/* into `p'. Always returns YES (for compatibility  with  the */
/* other versions of the function).                           */
/* ---------------------------------------------------------- */


	long    l;

	if_from_observer ("getPacket: called from an observer");
	assert (p->isEmpty (), "getPacket: the buffer is full");
	assert (Length >= 0, "getPacket: negative message length");

	p->QTime = QTime;
	if (p->QTime > p->TTime) p->TTime = p->QTime;
	p->Sender      = TheStation->Id;
	p->Receiver    = Receiver;
	p->TP          = TP;

	if (max && (Length > max))
		l = p->ILength = max;
	else
		l = p->ILength = Length;

	p->Flags = 1 << PF_full;        // The only flag
	if (!isStationId (p->Receiver)) {
		// NOTE !!!!  This trick only works under the assumption that
		// an address cannot be confused with a (not-too-big) integer
		// number representing a station Id.
		if (p->Receiver != NONE) setFlag (p->Flags, PF_broadcast);
	}

	p->TLength = frm + ((l < min) ? min : l);
	Info02 = (void*) (p->TP);
	zz_setptr (p) = idToTraffic (p->TP)->ptviptr;
	p->setup (this);               // Private attributes

	if ((Length -= l) == 0) {
		// The last packet of the message
		setFlag (p->Flags, PF_last);
		Info01 = NULL;

		// Deallocate the message
		if (prev != NULL) {
			// The message was taken from a queue
			assert ((TheStation->MQHead [p->TP] != NULL)
				&& (TheStation->MQTail [p->TP]
					!= NULL),
				"getPacket: message queue corrupted");
			queue_out (this);
			if (TheStation->MQTail [p->TP] == this)
				// Adjust the tail pointer
				TheStation->MQTail [p->TP] =
					(TheStation->MQHead [p->TP] ==
						NULL) ? NULL : prev;
		}
		delete (this);
	} else
		Info01 = (void*)this;
	return (YES);
}

void    zz_client::receive (Packet *p, Link *l) {

	Traffic *tp;

	if (l != NULL) l->spfmPRC (p);

	if (p->isStandard ()) {

		tp = idToTraffic (p->TP);

		// After receiving a packet

		tp->spfmPRC (p);

		// Check if it is the last packet of a message

		if (p->isLast ()) {

			// After receiving a message

			tp->spfmMRC (p);
			if (l != NULL) l->spfmMRC (p);
		}
	}
};

void    zz_client::release (Packet *p) {

/* ---------------------------------------------------------- */
/* Executed by the sender after the packet `p'is successfully */
/* sent                                                       */
/* ---------------------------------------------------------- */

	Traffic         *mtype;

	if_from_observer ("release: called from an observer");

	assert (p->isFull (), "release: packet buffer is empty");
	assert (isTrafficId (p->TP), "release: illegal packet's Traffic Id");

	mtype = idToTraffic (p->TP);
	mtype -> spfmPTR (p);
	if (p->isLast ()) mtype -> spfmMTR (p);
	clearFlag (p->Flags, PF_full);
	p->TTime = Time;
}

int     Packet::isMy () {

/* ---------------------------------------------------------- */
/* Determines  if  the  packet  is  addressed to this station */
/* (TheStation)                                               */
/* ---------------------------------------------------------- */

	if (isStationId (Receiver)) {
		return (Receiver == TheStation->Id);
	}
	if (Receiver == NONE) return (NO);

	// A broadcast packet
	return (((SGroup*)Receiver)->occurs (TheStation->Id));
}

static  char    fg [8];

char    *Packet::zz_pflags () {

/* ---------------------------------------------------- */
/* Converts standard packet flags to a printable string */
/* ---------------------------------------------------- */

	int     i, j;

	for (j = 0, i = PF_esignal; i <= PF_broadcast; i++, j++)
		if (flagSet (Flags, i))
			fg [j] = 'Y';
		else
			fg [j] = 'N';
	fg [j] = '\0';
	return (fg);
}

sexposure (zz_client)

	sonpaper {

		sexmode (0)

			exPrint0 (Hdr, (int) SId);    // Requests

		sexmode (1)

			exPrint1 (Hdr);               // Performance

		sexmode (2)

			exPrint2 (Hdr, (int) SId);    // Queues

		sexmode (3)

			exPrint3 (Hdr);               // Definitions (traffic)
	}

	sonscreen {

		sexmode (0)

			exDisplay0 ((int) SId);       // Requests

		sexmode (1)

			exDisplay1 ();                // Performance

		sexmode (2)

			exDisplay2 ((int) SId);       // Queues

		sexmode (3)

			exDisplay3 ();          // Sent-received statistics

	}
}

void    zz_client::exPrint0 (char *hdr, int sid) {

/* --------------------- */
/* Print Client requests */
/* --------------------- */

	ZZ_REQUEST         *r;
	ZZ_EVENT           *e;
	long               l;

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () << ") Full AI wait list";
		if (isStationId (sid))
			Ouf << " for " << idToStation (sid)->getOName ();
		Ouf << ":\n\n";
	}

	if (!isStationId (sid))
		Ouf << "       Time   St";
	else
		Ouf << "           Time ";

	Ouf << "    Process/Idn     CState      Event         AI/Idn" <<
		"      State\n";

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (isStationId (sid) && (e->station == NULL ||
			e->chain == NULL || ident (e->station) != sid))
				continue;

		if ((r = e->chain) == NULL) {
			// System event
			if (zz_flg_nosysdisp || e->ai->Class != AIC_client)
				continue;
			if (!isStationId (sid))
				ptime (e->waketime, 11);
			else
				ptime (e->waketime, 15);

			Ouf << "* --- ";
			print (e->process->getTName (), 10);
			Ouf << form ("/%3d ", zz_trunc (e->process->Id, 3));
			// State name
			print (e->process->zz_sn (e->pstate), 10);

			Ouf << ' ';
			print (e->ai->zz_eid (e->event_id), 10);

			Ouf << ' ';
			print (e->ai->getTName (), 10);
			if ((l = e->ai->Id) != NONE) {
				if (e->ai->Class == AIC_port)
					l = l & 0177777;
				Ouf << form ("/%03d ", zz_trunc (l, 3));
			} else
				Ouf << "     ";

			print (e->process->zz_sn (e->pstate), 10);

			Ouf << '\n';
		} else {
			
			// A station event

			while (1) {

				if (r->ai->Class == AIC_client ||
					r->ai->Class == AIC_traffic) {

					if (!isStationId (sid))
						ptime (r->when, 11);
					else
						ptime (r->when, 15);

					if (e->waketime < r->when)
						Ouf << ' ';     // Obsolete
					else if (e->chain == r)
						Ouf << '*';     // Current
					else
						Ouf << '?';

					if (!isStationId (sid)) {
						// Identify the station
						if (e->station != NULL &&
							ident (e->station) >= 0)
							Ouf << form ("%4d ",
					zz_trunc (ident (e->station), 3));
						else
							Ouf << " Sys ";
					}
					print (e->process->getTName (), 10);
					Ouf << form ("/%3d ",
						zz_trunc (e->process->Id, 3));
					print (e->process->zz_sn (r->pstate),
						10);

					Ouf << ' ';
					print (r->ai->zz_eid (r->event_id),
						10);

					Ouf << ' ';

					print (e->ai->getTName (), 10);
					if ((l = e->ai->Id) != NONE) {
						if (e->ai->Class == AIC_port)
							l = l & 0177777;
						Ouf << form ("/%03d ",
							zz_trunc (l, 3));
					} else
						Ouf << "     ";

					print (e->process->zz_sn
							(e->chain->pstate), 10);
					Ouf << '\n';
				}

				if ((r = r->other) == e->chain) break;
			}
		}
	}

	Ouf << "\n(" << getOName () << ") End of list\n\n";
}

void    zz_client::exPrint1 (char *hdr) {

/* -------------------------- */
/* Print performance measures */
/* -------------------------- */

	int             l, i;
	ZZ_TRVariable   *trv;

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () << ") Global performance measures";
		Ouf << ":\n\n";
	}

	// AMD

	for (i = 0; i < NTraffics; i++) {
		Assert (idToTraffic (i)->FlgSPF == OFF ||
		    idToTraffic (i) -> RVAMD != NULL,
			"Client->exposure: standard Client not started");
	}

	tcreate (TYPE_long);
	for (i = l = 0; i < NTraffics; i++) {
		if (idToTraffic(i)->FlgSPF == OFF) continue;
		if (l == 0) {
			l++;
			*trv = *((ZZ_TRVariable*) ((idToTraffic (i)->RVAMD)));
			continue;
		}
		combineRV (idToTraffic (i)->RVAMD, trv);
	}
	if (l) trv->printCnt ("AMD - Absolute message delay:");
	delete (trv);

	// APD

	tcreate (TYPE_long);
	for (i = l = 0; i < NTraffics; i++) {
		if (idToTraffic(i)->FlgSPF == OFF) continue;
		if (l == 0) {
			l++;
			*trv = *((ZZ_TRVariable*) ((idToTraffic (i)->RVAPD)));
			continue;
		}
		combineRV (idToTraffic (i)->RVAPD, trv);
	}
	if (l) trv->printCnt ("APD - Absolute packet delay:");
	delete (trv);

	// WMD

	tcreate (TYPE_BITCOUNT);
	for (i = l = 0; i < NTraffics; i++) {
		if (idToTraffic(i)->FlgSPF == OFF) continue;
		if (l == 0) {
			l++;
			*trv = *((ZZ_TRVariable*) ((idToTraffic (i)->RVWMD)));
			continue;
		}
		combineRV (idToTraffic (i)->RVWMD, trv);
	}
	if (l) trv->printCnt ("WMD - Weighted message delay:");
	delete (trv);

	// MAT

	tcreate (TYPE_long);
	for (i = l = 0; i < NTraffics; i++) {
		if (idToTraffic(i)->FlgSPF == OFF) continue;
		if (l == 0) {
			l++;
			*trv = *((ZZ_TRVariable*) ((idToTraffic (i)->RVMAT)));
			continue;
		}
		combineRV (idToTraffic (i)->RVMAT, trv);
	}
	if (l) trv->printCnt ("MAT - Message access time:");
	delete (trv);

	// PAT

	tcreate (TYPE_long);
	for (i = l = 0; i < NTraffics; i++) {
		if (idToTraffic(i)->FlgSPF == OFF) continue;
		if (l == 0) {
			l++;
			*trv = *((ZZ_TRVariable*) ((idToTraffic (i)->RVPAT)));
			continue;
		}
		combineRV (idToTraffic (i)->RVPAT, trv);
	}
	if (l) trv->printCnt ("PAT - Packet access time:");
	delete (trv);

	// MLS

	tcreate (TYPE_long);
	for (i = l = 0; i < NTraffics; i++) {
		if (idToTraffic(i)->FlgSPF == OFF) continue;
		if (l == 0) {
			l++;
			*trv = *((ZZ_TRVariable*) ((idToTraffic (i)->RVMLS)));
			continue;
		}
		combineRV (idToTraffic (i)->RVMLS, trv);
	}
	if (l) trv->printCnt ("MLS - Message length distribution:");
	delete (trv);

	print (zz_NGMessages, "    Number of generated messages:   ");
	print (zz_NQMessages, "    Number of queued messages:      ");
	print (zz_NTMessages, "    Number of transmitted messages: ");
	print (zz_NRMessages, "    Number of received messages:    ");
	print (zz_NTPackets , "    Number of transmitted packets:  ");
	print (zz_NRPackets , "    Number of received packets:     ");
	print (zz_NQBits    , "    Number of queued bits:          ");
	print (zz_NTBits    , "    Number of transmitted bits:     ");
	print (zz_NRBits    , "    Number of received bits:        ");
	print (Etu * ((double)zz_NRBits / (Time == TIME_0 ? TIME_1 : Time)),
			      "    Throughput:                     ");

	Ouf << "\n(" << getOName () << ") End of list\n\n";
}

void    zz_client::exPrint2 (char *hdr, int sid) {

/* -------------------- */
/* Print message queues */
/* -------------------- */

	int             l, i;
	long            mc, bc;
	Message         *m, **mm;
	TIME            t;

	if (hdr != NULL) {
		Ouf << hdr;
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () << ") Message queues at";
		if (isStationId (sid))
			Ouf << ' ' << idToStation (sid)->getOName ();
		else
			Ouf << " stations";
		Ouf << ':';
	}

	Ouf << "\n\n";

	if (isStationId (sid)) {

		// Station-specific printout

	Ouf << "                Time   Traffic     Length   Receiver\n";

		if (idToStation (sid)->MQHead != NULL) {

			mm = new Message* [NTraffics];
			for (i = 0; i < NTraffics; i++)
				mm [i] = idToStation (sid)->MQHead [i];


			while (1) {

				for (m = NULL, i = l = 0, t = TIME_inf;
					i < NTraffics; i++) {

					if (mm [i] == NULL) continue;
					if (mm [i] -> QTime < t)
						t = (m = mm [l = i])->QTime;
				}

				if (m == NULL) break;

				mm [l] = m->next;

				print (m->QTime, 20);
				print (l, 10);
				print (m->Length, 11);
				if (isStationId (m->Receiver))
					print (m->Receiver, 11);
				else if (m->Receiver == NONE)
					print ("none", 11);
				else
					print ("bcast", 11);

				Ouf << '\n';
			}

			delete (mm);

		}

	} else {

		Ouf << "Station   Messages       Bits\n";

		for (i = 0; i < NStations; i++) {
			mc = bc = 0;
			if (idToStation (i)->MQHead != NULL) {
				for (l = 0; l < NTraffics; l++) {
					for (m = idToStation(i)->MQHead[l];
						m != NULL;
							m = m->next) {

						mc ++;
						bc += m->Length;
					}
				}
			}

			print (i, 7);
			print (mc, 11);
			print (bc, 11);
			Ouf << '\n';
		}
	}

	Ouf << "\n(" << getOName () << ") End of list\n\n";
}

void    zz_client::exPrint3 (char *hdr) {

/* --------------------------------- */
/* Print traffic pattern definitions */
/* --------------------------------- */

	int             i;

	if (hdr != NULL) {
		Ouf << hdr;
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () << ") Traffic definitions:";
	}

	Ouf << "\n\n";

	for (i = 0; i < NTraffics; i++) {

		char    tt [64];

		strcpy (tt, form ("%s:", idToTraffic (i) -> getOName ()));

		idToTraffic (i) -> printDef (tt);
	}

	Ouf << "(" << getOName () << ") End of list\n\n";
}

void    zz_client::exDisplay0 (int sid) {

/* ----------------------- */
/* Display Client requests */
/* ----------------------- */

	ZZ_REQUEST         *r;
	ZZ_EVENT           *e;
	long               l;

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (isStationId (sid) && (e->station == NULL ||
			e->chain == NULL || ident (e->station) != sid))
				continue;

		if ((r = e->chain) == NULL) {
			// System event
			if (zz_flg_nosysdisp || e->ai->Class != AIC_client)
				continue;
			dtime (e->waketime);
			display ('*');
			display ("---");
			display (e->process->getTName ());
			display (e->process->Id);
			display (e->process->zz_sn (e->pstate));
			display (e->ai->zz_eid (e->event_id));
			display (e->ai->getTName ());
			if ((l = e->ai->Id) != NONE) {
				if (e->ai->Class == AIC_port)
					l = l & 0177777;
				display (l);
			} else
				display (' ');

			display (e->process->zz_sn (e->pstate));

		} else {
			
			// A station event

			while (1) {
				if (r->ai->Class == AIC_client ||
					r->ai->Class == AIC_traffic) {

					dtime (r->when);
					if (e->waketime < r->when)
						display (' ');
					else if (e->chain == r)
						display ('*');
					else
						display ('?');

					if (!isStationId (sid)) {
						// Identify the station
						if (e->station != NULL &&
							ident (e->station) >= 0)
						   display (ident (e->station));
						else
						   display ("Sys");
					}
					display (e->process->getTName ());
					display (e->process->Id);
					display (e->process->zz_sn (r->pstate));
					display (r->ai->zz_eid (r->event_id));
					display (e->ai->getTName ());
					if ((l = e->ai->Id) != NONE) {
						if (e->ai->Class == AIC_port)
							l = l & 0177777;
						display (l);
					} else
						display (' ');

					display (e->process->zz_sn
							    (e->chain->pstate));
				}
				if ((r = r->other) == e->chain) break;
			}
		}
	}
}

void    zz_client::exDisplay1 () {

/* ---------------------------- */
/* Display performance measures */
/* ---------------------------- */

	int             l, i;
	ZZ_TRVariable   *trv;

	// AMD

	for (i = 0; i < NTraffics; i++) {
		if (idToTraffic (i)->FlgSPF != OFF &&
			idToTraffic (i) -> RVAMD == NULL) {
				display (' ');
				return;         // Client not started yet
		}
	}

	tcreate (TYPE_long);
	for (i = l = 0; i < NTraffics; i++) {
		if (idToTraffic(i)->FlgSPF == OFF) continue;
		if (l == 0) {
			l++;
			*trv = *((ZZ_TRVariable*) ((idToTraffic (i)->RVAMD)));
			continue;
		}
		combineRV (idToTraffic (i)->RVAMD, trv);
	}
	if (l) {
		display ("AMD");
		trv->displayOut (2);    // Abbreviated mode - one line
	}
	delete (trv);

	// APD

	tcreate (TYPE_long);
	for (i = l = 0; i < NTraffics; i++) {
		if (idToTraffic(i)->FlgSPF == OFF) continue;
		if (l == 0) {
			l++;
			*trv = *((ZZ_TRVariable*) ((idToTraffic (i)->RVAPD)));
			continue;
		}
		combineRV (idToTraffic (i)->RVAPD, trv);
	}
	if (l) {
		display ("APD");
		trv->displayOut (2);
	}
	delete (trv);

	// WMD

	tcreate (TYPE_BITCOUNT);
	for (i = l = 0; i < NTraffics; i++) {
		if (idToTraffic(i)->FlgSPF == OFF) continue;
		if (l == 0) {
			l++;
			*trv = *((ZZ_TRVariable*) ((idToTraffic (i)->RVWMD)));
			continue;
		}
		combineRV (idToTraffic (i)->RVWMD, trv);
	}
	if (l) {
		display ("WMD");
		trv->displayOut (2);
	}
	delete (trv);

	// MAT

	tcreate (TYPE_long);
	for (i = l = 0; i < NTraffics; i++) {
		if (idToTraffic(i)->FlgSPF == OFF) continue;
		if (l == 0) {
			l++;
			*trv = *((ZZ_TRVariable*) ((idToTraffic (i)->RVMAT)));
			continue;
		}
		combineRV (idToTraffic (i)->RVMAT, trv);
	}
	if (l) {
		display ("MAT");
		trv->displayOut (2);
	}
	delete (trv);

	// PAT

	tcreate (TYPE_long);
	for (i = l = 0; i < NTraffics; i++) {
		if (idToTraffic(i)->FlgSPF == OFF) continue;
		if (l == 0) {
			l++;
			*trv = *((ZZ_TRVariable*) ((idToTraffic (i)->RVPAT)));
			continue;
		}
		combineRV (idToTraffic (i)->RVPAT, trv);
	}
	if (l) {
		display ("PAT");
		trv->displayOut (2);
	}
	delete (trv);
	
	// MLS

	tcreate (TYPE_long);
	for (i = l = 0; i < NTraffics; i++) {
		if (idToTraffic(i)->FlgSPF == OFF) continue;
		if (l == 0) {
			l++;
			*trv = *((ZZ_TRVariable*) ((idToTraffic (i)->RVMLS)));
			continue;
		}
		combineRV (idToTraffic (i)->RVMLS, trv);
	}
	if (l) {
		display ("MLS");
		trv->displayOut (2);
	}
	delete (trv);
}

void    zz_client::exDisplay3 () {

/* -------------------------------- */
/* Display sent-received statistics */
/* -------------------------------- */

	display (Etu * ((double)zz_NRBits / (Time == TIME_0 ? TIME_1 : Time)));
	display (zz_NGMessages);
	display (zz_NQMessages);
	display (zz_NTMessages);
	display (zz_NRMessages);
	display (zz_NTPackets );
	display (zz_NRPackets );
	display (zz_NQBits    );
	display (zz_NTBits    );
	display (zz_NRBits    );
}

void    zz_client::exDisplay2 (int sid) {

/* ---------------------- */
/* Display message queues */
/* ---------------------- */

	int             l, i;
	long            bc, maxbc;
	Message         *m, **mm;
	TIME            t;


	if (isStationId (sid)) {

		// Station-specific display

		if (idToStation (sid)->MQHead != NULL) {

			mm = new Message* [NTraffics];
			for (i = 0; i < NTraffics; i++)
				mm [i] = idToStation (sid)->MQHead [i];

			while (1) {

				for (m = NULL, i = l = 0, t = TIME_inf;
					i < NTraffics; i++) {

					if (mm [i] == NULL) continue;
					if (mm [i] -> QTime < t)
						t = (m = mm [l = i])->QTime;
				}

				if (m == NULL) break;

				mm [l] = m->next;

				display (m->QTime);
				display (l);
				display (m->Length);
				if (isStationId (m->Receiver))
					display (m->Receiver);
				else if (m->Receiver == NONE)
					display ("none");
				else
					display ("bcast");
			}
			delete (mm);
		}

		return;

	}

	// Graphical display for all stations

	maxbc = 0;
	// Calculate the maximum
	for (i = 0; i < NStations; i++) {
		bc = 0;
		if (idToStation (i)->MQHead != NULL) {
			for (l = 0; l < NTraffics; l++) {
				for (m = idToStation(i)->MQHead[l];
					m != NULL;
						m = m->next) {

					bc ++;
				}
			}
		}

		if (maxbc < bc) maxbc = bc;
	}

	if (bc == 0) bc = 1;

	startRegion (0.0, (double)(NStations-1), 0.0, (double) maxbc);
	startSegment (0);       // Unconnected points

	for (i = 0; i < NStations; i++) {
		bc = 0;
		if (idToStation (i)->MQHead != NULL) {
			for (l = 0; l < NTraffics; l++) {
				for (m = idToStation(i)->MQHead[l];
					m != NULL;
						m = m->next) {

					bc ++;
				}
			}
		}
		displayPoint ((double)i, (double)bc);
	}

	endSegment ();
	endRegion ();
}

sexposure (Traffic)

	sonpaper {

		sexmode (0)

			exPrint0 (Hdr, (int) SId);    // Request queue

		sexmode (1)

			exPrint1 (Hdr);               // Performance

		sexmode (2)

			exPrint2 (Hdr, (int) SId);    // Queues

		sexmode (3)

			exPrint3 (Hdr);
	}

	sonscreen {

		sexmode (0)

			exDisplay0 ((int) SId);       // Request queue

		sexmode (1)

			exDisplay1 ();                // Performance

		sexmode (2)

			exDisplay2 ((int) SId);       // Sent-received

		sexmode (3)

			exDisplay3 ();                // Queues
	}
}

void    Traffic::exPrint0 (char *hdr, int sid) {

/* ----------------------- */
/* Print the request queue */
/* ----------------------- */

	ZZ_REQUEST         *r;
	ZZ_EVENT           *e;
	long               l;

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () << ") Full AI wait list";
		if (isStationId (sid))
			Ouf << " for " << idToStation (sid)->getOName ();
		Ouf << ":\n\n";
	}

	if (!isStationId (sid))
		Ouf << "       Time   St";
	else
		Ouf << "           Time ";

	Ouf << "    Process/Idn     TState      Event         AI/Idn" <<
		"      State\n";

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (isStationId (sid) && (e->station == NULL ||
			e->chain == NULL || ident (e->station) != sid))
				continue;

		if ((r = e->chain) == NULL) {
			// Generation event (system)
			if (zz_flg_nosysdisp || e->ai->Class != AIC_client ||
					(int)(e->Info02) != Id) continue;

			if (!isStationId (sid))
				ptime (e->waketime, 11);
			else
				ptime (e->waketime, 15);

			Ouf << "* --- ";
			print (e->process->getTName (), 10);
			Ouf << form ("/%03d ",
					zz_trunc (e->process->Id, 3));
			// State name
			print (e->process->zz_sn (e->pstate), 10);

			Ouf << ' ';
			print (e->ai->zz_eid (e->event_id), 10);

			Ouf << ' ';

			print (e->ai->getTName (), 10);
			if ((l = e->ai->Id) != NONE) {
				if (e->ai->Class == AIC_port)
					l = l & 0177777;
				Ouf << form ("/%03d ", zz_trunc (l, 3));
			} else
				Ouf << "     ";

			print (e->process->zz_sn (e->pstate), 10);

			Ouf << '\n';
		} else {
			
			// A station event

			while (1) {

				if ((r->ai->Class == AIC_client &&
					r->event_id == Id) ||
							r->ai == this) {

					if (!isStationId (sid))
						ptime (r->when, 11);
					else
						ptime (r->when, 15);

					if (e->waketime < r->when)
						Ouf << ' ';     // Obsolete
					else if (e->chain == r)
						Ouf << '*';     // Current
					else
						Ouf << '?';

					if (!isStationId (sid)) {
						// Identify the station
						if (e->station != NULL &&
							ident (e->station) >= 0)
							Ouf << form ("%4d ",
					zz_trunc (ident (e->station), 3));
						else
							Ouf << " Sys ";
					}
					print (e->process->getTName (), 10);
					Ouf << form ("/%03d ",
						zz_trunc (e->process->Id, 3));
					print (e->process->zz_sn (r->pstate),
						10);

					Ouf << ' ';
					print (r->ai->zz_eid (r->event_id),
						10);

					Ouf << ' ';

					print (e->ai->getTName (), 10);
					if ((l = e->ai->Id) != NONE) {
						if (e->ai->Class == AIC_port)
							l = l & 0177777;
						Ouf << form ("/%03d ",
							zz_trunc (l, 3));
					} else
						Ouf << "     ";

					print (e->process->zz_sn
							(e->chain->pstate), 10);
					Ouf << '\n';
				}

				if ((r = r->other) == e->chain) break;
			}
		}
	}

	Ouf << "\n(" << getOName () << ") End of list\n\n";
}

void    Traffic::exPrint1 (char *hdr) {

/* -------------------------- */
/* Print performance measures */
/* -------------------------- */

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () << ") Performance measures:\n\n";
	}

	if (FlgSPF == OFF) {
		Ouf << ">>> Switched off <<<\n";
	} else {

		Assert (RVAMD != NULL,
			"Traffic->exposure: standard Client not started");
		// AMD
		RVAMD->printCnt ("AMD - Absolute message delay:");
		// APD
		RVAPD->printCnt ("APD - Absolute packet delay:");
		// WMD
		RVWMD->printCnt ("WMD - Weighted message delay:");
		// MAT
		RVMAT->printCnt ("MAT - Message access time:");
		// PAT
		RVPAT->printCnt ("PAT - Packet access time:");
		// MLS
		RVMLS->printCnt ("MLS - Message length statistics:");
	}

	print (NTMessages + NQMessages, "    Number of generated messages:   ");
	print (NQMessages             , "    Number of queued messages:      ");
	print (NTMessages             , "    Number of transmitted messages: ");
	print (NRMessages             , "    Number of received messages:    ");
	print (NTPackets              , "    Number of transmitted packets:  ");
	print (NRPackets              , "    Number of received packets:     ");
	print (NQBits                 , "    Number of queued bits:          ");
	print (NTBits                 , "    Number of transmitted bits:     ");
	print (NRBits                 , "    Number of received bits:        ");

	Ouf << "\n(" << getOName () << ") End of list\n\n";
}

void    Traffic::exPrint2 (char *hdr, int sid) {

/* -------------------- */
/* Print message queues */
/* -------------------- */

	int             i;
	long            mc, bc;
	Message         *m;

	if (hdr != NULL) {
		Ouf << hdr;
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () << ") Message queue";
		if (isStationId (sid))
			Ouf << " at " << idToStation (sid)->getOName ();
		else
			Ouf << "s at stations";
			
		Ouf << ':';
	}

	Ouf << "\n\n";

	if (isStationId (sid)) {

		// Station-specific printout

		Ouf << "                Time     Length   Receiver\n";

		if (idToStation (sid)->MQHead != NULL) {

			for (m = idToStation (sid)->MQHead [Id]; m != NULL;
				m = m->next) {

				print (m->QTime, 20);
				print (m->Length, 11);
				if (isStationId (m->Receiver))
					print (m->Receiver, 11);
				else if (m->Receiver == NONE)
					print ("none", 11);
				else
					print ("bcast", 11);

				Ouf << '\n';
			}
		}

	} else {

		Ouf << "Station   Messages       Bits\n";

		for (i = 0; i < NStations; i++) {
			mc = bc = 0;
			if (idToStation (i)->MQHead != NULL) {
				for (m = idToStation(i)->MQHead[Id];
					m != NULL; m = m->next) {

					mc ++;
					bc += m->Length;
				}
			}

			print (i, 7);
			print (mc, 11);
			print (bc, 11);
			Ouf << '\n';
		}
	}

	Ouf << "\n(" << getOName () << ") End of list\n\n";
}

void    Traffic::exPrint3 (char *hdr) {

/* -------------------------------- */
/* Print traffic pattern definition */
/* -------------------------------- */

	long            l;
	int             i, j;
	CGroup          *cg;
	SGroup          *sg;

	if (hdr != NULL) {
		Ouf << hdr;
	} else  {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () << ") Definition:";
	}

	Ouf << "\n\n";

	print ("MIT distribution:      ");

	if (DstMIT == EXPONENTIAL) {
		print ("EXPONENTIAL (Mean = ");
		print (ParMnMIT/Etu, 8);
		Ouf << ')';
	} else if (DstMIT == UNIFORM) {
		print ("UNIFORM     (Min  = ");
		print (ParMnMIT/Etu, 8);
		print (", Max  = ");
		print (ParMxMIT/Etu, 8);
		Ouf << ')';
	} else
		print ("UNDEFINED");
	Ouf << '\n';

	print ("MLE distribution:      ");

	if (DstMLE == EXPONENTIAL) {
		print ("EXPONENTIAL (Mean = ");
		print (ParMnMLE, 8);
		Ouf << ')';
	} else if (DstMLE == UNIFORM) {
		print ("UNIFORM     (Min  = ");
		print (ParMnMLE, 8);
		print (", Max  = ");
		print (ParMxMLE, 8);
		Ouf << ')';
	} else
		print ("UNDEFINED");
	Ouf << '\n';

	if (DstBIT != UNDEFINED) {      // Bursty traffic

		print ("BIT distribution:      ");

		if (DstBIT == EXPONENTIAL) {
			print ("EXPONENTIAL (Mean = ");
			print (ParMnBIT/Etu, 8);
			Ouf << ')';
		} else if (DstBIT == UNIFORM) {
			print ("UNIFORM     (Min  = ");
			print (ParMnBIT/Etu, 8);
			print (", Max  = ");
			print (ParMxBIT/Etu, 8);
			Ouf << ')';
		} else
			print ("UNDEFINED");
		Ouf << '\n';

		print ("BSI distribution:      ");

		if (DstBSI == EXPONENTIAL) {
			print ("EXPONENTIAL (Mean = ");
			print (ParMnBSI, 8);
			Ouf << ')';
		} else if (DstBSI == UNIFORM) {
			print ("UNIFORM     (Min  = ");
			print (ParMnBSI, 8);
			print (", Max  = ");
			print (ParMxBSI, 8);
			Ouf << ')';
		} else
			print ("UNDEFINED");
		Ouf << '\n';
	}

	print ("Standard Client is:    ");
	print ((FlgSCL == ON) ? "ON" : "OFF");
	Ouf << '\n';

	print ("Standard measures are: ");
	print ((FlgSPF == ON) ? "ON" : "OFF");
	Ouf << "\n\n";

	for (i = 0; i < NCGroups; i++) {

		cg = CGCluster [i];
		Ouf << form ("Communication group #%03d (density %8.2g)\n\n",
			i, cg->CGWeight);

		print ("Senders:   ");
		if ((sg = cg->Senders) == NULL || sg->NStations == NStations ||
			sg->SIdents == NULL) {

			l = NStations;
			print ("all stations");
		} else {
			l = sg->NStations;
			print (l, 5);
		}

		if (cg->SWeights == NULL) {
			print (", equal weights totaling to 1.0");
			if (l != NStations) {
				Ouf << '\n';
				for (j = 0; j < l; j++) {
					if (j % 10 == 0) Ouf << '\n';
					print (sg->SIdents [j], 6);
				}
			}
		} else {
			Ouf << '\n';
			for (j = 0; j < l; j++) {
				if (j % 4 == 0) Ouf << '\n';
				Ouf << '(';
				print ((l<NStations)? sg->SIdents [j] : j, 5);
				Ouf << ' ';
				print (cg->SWeights [j], 8);
				Ouf << ")   ";
			}
		}

		Ouf << "\n\n";

		print ("Receivers: ");
		if ((sg = cg->Receivers) == NULL || sg->NStations == NStations
			|| sg->SIdents == NULL) {

			l = NStations;
			print ("all stations");
		} else {
			l = sg->NStations;
			print (l, 5);
		}

		if (cg->RWeights == NULL || cg->RWeights == (float*) NONE) {
			if (cg->RWeights == NULL)
				print (", equal weights totaling to 1.0");
			else
				print (", broadcast group");
			Ouf << '\n';
			if (l != NStations) {
				for (j = 0; j < l; j++) {
					if (j % 10 == 0) Ouf << '\n';
					print (sg->SIdents [j], 6);
				}
			}
		} else {
			Ouf << '\n';
			for (j = 0; j < l; j++) {
				if (j % 4 == 0) Ouf << '\n';
				Ouf << '(';
				print ((l<NStations)? sg->SIdents [j] : j, 5);
				Ouf << ' ';
				print (cg->RWeights [j], 8);
				Ouf << ")   ";
			}
		}

		Ouf << "\n\n";
	}

	Ouf << "(" << getOName () << ") End of list\n\n";
}

void    Traffic::exDisplay0 (int sid) {

/* ------------------------- */
/* Display the request queue */
/* ------------------------- */

	ZZ_REQUEST         *r;
	ZZ_EVENT           *e;
	long               l;

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (isStationId (sid) && (e->station == NULL ||
			e->chain == NULL || ident (e->station) != sid))
				continue;

		if ((r = e->chain) == NULL) {
			// Generation event (system)
			if (zz_flg_nosysdisp || e->ai->Class != AIC_client ||
					(int)(e->Info02) != Id) continue;

			dtime (e->waketime);
			display ('*');
			display ("---");
			display (e->process->getTName ());
			display (e->process->Id);
			display (e->process->zz_sn (e->pstate));
			display (e->ai->zz_eid (e->event_id));
			display (e->ai->getTName ());
			if ((l = e->ai->Id) != NONE) {
				if (e->ai->Class == AIC_port)
					l = l & 0177777;
				display (l);
			} else
				display (' ');

			display (e->process->zz_sn (e->pstate));

		} else {
			
			// A station event

			while (1) {
				if ((r->ai->Class == AIC_client &&
					r->event_id == Id) ||
							r->ai == this) {

					dtime (r->when);

					if (e->waketime < r->when)
						display (' ');
					else if (e->chain == r)
						display ('*');
					else
						display ('?');

					if (!isStationId (sid)) {
						// Identify the station
						if (e->station != NULL &&
							ident (e->station) >= 0)
						   display (ident (e->station));
						else
						   display ("Sys");
					}
					display (e->process->getTName ());
					display (e->process->Id);
					display (e->process->zz_sn (r->pstate));
					display (r->ai->zz_eid (r->event_id));
					display (e->ai->getTName ());
					if ((l = e->ai->Id) != NONE) {
						if (e->ai->Class == AIC_port)
							l = l & 0177777;
						display (l);
					} else
						display (' ');

					display (e->process->zz_sn
							    (e->chain->pstate));
				}
				if ((r = r->other) == e->chain) break;
			}
		}
	}
}

void    Traffic::exDisplay1 () {

/* ---------------------------- */
/* Display performance measures */
/* ---------------------------- */

	if (FlgSPF == OFF) {
		display (' ');
		display ("OFF");
		return;
	}

	if (RVAMD == NULL) {            // Standard Client not started
		display (' ');
		return;
	}

	display ("AMD");
	RVAMD->displayOut (2);
	display ("APD");
	RVAPD->displayOut (2);
	display ("WMD");
	RVWMD->displayOut (2);
	display ("MAT");
	RVMAT->displayOut (2);
	display ("PAT");
	RVPAT->displayOut (2);
	display ("MLS");
	RVMLS->displayOut (2);
}

void    Traffic::exDisplay3 () {

/* -------------------------------- */
/* Display sent-received statistics */
/* -------------------------------- */

	display (NTMessages + NQMessages);
	display (NQMessages             );
	display (NTMessages             );
	display (NRMessages             );
	display (NTPackets              );
	display (NRPackets              );
	display (NQBits                 );
	display (NTBits                 );
	display (NRBits                 );
}

void    Traffic::exDisplay2 (int sid) {

/* ---------------------- */
/* Display message queues */
/* ---------------------- */


	int             i;
	long            bc, maxbc;
	Message         *m, **mm;

	if (isStationId (sid)) {

		// Station-specific display

		if ((mm = idToStation (sid)->MQHead) != NULL) {

			for (m = mm [Id]; m != NULL; m = m->next) {

				display (m->QTime);
				display (m->Length);
				if (isStationId (m->Receiver))
					display (m->Receiver);
				else if (m->Receiver == NONE)
					display ("none");
				else
					display ("bcast");
			}
		}

		return;
	}

	// Graphical display for all stations

	maxbc = 0;
	// Calculate the maximum
	for (i = 0; i < NStations; i++) {
		bc = 0;
		if (idToStation (i)->MQHead != NULL) {
			for (m = idToStation(i)->MQHead[Id]; m != NULL;
				m = m->next)
					bc += m->Length;
		}
		if (maxbc < bc) maxbc = bc;
	}

	if (bc == 0) bc = 1;

	startRegion (0.0, (double)(NStations-1), 0.0, (double) maxbc);
	startSegment (0);       // Unconnected points

	for (i = 0; i < NStations; i++) {
		bc = 0;
		if (idToStation (i)->MQHead != NULL) {
			for (m = idToStation(i)->MQHead[Id]; m != NULL;
				m = m->next)
					bc += m->Length;
		}
		displayPoint ((double)i, (double)bc);
	}

	endSegment ();
	endRegion ();
}
