/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ---------------- */
/* Observer service */
/* ---------------- */

#include        "system.h"

#if     ZZ_OBS

static  Process *phandle = NULL,        // Handle for observer service process
	*PFound;                        // For pttrav

enum { Start, ObsTimeout };             // States made global within this file

/* ----------------------------------------- */
/* Observer service process type declaration */
/* ----------------------------------------- */

class   ObserverService : public ZZ_SProcess {

	public:

	Station *S;

	static  int     zz_ObserverService_prcs;

	ObserverService () {
		S = TheStation;
		zz_typeid = (void*) (&zz_ObserverService_prcs);
	};

	virtual char *getTName () { return ("ObserverService"); };

	ZZ_INSPECT *ins, *jns;

	void setup () {

		// Initialize state name list

		zz_ns = 2;
		zz_sl = new char* [ObsTimeout + 1];
		zz_sl [0] = "Start";
		zz_sl [1] = "ObsTimeout";
	};

	void zz_code () {

	 switch (TheState) {

	  case Start:          // Void

	  break; case ObsTimeout:

		zz_current_observer = (Observer*) Info01;
		TheObserverState = (int) Info02;

		// Erase the inspects
		
		for (ins = zz_current_observer->inlistHead; ins != NULL;
			ins=jns) {

			jns = ins -> next;
			delete ((void*) ins);
		}

		zz_current_observer->inlistHead = zz_inlist_tail = NULL;

		// Clear inspect masks

		zz_current_observer->amask = zz_current_observer->pmask =
		zz_current_observer->nmask = zz_current_observer->smask = 0;

		zz_current_observer->tevent = NULL;     // No Timer event

		// Call the observer

		zz_observer_running = YES;

		do {
		  zz_jump_flag = NO;
		  zz_current_observer->zz_code ();
		  if (DisplayActive) {
		    // Observer stepping check
		    ZZ_SIT   *ws;

		    for (ws = zz_steplist; ws != NULL; ws = ws->next) {
		      if (ws->obs != zz_current_observer) continue;
		      if (ws->station != NULL && ws->station != TheStation)
			continue;
		      if (ws->process != NULL && ws->process != TheProcess)
			continue;
		      if (ws->ai != NULL && ws->ai != zz_CE->ai)
			continue;

		      zz_events_to_display = 0;
		      zz_send_step_phrase = YES;
		      break;
		    }
		  }
		  TheObserverState = zz_state_to_jump;
		} while (zz_jump_flag);

		zz_observer_running = NO;
	 }
	}

};

#endif

void    Observer::zz_start () {

#if     ZZ_OBS

	Class = AIC_observer;
	Id = sernum++;

	if (phandle == NULL) {

		Station *st;
		Process *tp;

		// Create the service process (belonging to the Kernel
		st = TheStation;
		tp = TheProcess;
		TheStation = System;
		TheProcess = Kernel;
		// phandle = create ObserverService;
		phandle = new ObserverService;
		((ObserverService*) phandle)->zz_start ();
		((ObserverService*) phandle)->setup ();
		TheStation = st;
		TheProcess = tp;
	}

	// Add the observer to the global observer list
	queue_head_AI (this, zz_obslist, Observer);
	inlistHead = zz_inlist_tail = NULL;
	tevent = NULL;
	smask = pmask = nmask = amask = 0L;

	// Add the observer to the owner's list
	zz_queue_head ((ZZ_Object*)this, TheProcess->ChList, ZZ_Object);
	// Notify the server about a new displayable object.
	zz_DADD (TheProcess, this);

	zz_observer_running = YES;
	zz_current_observer = this;
	zz_jump_flag = NO;
	TheObserverState = 0;                   // First time around
	zz_code ();
	while (zz_jump_flag) {  // Process jump
		zz_jump_flag = NO;
		TheObserverState = zz_state_to_jump;
		zz_code ();
	}
	zz_observer_running = NO;

#endif
}

char    *Observer::zz_sn (int) {

/* -------------------------------------- */
/* Returns the name of the state number i */
/* -------------------------------------- */

	// This is a virtual function
	return ("undefined");
}

void terminate (Observer *o) {

/* ---------------------- */
/* Terminates an observer */
/* ---------------------- */

#if     ZZ_OBS

	ZZ_INSPECT         *ins, *jns;

	Assert (o->ChList == NULL, "terminate: the observer has descendants");

	// Erase any inspects that the observer may have created

	for (ins = o->inlistHead; ins != NULL; ins=jns) {
		jns = ins -> next;
		delete ((void*) ins);
	}

	// Remove pending timeout event

	if (o->tevent != NULL) {
		o->tevent->cancel ();
		o->tevent = NULL;
	}

	// Remove the observer from the observer pool
	assert (o->ChList == NULL,
		form ("Can't terminate %s -- ownership list not empty",
			o->getOName ()));

	queue_out (o);
	queue_out ((ZZ_Object*)o);

	zz_DREM (o);
	if (o->zz_nickname != NULL)
		delete (o->zz_nickname);
	delete ((void*) o);
#endif
}

void zz_inspect_rq (Station *s, void *p, char *n, int pstate, int os) {

/* ---------------------------------------------------------- */
/* Issues an inspect request from an observer (called through */
/* a macro)                                                   */
/* ---------------------------------------------------------- */

#if  ZZ_OBS

	ZZ_INSPECT         *ins;

	if_not_from_observer ("inspect: called not from an observer");

	if (isStationId ((long) s))
		// Again this dirty trick, but there seems to be no way out
		s = idToStation ((long) s);

	// Update observer masks
	//
	// Note: the following instructions assume that words are (at least)
	//       32 bits long.

	if ((int) s == ANY)
		zz_current_observer->smask = 037777777777;
	else
		zz_current_observer->smask |= 1 << ((s->Id) & 037);

	if (p == (void*) (&Process::zz_Process_prcs)) {
		p = (void*) ANY;
		zz_current_observer->pmask = 037777777777;
	} else
		zz_current_observer->pmask |= 1 << ((((int) p) >> 2) & 037);

	if ((int)n == ANY)
		zz_current_observer->nmask = 037777777777;
	else
		zz_current_observer->nmask |= 1 << ((*n) & 037);

	if (pstate == ANY)
		zz_current_observer->amask = 037777777777;
	else
		zz_current_observer->amask |= 1 << (pstate & 037);

	// Create and queue the inspect request
	ins = new ZZ_INSPECT (s, p, n, pstate, os);
#endif
}

void    zz_timeout_rq (TIME t, int a) {

/* ----------------------------------------- */
/* Issues a timeout request from an observer */
/* ----------------------------------------- */

#if     ZZ_OBS

	if_not_from_observer ("timeout: called not from an observer");

	assert (zz_current_observer->tevent == NULL,
		"timeout: duplicate timeout request from an observer");

	zz_current_observer->tevent = new ZZ_EVENT (zz_sentinel_event,
		Time + t, System, (void*) zz_current_observer, (void*)a,
			phandle, zz_current_observer, OBS_TSERV, ObsTimeout,
				NULL);
#endif
}

void    zz_jump_rq (int a) {

/* --------------------------------------------- */
/* Resume the current observer at a given action */
/* --------------------------------------------- */

#if     ZZ_OBS

	if_not_from_observer ("jumpto: called not from an observer");
	zz_state_to_jump = a;
	zz_jump_flag    = YES;
#endif
}

char    *ZZ_INSPECT::getptype () {

#if     ZZ_OBS

/* ---------------------------------------------------------- */
/* Returns  the  type  name of the process represented by the */
/* inspect entry                                              */
/* ---------------------------------------------------------- */

	ZZ_Object  *o;
	char       *res;
	int        i;

	if (typeid == (void*)(ANY)) return ("ANY");

	for (i = 0; i < NStations; i++) {
		for (o = idToStation (i)->ChList; o != NULL; o = o->next) {
			if (o->Class == AIC_process) {
				if ((res = pttrav (o)) != NULL) {
					return (res);
				}
			}
		}
	}

	for (o = Kernel->ChList; o != NULL; o = o->next) {
		if (o->Class == AIC_process) {
			if ((res = pttrav (o)) != NULL) {
				return (res);
			}
		}
	}

	return ("unknown");
#else
#if	ZZ_ATT
	return (NULL);
#endif
#endif

}

char    *ZZ_INSPECT::pttrav (ZZ_Object *o) {

#if     ZZ_OBS

/* --------------------------------- */
/* A recursive searcher for getptype */
/* --------------------------------- */

	if (((Process*)o)->zz_typeid == typeid) {
		PFound = (Process*) o;
		return (o->getTName ());
	}

	for (ZZ_Object *b = ((Process*)o)->ChList; b != NULL; b = b->next)
		if (b->Class == AIC_process) return (pttrav (b));

	return (NULL);
#else
#if	ZZ_ATT
	return ((char*)o);
#endif
#endif
}

sexposure (Observer)

	sonpaper {

		sexmode (0)

			exPrint0 (Hdr);         // Global information

		sexmode (1)

			exPrint1 (Hdr);         // Inspect list
	}

	sonscreen {

		sexmode (0)

			exDisplay0 ();          // Global information

		sexmode (1)

			exDisplay1 ();          // Inspect list
	}
	USESID;
}

void    Observer::exPrint0 (char *hdr) {

#if     ZZ_OBS

/* ------------------------------------- */
/* Print information about all observers */
/* ------------------------------------- */

	Observer                        *o;
	ZZ_INSPECT                      *in;

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << "(Observer) Inspect lists of all observers:\n\n";
	}

	Ouf << "            Observer    St      PType  PNickname      State" <<
		"     OState\n";

	for (o = zz_obslist; o != NULL; o = o -> next) {
		for (in = o->inlistHead; in != NULL; in = in -> next) {

			if (in == o->inlistHead)
				// First time for this observer
				print (o->getOName (), 20);
			else
				Ouf << "                    ";
			Ouf << ' ';
			if ((int)(in->station) == ANY)
				print ("ANY", 5);
			else
				print (zz_trunc (ident (in->station), 5), 5);
			Ouf << ' ';
	
			PFound = NULL;  // Will be set by getptype
			print (in->getptype (), 10);

			Ouf << ' ';
			if ((int)(in->nickname) == ANY)
				print ("ANY", 10);
			else
				print (in->nickname, 10);
			Ouf << ' ';
			if (in->pstate == ANY)
				print ("ANY", 10);
			else if (PFound != NULL)
				print (PFound->zz_eid (in->pstate), 10);
			else
				print (zz_trunc (in->pstate, 10), 10);
			Ouf << ' ';
			print (o->zz_sn (in->ostate), 10);
			Ouf << '\n';
		}
		if (o->tevent != NULL) {
			// Display timeout info
			if (o->inlistHead == NULL)
				print (o->getOName (), 20);
			else
				Ouf << "                    ";
			Ouf << "    >> Timeout at ";
			ptime (tevent->waketime, 15);
			Ouf << " <<\n";
		}
	}

	Ouf << "\n(Observer) End of list\n\n";
#endif
}

void    Observer::exPrint1 (char *hdr) {

#if     ZZ_OBS

/* --------------------------------------- */
/* Print the inspect list of this observer */
/* --------------------------------------- */

	ZZ_INSPECT                      *in;

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () << ") Inspect list:\n\n";
	}

	Ouf << "    St      PType         Id      State     OState\n";

	for (in = inlistHead; in != NULL; in = in -> next) {

		if ((int)(in->station) == ANY)
			print ("ANY", 5);
		else
			print (zz_trunc (ident (in->station), 5), 5);
		Ouf << ' ';
	
		PFound = NULL;
		print (in->getptype (), 10);

		Ouf << ' ';
		if ((int)(in->nickname) == ANY)
			print ("ANY", 10);
		else
			print (in->nickname, 10);
		Ouf << ' ';
		if (in->pstate == ANY)
			print ("ANY", 10);
		else if (PFound != NULL)
			print (PFound->zz_eid (in->pstate), 10);
		else
			print (zz_trunc (in->pstate, 10), 10);

		Ouf << ' ';
		print (zz_sn (in->ostate), 10);
		Ouf << '\n';
	}

	if (tevent != NULL) {
		// Display timeout info
		Ouf << "    >> Timeout at ";
		ptime (tevent->waketime, 15);
		Ouf << " <<\n";
	}

	Ouf << "\n(" << getOName () << ") End of list\n\n";
#endif
}

void    Observer::exDisplay0 () {

#if     ZZ_OBS

/* ------------------------------------- */
/* Print information about all observers */
/* ------------------------------------- */

	Observer                        *o;
	ZZ_INSPECT                      *in;

	for (o = zz_obslist; o != NULL; o = o -> next) {
		for (in = o->inlistHead; in != NULL; in = in -> next) {

			if (in == o->inlistHead)
				// First time for this observer
				display (o->getOName ());
			else
				display (' ');

			if ((int)(in->station) == ANY)
				display ("ANY");
			else
				display (ident (in->station));
	
			PFound = NULL;
			display (in->getptype ());

			if ((int)(in->nickname) == ANY)
				display ("ANY");
			else
				display (in->nickname);

			if (in->pstate == ANY)
				display ("ANY");
			else if (PFound != NULL)
				display (PFound->zz_eid (in->pstate));
			else
				display (in->pstate);

			display (o->zz_sn (in->ostate));
		}
		if (o->tevent != NULL) {
			// Display timeout info
			if (o->inlistHead == NULL)
				display (o->getOName ());
			else
				display (' ');

			display ("Timeout at");
			dtime (tevent->waketime);
			display (' ');
			display (' ');
			display (' ');
		}
	}
#endif
}

void    Observer::exDisplay1 () {

#if     ZZ_OBS

/* ----------------------------------------- */
/* Display the inspect list of this observer */
/* ----------------------------------------- */

	ZZ_INSPECT                      *in;

	for (in = inlistHead; in != NULL; in = in -> next) {

		if ((int)(in->station) == ANY)
			display ("ANY");
		else
			display (ident (in->station));
	
		PFound = NULL;
		display (in->getptype ());

		if ((int)(in->nickname) == ANY)
			display ("ANY");
		else
			display (in->nickname);

		if (in->pstate == ANY)
			display ("ANY");
		else if (PFound != NULL)
			display (PFound->zz_eid (in->pstate));
		else
			display (in->pstate);

		display (zz_sn (in->ostate));
	}

	if (tevent != NULL) {
		// Display timeout info
		display ("Timeout at");
		dtime (tevent->waketime);
		display (' ');
		display (' ');
		display (' ');
	}
#endif
}
