/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ------------------------------ */
/* Global variables (definitions) */
/* ------------------------------ */

#include        <signal.h>
#include        <setjmp.h>
#include        "global.h"
#include        "display.h"

/* ------------ */
/* Standard AIs */
/* ------------ */

zz_timer        zz_AI_timer;
zz_client       zz_AI_client;
zz_queue        zz_AI_queue;
zz_signal       zz_AI_signal;

ZZ_SYSTEM       *System = NULL;         // The system station
ZZ_KERNEL       *Kernel = NULL,         // The system process
		*ZZ_Kernel = NULL;      // Another copy of the above pointer
Process         *ZZ_Main = NULL;        // The user's root process

/* ----------- */
/* Environment */
/* ----------- */
double          Itu = 1.0, Etu = 1.0;   // Time units
TIME            Time = TIME_0;          // The simulated time
Station         *TheStation = NULL;     // The current station
Process         *TheProcess = NULL;     // The current process
int             TheState,               // The process state
		TheObserverState;       // The observer state
long            MemoryUsed = 0;         // Dynamicaly allocated storage

void            *Info01 = NULL,         // GP environment variables
		*Info02 = NULL;

ZZ_EVENT        *zz_CE;                 // Current event

/* ------ */
/* Counts */
/* ------ */
long            NStations = 0,          // The total number of stations
		NLinks = 0,             // The number of links
		NTraffics = 0;          // The number of message patterns

Station         **zz_st = NULL;         // Station table
Link            **zz_lk = NULL;         // Link table

Traffic         **zz_tp = NULL;         // Array of Traffics

/* ------------------------------ */
/* Global counters for the Client */
/* ------------------------------ */
long            zz_NQMessages,          // Number of queued messages
		zz_NTMessages,          // Number of transmitted messages
		zz_NRMessages,          // Number of received messages
		zz_NTPackets,           // Number of transmitted packets
		zz_NRPackets,           // Number of received packets
		zz_NGMessages;          // Number of generated messages

BITCOUNT        zz_NQBits,              // Number of queued bits
		zz_NTBits,              // Number of transmitted bits
		zz_NRBits;              // Number of received bits

/* ---------------------------------------------------------- */
/* Temporary list heads for creating station ports and packet */
/* buffers                                                    */
/* ---------------------------------------------------------- */
Port            *zz_ncport = NULL;
ZZ_PFItem       *zz_ncpframe = NULL;

/* --------- */
/* Debugging */
/* --------- */
TIME            TracingTime;            // Trace flag (initially off)
int             FullTracing = NO,       // Full trace flag
		EndOfData = NO;         // EOF flag

/* ------------------------------------- */
/* Boundary conditions of the simulation */
/* ------------------------------------- */
long    zz_max_NMessages = MAX_long;    // Default limits and flags
TIME    zz_max_Time;                    // Everything is by default
double  zz_max_cpuTime   = HUGE;        // unlimited
int     zz_exit_code     = EXIT_user;   // Termination code

/* ----------------------- */
/* Program call parameters */
/* ----------------------- */
char    **zz_callArgs;
int     zz_callArgc;                    // Argument count

	// Original RNG seeds to be included in error messages; initialized
	// in a standard way (changeable by a call parameter)
long    zz_OrgSeeds [ZZ_N_SEEDS] = {1777, 98733, 111329};

#if     ZZ_R48

	// Working seeds

#if     ZZ_ATT
unsigned  short  zz_Seeds [ZZ_N_SEEDS] [ 3 ];
#else
short  zz_Seeds [ZZ_N_SEEDS] [ 3 ];
#endif

#else
long    zz_Seeds [ZZ_N_SEEDS];

	// A divisor for rnd. This is the minimum double number greater than
	// MAX_long such that (double)MAX_long divided by it is less than 1.0
double  zz_rndd;
#endif

/* ----- */
/* Flags */
/* ----- */
	// A flag used to tell whether message queues should be
	// emptied after zz_max_NMessages have been generated
int     zz_flg_emptyMQueues = NO,

	// Tells whether definitions of standard objects should be
	// automatically echoed to the output file
	zz_flg_printDefs = NO,

	// Tells whether the standard Client is to be started. Can be
	// switched off by a program call parameter.
	zz_flg_stdClient = YES,

	// Tells whether information about system events should be
	// suppressed from exposures (which is the default)
	zz_flg_nosysdisp = YES,

	// Tells whether the definition of stations is complete and the
	// protocol (the first protocol process) has been started.
	zz_flg_started = NO,

	// Tells whether the structure of the ownership three has been
	// modified by zz_adjust_ownership.
	zz_flg_otreeAdjusted = NO;

#if     ZZ_QSL
	// The global limit on the number of messages to be queued at stations
	// awaiting transmission. When this number is reached new generated
	// messages are ignored.
extern  long    zz_QSLimit = MAX_long;
#endif

/* ---------------- */
/* Event processing */
/* ---------------- */
ZZ_EVENT   *zz_pe = NULL,               // Priority event
	   *zz_c_wait_event,            // Currently built event
	   zz_fsent,                    // Front sentinel event
	   zz_rsent,                    // Rear sentinel event
	   *zz_sentinel_event;          // Pointer to the rear sentinel

ZZ_REQUEST *zz_pr = NULL;               // Priority request

long    zz_npee = 0,                    // Number of pending events
	zz_npre = 0,                    // Number of processed events
	zz_event_id;                    // Current event id

int     zz_c_first_wait = YES,          // Wait issued for the first time
	zz_port_event_id = NONE,        // Last port event id
	zz_chckint;                     // Time limit check interval
					// Also used as checkpoint rqst flag

ZZ_REQUEST      *zz_c_other,            // Next wait request pointer
		*zz_c_whead,            // Built wait list head
		*zz_nqrqs;              // Dummy head for some request lists

TIME    zz_c_wait_tmin;                 // Earliest request so far

AI      *zz_ai;                         // The current AI pointer

#if     ZZ_OBS
/* ------------------- */
/* Observers internals */
/* ------------------- */
Observer   *zz_obslist = NULL,          // The list of all observers
	   *zz_current_observer;        // The observer currently running

int        zz_observer_running = NO,    // YES, if an observer is running
	   zz_jump_flag = NO,           // Resume in progress
	   zz_state_to_jump;            // The observer's state to jump to

ZZ_INSPECT *zz_inlist_tail;             // The tail of the created inlist

#endif
#if     ZZ_TOL
/* -------------------------- */
/* Clock tolerance parameters */
/* -------------------------- */
int    zz_quality   = 0;                // The default clock quality (perfect)
double zz_tolerance = 0.0;              // The default clock tolerance (none)
#endif

/* ---------------------- */
/* Display (visible part) */
/* ---------------------- */
volatile int    DisplayActive = NO;     // Display activity flag

int     DisplayInterval = DEFDINTVL,    // Display interval
	DisplayOpening = NO,            // A brand new window
	DisplayClosing = NO;            // Closing the window

/* ------------------------ */
/* Display (invisible part) */
/* ------------------------ */

ZZ_WINDOW  *zz_windows = NULL,          // Active windows
	   *zz_the_window = NULL;       // The current window
ZZ_SIT     *zz_steplist = NULL;         // Stepped windows
ZZ_WSIT    *zz_steplist_delayed = NULL; // Waiting to be stepped
ZZ_SUPPLI  *zz_supplist = NULL;         // The supplement list
ZZ_SUPPLD  *zz_suppldst = NULL;         // The supplement list
int        zz_suppcount = 0;            // The number of supplements
long       zz_commit_event;             // Starting event number
TIME       zz_commit_time;              // Starting time
volatile long zz_events_to_display = 0; // Delay to next display refresh
int        zz_ctimeout,                 // Server i/o timeout
	   zz_drequest = NO,            // Display requested by '-d'
	   zz_send_step_phrase = NO,    // Flag: stepped window sent
	   zz_dispfound;                // Flag: exmode found

/* ---------------- */
/* Standard streams */
/* ---------------- */
istream *zz_ifpp = NULL;                // Input stream
ostream *zz_ofpp = NULL;                // Output file

char    *zz_ifn = NULL,                 // Pointer to input filename
	*zz_ofn = NULL;                 // Pointer to output filename

#if     ZZ_SMURPHDEBUG  == 2
ostream *zz_dbgf;
#endif

#if     ZZ_ATT
/* ---------------------------------------------------------- */
/* The constant below specifies the maximum nesting of create */
/* for  the  ATT  version.  It  is never checked whether this */
/* maximum is not exceeded. It is assumed  that  nobody  sane */
/* will even get close to this number.                        */
/* ---------------------------------------------------------- */
#define MXOCLEVEL               16
/* --------------------------------------- */
/* A dummy object pointer (used by create) */
/* --------------------------------------- */
void    *zz_COBJ [MXOCLEVEL];
int     zz_clv = 0;
#endif

/* ----------------- */
/* Smurph process id */
/* ----------------- */
int     zz_processId;

/* ------------------- */
/* Initialized statics */
/* ------------------- */
long    EObject::sernum = 0;
long    RVariable::sernum = 0;
long    Station::sernum = 0;
long    Link::sernum = 0;
long    Traffic::sernum = 0;
long    Process::sernum = 0;
long    Observer::sernum = 0;

#if     ZZ_ATT
	// The variable below is supposed to be used as a garbage location
	// to force usage of unused parameters that Cfront complains about
long    zz_Stroustrup = 0;
#include        "inlines.h"
#endif
